/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* ###--------------------------------------------------------------### */
/* file		: c_fsyn_sr4k_10.c					*/
/* date		: Jun  3 1997						*/
/* version	: v2.02							*/
/* author	: Pirouz Bazargan Sabet					*/
/* content	: C model of a circuit					*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include BEH_H
#include "vh_ltype.h"
#include "vh_lspec.h"
#include "vh_xspec.h"

/* ###--------------------------------------------------------------### */
/* function	: c_fsyn_sr4k_10					*/
/* description	: 4K synchronous (clock's falling edge) static 10-bit	*/
/*		  Ram							*/
/* called func.	: vhx_getvalu, vhx_readinput, vhx_writebus		*/
/* ###--------------------------------------------------------------### */

void c_fsyn_sr4k_10 (pt_lkdins)

struct lkdins *pt_lkdins;

  {
  char         ck    ;			/* clock			*/
  unsigned int d_out ;			/* output 10-bit data		*/
  unsigned int d_in  ;			/* input  10-bit data		*/
  unsigned int adr   ;			/* 12-bit address		*/
  char         oe    ;			/* output enable		*/
  char         we    ;			/* write enable			*/

#define POSI_we                 0
#define POSI_oe    POSI_we   +  1
#define POSI_adr   POSI_oe   +  1
#define POSI_d_in  POSI_adr  + 12
#define POSI_ck    POSI_d_in + 10

#define POSO_d_out              0

  struct local_mem
    {
    unsigned int   LAST_CK       ;
    unsigned short RAM     [4096];
    };

  struct local_mem *local   ;
  static void       init () ;

	/* ###------------------------------------------------------### */
	/*   get the memory space reserved for the current instance	*/
	/* ###------------------------------------------------------### */

  vhx_getvalu (pt_lkdins, &local, sizeof(struct local_mem), init);

	/* ###------------------------------------------------------### */
	/*   read the current value of input signals (ck)		*/
	/* ###------------------------------------------------------### */

  ck = vhx_readinput (pt_lkdins, POSI_ck, 1);

  if (ck == 1)
    local->LAST_CK = 1;
  else
    {
	/* ###------------------------------------------------------### */
	/*   check for the falling edge of the clock			*/
	/* ###------------------------------------------------------### */

    if (local->LAST_CK == 1)
      {
      local->LAST_CK = 0;

      oe   = vhx_readinput (pt_lkdins, POSI_oe  ,  1);
      we   = vhx_readinput (pt_lkdins, POSI_we  ,  1);
      adr  = vhx_readinput (pt_lkdins, POSI_adr , 12);

      if (we == 1)
        {
	/* ###------------------------------------------------------### */
	/*   write the input data into the ram				*/
	/* ###------------------------------------------------------### */

        d_in = vhx_readinput (pt_lkdins, POSI_d_in, 10);
        local->RAM [adr] = d_in;
        }
      else
        {
	/* ###------------------------------------------------------### */
	/*   read the content of the ram and write the result into the	*/
	/* projected value of ouput data				*/
	/* ###------------------------------------------------------### */

        if (oe == 1)
          d_out = local->RAM [adr];
        else
          d_out = 0;

        vhx_writeout (d_out, pt_lkdins, POSO_d_out, 10);
        }
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: init							*/
/* description	: initializations					*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static void init (pt_lkdins, pt_mem)

struct lkdins *pt_lkdins;
char          *pt_mem   ;

  {
  printf ("coucou %s\n", pt_lkdins->INSNAME);
  }
