/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* ###--------------------------------------------------------------### */
/* file		: vh_asimut.c						*/
/* date		: Dec 13 1999						*/
/* version	: v3.02							*/
/* author	: Huu Nghia VUONG, Pirouz BAZARGAN SABET		*/
/* description	: VHDL simulation tool					*/
/* ###--------------------------------------------------------------### */

#ident "@(#) ASIMUT : A SIMUlation Tool V2.02"
#include <stdio.h>
#include <string.h>
#include MUT_H
#include MLO_H
#include MLU_H
#include LOG_H
#include BEH_H
#include BHL_H
#include BVL_H
#include PAT_H
#include PPT_H
#include "vh_util.h"
#include "vh_ltype.h"
#include "vh_asimut.h"
#include "vh_asdef.h"

extern char MVL_ERRFLG;

main (argc, argv)

int   argc;
char **argv;

  {
  int                i , j , patindx;
  char               asimut_err_flg;
  int                nbr_errlft;		/* how many errors left	*/
						/* before aborting	*/
  unsigned int       max_delta   = VHX_DLTDFN;	/* max # of delta cyc.	*/

  char               usage_flg   = 0;		/* print usage		*/
  char               help_flg    = 0;		/* display help		*/
  char               init_flg    = 0;		/* description's init.	*/
  char               core_flg    = 0;		/* core file at error	*/
  char               sdebug_flg  = 0;		/* debug structural	*/
  char               bdebug_flg  = 0;		/* debug behavioural	*/
  char               pdebug_flg  = 0;		/* debug pattern	*/
  char               ldebug_flg  = 0;		/* debug linker		*/
  char               cdebug_flg  = 0;		/* dbeug cycle by cycle	*/
  char               ddebug_flg  = 0;		/* debug delta by delta	*/
  char               behav_flg   = 0;		/* behav. description	*/
  char               nosim_flg   = 0;		/* no simulation	*/
  char               nores_flg   = 0;		/* no result file	*/
  char               makbdd_flg  = 0;		/* trace bdd making	*/
  char               bdd_flg     = 0;		/* use bdd or gex	*/
  char               level_flg   = 0;		/* levelize or not	*/
  char               sav_end     ;		/* saved end flag	*/
  int                max_err     = 10;		/* # of simul. errors	*/
  unsigned int       max_pat     = 0 ;		/* # of patterns	*/
  unsigned int       labelsiz    = 15;		/* longest patt. label	*/
  char              *arg_flg     = NULL;	/* recongnized arg. flg	*/
  unsigned int       last_opt    = 0;		/* last option's index	*/
  unsigned int       frst_fil    = 0;		/* first file's index	*/
  unsigned int       init_val    = 0;		/* init. constant	*/

  unsigned int       comp_mode   = 0;		/* beh compiler mode	*/
  struct lofig      *ptlofig     ;
  struct befig      *ptbefig     = NULL;
  struct befig      *ptbef       ;
  struct chain      *ptchain     ;
  struct papat      *pt_papat    ;
  struct paseq      *pt_paseq    ;
  struct pains      *pt_pains    = NULL;	/* inspected instances	*/
  struct pains      *tmp_pains   = NULL;
  char              *str;
  char               filename [200];
  FILE              *filepnt     ;
  struct beden     **hshtab      ;

	/* ###------------------------------------------------------### */
	/*    analyse the command line, set option flags and find the	*/
	/* first file in the argument list				*/
	/* ###------------------------------------------------------### */

  arg_flg = (char *) vhu_alloc (argc * sizeof (char), VHU_NOALIGN);

  arg_flg [0] = 1;
  for (i=1 ; i<argc ; i++)
    arg_flg [i] = 0;

  for (i=1 ; i<argc ; i++)
    {
    if (arg_flg [i] == 0)
      {
      if (!strcmp (argv[i], "-h"))
        {
        help_flg    = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-p"))
        {
        sscanf (argv[i+1], "%u", &max_pat);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-l"))
        {
        sscanf (argv[i+1], "%u", &labelsiz);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-t"))
        {
        makbdd_flg  = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-i"))
        {
        init_flg      = i + 1;
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-core"))
        {
        core_flg      = i + 1;
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-inspect"))
        {
        str      = namealloc (argv [i+1]);
        pt_pains = pat_addpains (pt_pains, str, str);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-level"))
        {
        level_flg   = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-nores"))
        {
        nores_flg   = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-c"))
        {
        nosim_flg   = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-bdd"))
        {
        bdd_flg     = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-b"))
        {
        behav_flg   = 1;
        arg_flg [i] = 1;
        }

      if (!strncmp (argv[i], "-d", 2))
        {
        arg_flg [i] = 1;
        for (j=2 ; argv[i][j] != '\0' ; j++)
          {
          switch (argv [i] [j])
            {
            case 's':
              sdebug_flg = 1; break;
            case 'b':
              bdebug_flg = 1; break;
            case 'l':
              ldebug_flg = 1; break;
            case 'p':
              pdebug_flg = 1; break;
            case 'c':
              cdebug_flg = 1; break;
            case 'd':
              ddebug_flg = 1; break;
            }
          }
        }
      }
    }

  for (i=0 ; i<argc ; i++)
    {
    if (arg_flg [i] == 1)
      last_opt = i;
    }

  for (frst_fil=0 ; frst_fil < argc ; frst_fil++)
    {
    if (arg_flg [frst_fil] == 0)
      break;
    }

	/* ###------------------------------------------------------### */
	/* cases of error in the command line:				*/
	/*   - no argument						*/
	/*   - argument unrecognized as an option			*/
	/*   - number of files do not match:				*/
	/*       - simulation with result    => 3 files needed		*/
	/*       - simulation without result => 2 files needed		*/
	/*       - no simulation             => 1 file  needed		*/
	/* ###------------------------------------------------------### */

  if (help_flg == 1)
    {
    vhu_message (0, NULL, 0);
    system      ("man asimut");
    exit (0);
    }

  if ((argc == 1) || (last_opt != (frst_fil - 1)))
    usage_flg = 1;

  if (((nosim_flg == 0) && (nores_flg == 0) && (argc != (frst_fil + 3))) ||
      ((nosim_flg == 0) && (nores_flg == 1) && (argc != (frst_fil + 2))) ||
      ((nosim_flg == 1) &&                     (argc != (frst_fil + 1))))
    usage_flg = 1;

  if (usage_flg == 1)
    {
    vhu_message (0, NULL, 0);
    vhu_message (1, NULL, 0);
    exit (1);
    }

  alliancebanner ("ASimuT", vhu_avers(), "A SIMUlation Tool", "1991...1999",
                  ALC_V);

  vhu_message (21, NULL, 0);

	/* ###------------------------------------------------------### */
	/*    Initialize :						*/
	/*      - BDD hash tables					*/
	/*      - get MBK environment variables				*/
	/*      - set the fast mode on MBK functions			*/
	/*      - set the behavioural compiler's mode			*/
	/* ###------------------------------------------------------### */

  vhu_message (2, NULL, 0);
  initializeBdd (0);
  mbkenv ();

  vhi_models ();

  FAST_MODE = 'Y';
  comp_mode = BVL_TRACE | BVL_KEEPAUX;

	/* ###------------------------------------------------------### */
	/*    Read the maximum number of simulation errors from the	*/
	/* environment variable VH_MAXERR				*/
	/* ###------------------------------------------------------### */

  str = getenv ("VH_MAXERR");
  if (str != NULL)
    {
    if (sscanf (str, "%u", &max_err) == 0)
      max_err = 10;
    }

	/* ###------------------------------------------------------### */
	/*    Searching the root file					*/
	/* ###------------------------------------------------------### */

  vhu_message (5, argv[frst_fil], 0);
  
	/* ###------------------------------------------------------### */
	/*    If there is no structural level call directly the		*/
	/* behaviour compiler						*/
	/* ###------------------------------------------------------### */

  if (behav_flg == 1)
    {

	/* ###------------------------------------------------------### */
	/*    run the behavioural compiler on the current file. Call	*/
	/* the debugger is necessary					*/
	/* ###------------------------------------------------------### */

    if (vhl_iscmodel (argv[frst_fil]) == NULL)
      comp_mode = BVL_TRACE | BVL_KEEPAUX ;
    else
      comp_mode = BVL_TRACE | BVL_KEEPAUX | BVL_CHECKEMPTY;

    ptbefig        = vhdlloadbefig (ptbefig, argv[frst_fil], comp_mode);
    asimut_err_flg = ptbefig->ERRFLG;

    if (bdebug_flg == 1)
      dast_dbg (ptbefig, "befig");

	/* ###------------------------------------------------------### */
	/*    if the architecture of the compiled description is VHDL	*/
	/* and if there is no error in the description :		*/
	/*								*/
	/*    - identify dependencies in the description		*/
	/*    - transform boolean expression into BDDs or GEXs		*/
	/* ###------------------------------------------------------### */

    if ((ptbefig->ERRFLG == 0) && (ptbefig->FLAG == BEH_ARC_VHDL))
      {
      beh_depend (ptbefig);
      if (bdd_flg == 1)
        {
        vhu_message (15, "BDD", 0);
        beh_makbdd  (ptbefig);
        }
      else
        {
        vhu_message (15, "GEX", 0);
        beh_makgex  (ptbefig);
        }

      beh_freabl  (ptbefig);
      asimut_err_flg = ptbefig->ERRFLG;
      }

    ptlofig = vhl_makgst (ptbefig);

    }
  else
    {
	/* ###------------------------------------------------------### */
	/*    run the structural compiler on the root file and call the	*/
	/* debugger if it has been asked. Flatten the root description	*/
	/* and call the structural compiler as many times as necessary	*/
	/* to reach leaf cells (behavioural description)		*/
	/* ###------------------------------------------------------### */

    vhu_message (4, argv[frst_fil], 0);
    
    VHS_LOFPNT = getlofig (argv[frst_fil], 'A');
    asimut_err_flg = MVL_ERRFLG;

    if (sdebug_flg == 1)
      dast_dbg (HEAD_LOFIG, "lofig");

    if (asimut_err_flg != 0)
      {
      asimut_err_flg = vhu_error (200, NULL, NULL);
      exit (1);
      }

    vhu_message (18, NULL, 0);

    lofigchain (VHS_LOFPNT);
    rflattenlofig (VHS_LOFPNT, YES, YES);

    if (sdebug_flg == 1)
      dast_dbg (HEAD_LOFIG, "lofig");

    ptchain = VHS_LOFPNT->MODELCHAIN;
    while (ptchain != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    for each leaf cell of the description search the		*/
	/* behavioural file						*/
	/* ###------------------------------------------------------### */

      vhu_message (5, ptchain->DATA, 0);

	/* ###------------------------------------------------------### */
	/*    run the behavioural compiler on the current file		*/
	/* ###------------------------------------------------------### */

      ptbef = ptbefig;

      if (vhl_iscmodel ((char *)ptchain->DATA) == NULL)
        comp_mode = BVL_TRACE | BVL_KEEPAUX ;
      else
        comp_mode = BVL_TRACE | BVL_KEEPAUX | BVL_CHECKEMPTY;

      ptbefig         = vhdlloadbefig (ptbefig, ptchain->DATA, comp_mode);
      ptbefig->NEXT   = ptbef;
      asimut_err_flg += ptbefig->ERRFLG;

	/* ###------------------------------------------------------### */
	/*    if the architecture of the compiled description is VHDL	*/
	/* and if there is no error in the description ;		*/
	/*								*/
	/*    - identify dependencies in the description		*/
	/*    - transform boolean expression into BDDs or GEXs		*/
	/* ###------------------------------------------------------### */

      if ((asimut_err_flg == 0) && (ptbefig->FLAG == BEH_ARC_VHDL))
        {
        beh_depend (ptbefig);
        if (bdd_flg == 1)
          {
          vhu_message (15, "BDD", 0);
          beh_makbdd  (ptbefig);
          }
        else
          {
          vhu_message (15, "GEX", 0);
          beh_makgex  (ptbefig);
          }

        beh_freabl  (ptbefig);
        asimut_err_flg = ptbefig->ERRFLG;
        }
      ptchain = ptchain->NEXT;
      }
    }

	/* ###------------------------------------------------------### */
	/*    call the debugger if necessary				*/
	/* ###------------------------------------------------------### */

  if (bdebug_flg == 1)
    dast_dbg (ptbefig, "befig");

  if (asimut_err_flg != 0)
    {
    asimut_err_flg = vhu_error (200, NULL, NULL);
    exit (1);
    }

  if (nosim_flg == 1)
    exit (0);

	/* ###------------------------------------------------------### */
	/*    Running the pattern description compiler on the current	*/
	/* file								*/
	/* ###------------------------------------------------------### */

  vhu_message (6, argv[frst_fil + 1], 0);

  pt_paseq = pat_lodpaseq (argv[frst_fil + 1], NULL, max_pat);

  if ((pt_paseq == NULL) || (pt_paseq->ERRFLG != 0))
    {
    asimut_err_flg = vhu_error (200, NULL, NULL);
    exit (1);
    }

	/* ###------------------------------------------------------### */
	/*    add the list of inspected instances passed as argument to	*/
	/* the list of inspected instances found in the pattern file	*/
	/* ###------------------------------------------------------### */

  if (pt_pains != NULL)
    {
    tmp_pains = pt_pains;
    while (tmp_pains->NEXT != NULL)
      tmp_pains = tmp_pains->NEXT;

    tmp_pains->NEXT = pt_paseq->PAINS;
    pt_paseq->PAINS = pt_pains;
    }

  if (pdebug_flg == 1)
    dast_dbg (pt_paseq, "paseq");

	/* ###------------------------------------------------------### */
	/*    Link previous structures to create the simulator's data	*/
	/* base structures						*/
	/* ###------------------------------------------------------### */

  hshtab = beh_initab ();

  vhu_message (16, NULL, 0);

  if (behav_flg == 0)
    vhl_resord (hshtab, HEAD_LOFIG, ptbefig);

  if (sdebug_flg == 1)
    dast_dbg (HEAD_LOFIG, "lofig");

  asimut_err_flg = vhl_restor (hshtab, HEAD_LOFIG, ptbefig);

  if (sdebug_flg == 1)
    dast_dbg (HEAD_LOFIG, "lofig");

  if (asimut_err_flg != 0)
    {
    asimut_err_flg = vhu_error (200, NULL, NULL);
    exit (1);
    }

  vhu_message (9, NULL, 0);

  asimut_err_flg = vhl_linker (hshtab, pt_paseq);

  if (asimut_err_flg != 0)
    {
    asimut_err_flg = vhu_error (200, NULL, NULL);
    exit (1);
    }

	/* ###------------------------------------------------------### */
	/*    levelize the circuit if needed.				*/
	/* ###------------------------------------------------------### */

  if (level_flg != 0)
    {
    vhx_levelize (VHL_HEDLKF, bdd_flg);
    VHX_LEVEL = 0;
    }

/*-----
dast_dbg (VHL_HEDLKF, "lkdfig");
printf ("\n\t=============== after levelize ==============\n");
traceX ();
-------*/

	/* ###------------------------------------------------------### */
	/*    Initialize the circuit with a .sav file or a constant.	*/
	/* ###------------------------------------------------------### */

  if (init_flg != 0)
    {
    if ((sscanf (argv[init_flg], "%u", &init_val) == 1) && (init_val <= 1))
      vhx_initval (VHL_HEDLKF, init_val + '0');
    else
      {
      if ((filepnt = mbkfopen (argv[init_flg], "sav", READ_TEXT)) == NULL)
        vhu_error (116, argv [init_flg], NULL);
      else
        {
        asimut_err_flg = vhx_initfig (VHL_HEDLKF, filepnt);
        fclose (filepnt);
        }
      }
    }
  else
    vhx_initval (VHL_HEDLKF, '0');

  if (ldebug_flg == 1)
    dast_dbg (VHL_HEDLKF, "lkdfig");

  if (pdebug_flg == 1)
    dast_dbg (pt_paseq, "paseq");

  if (asimut_err_flg != 0)
    {
    asimut_err_flg = vhu_error (200, NULL, NULL);
    exit (1);
    }

	/* ###------------------------------------------------------### */
	/*    Execute the current figure				*/
	/* ###------------------------------------------------------### */

  vhx_execall (VHL_HEDLKF);
  patindx = 0;

  nbr_errlft = max_err;

  while (1)
    {
    pt_papat = pt_paseq->CURPAT;

    while ((pt_papat   != NULL) && (pt_papat != pt_paseq->NXTPAT) &&
           (nbr_errlft >  0))
      {
      max_delta = VHX_DLTDFN;

	/* ###------------------------------------------------------### */
	/*    Read a new pattern					*/
	/* ###------------------------------------------------------### */

      vhu_message (10, NULL, patindx);

      if (pt_papat->ACTFLAG == 'I')
        vhx_inireg (hshtab, pt_papat->PAINI);

      VHX_LEVEL = 0;
      vhx_redpat  (pt_paseq, pt_papat);

	/* ###------------------------------------------------------### */
	/*    Execute the pattern					*/
	/*      - first update signals (value read from the pattern)	*/
	/*      - then make execute-update iteration until no new	*/
	/*        event is generated (if the maximum number of		*/
	/*        iterations (delta-cycles) is reached and events	*/
	/*        continue to be produced abort the simulation with a	*/
	/*        'combinatory loop' error)				*/
	/* ###------------------------------------------------------### */

      vhx_update ();

      if (ddebug_flg == 1)
        dast_dbg (VHL_HEDLKF, "lkdfig");

/*-------
printf ("\n\t====== delta ====\n");
traceX ();
-------*/

      while ((vhx_execute (bdd_flg) == 0) && (max_delta > 0))
        {
        vhx_update ();
/*------
printf ("\n\t====== delta ====\n");
traceX ();
--------*/
        max_delta--;

/*--------------------
      if ((max_delta % 100) == 0)
        printf ("------ max_delta : %d\n", max_delta);

      if (max_delta <= 9000)
        {
        printf ("---------------- delta : %d\n", max_delta);
        vhx_core (stdout, VHL_HEDLKF);
        }
--------------------*/

        if (ddebug_flg == 1)
          dast_dbg (VHL_HEDLKF, "lkdfig");
        }

	/* ###------------------------------------------------------### */
	/*    Save the new state of the circuit after the pattern has	*/
	/* been processed						*/
	/* ###------------------------------------------------------### */

      if (cdebug_flg == 1)
        dast_dbg (VHL_HEDLKF, "lkdfig");

      if (max_delta <= 0)
        break;

      nbr_errlft -= vhx_wrtpat (pt_paseq, pt_papat);

	/* ###------------------------------------------------------### */
	/*    Create a core file at the first error encountered. Create	*/
	/* also a save file. If no result pattern file is asked, save	*/
	/* also the current and the previous sets of petterns		*/
	/*								*/
	/*    Reset the core generation flag for further errors		*/
	/* ###------------------------------------------------------### */

      if ((core_flg != 0) && (max_err != nbr_errlft))
        {
        vhu_message (17, NULL, 0);

        if ((filepnt = mbkfopen (argv[core_flg], "cor", WRITE_TEXT)) == NULL)
          {
          sprintf (filename, "%s.cor", argv[core_flg]);
          asimut_err_flg = vhu_error (116, filename, NULL);
          exit (1);
          }
        vhx_core (filepnt, VHL_HEDLKF);
        fclose (filepnt);

        if ((filepnt = mbkfopen (argv[core_flg], "sav", WRITE_TEXT)) == NULL)
          {
          sprintf (filename, "%s.sav", argv[core_flg]);
          asimut_err_flg = vhu_error (116, filename, NULL);
          exit (1);
          }

        vhx_savall (filepnt, VHL_HEDLKF);
        fclose (filepnt);

        if (nores_flg == 1)
          {
          sav_end          = pt_paseq->ENDFLG;
          pt_paseq->ENDFLG = 'Y';

          pat_savpaseq (argv[core_flg], pt_paseq, labelsiz);
          pt_paseq->ENDFLG = sav_end;
          }

        core_flg = 0;
        }

	/* ###------------------------------------------------------### */
	/*    If a fatal error has been occured during the simulation	*/
	/* (Assert Violation, Bus conflict), set 'number of errors to	*/
	/* abort' (nbr_errlft) to zero to abort remaining patterns	*/
	/*								*/
	/*   Otherwies, begin proceesing of the next pattern		*/
	/* ###------------------------------------------------------### */

      if (vhx_chkerr () == 1)
        {
        nbr_errlft = 0;
        break;
        }

      pt_papat = pt_papat->NEXT;
      patindx ++;
      }

	/* ###------------------------------------------------------### */
	/*    After all the patterns of a set have been processed ...	*/
	/* ###------------------------------------------------------### */

    if (pdebug_flg == 1)
      dast_dbg (pt_paseq, "paseq");

    if (nores_flg == 0)
      pat_savpaseq (argv[frst_fil + 2], pt_paseq, labelsiz);

    if (max_delta <= 0)
      break;

    if ((pt_paseq->ENDFLG != 'Y') && (pt_paseq->ERRFLG == 0))
      {
      pt_paseq = pat_lodpaseq (argv[frst_fil + 1], pt_paseq, max_pat);
      if (pdebug_flg == 1)
        dast_dbg (pt_paseq, "paseq");
      }
    else
      break;

    }

  if (max_delta <= 0)
    asimut_err_flg = vhu_error (105, NULL, NULL);

  if (nbr_errlft < 0)
    asimut_err_flg = vhu_error (106, NULL, NULL);

	/* ###------------------------------------------------------### */
	/*    if the save flag is set create a save file		*/
	/* ###------------------------------------------------------### */

  if (nores_flg == 0)
    {
    if (pt_paseq->SAVFLG == 'Y')
      {
      if ((filepnt = mbkfopen (argv[frst_fil + 1], "sav", WRITE_TEXT)) == NULL)
        {
        sprintf (filename, "%s.sav", argv[frst_fil + 1]);
        asimut_err_flg = vhu_error (116, filename, NULL);
        exit (1);
        }
      vhx_savall (filepnt, VHL_HEDLKF);
      fclose (filepnt);
      }
    }

  if (nbr_errlft < max_err)
    exit (2);
  else
    exit (0);
  }
