/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* ###--------------------------------------------------------------### */
/* file		: vh_init.c						*/
/* date		: Dec 13 1999						*/
/* version	: v3.02							*/
/* author	: Pirouz BAZARGAN SABET					*/
/* description	: this file contains the followinf functions		*/
/*		  vhi_addcmodel, vhi_models				*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include BEH_H
#include "vh_util.h"
#include "vh_ltype.h"

/* ###--------------------------------------------------------------### */
/* function	: vhi_addcmodel						*/
/* description	: create a data structure that associate the c model's	*/
/*		  name and the pointer of the function that represents	*/
/*		  that C model						*/
/* called func.	: vhu_alloc, namealloc					*/
/* ###--------------------------------------------------------------### */

void vhi_addcmodel (str, fonc_pnt)

char *str;
void (*fonc_pnt) ();

  {
  struct lkdmdl        *model     ;
  extern struct lkdmdl *VHL_MODELS;

  model = (struct lkdmdl *) vhu_alloc (sizeof (struct lkdmdl), VHU_ALIGN);

  model->NAME  = namealloc (str);
  model->FONC  = fonc_pnt       ;
  model->NEXT  = VHL_MODELS     ;

  VHL_MODELS   = model;
  }

/* ###--------------------------------------------------------------### */
/* function	: vhi_models						*/
/* description	: declare available C models				*/
/* called func.	: vhi_addcmodel						*/
/* ###--------------------------------------------------------------### */

void vhi_models ()
  {
  extern void c_sr1k_8a      ();
  extern void c_sr1k_8b      ();
  extern void c_sr8k_8a      ();
  extern void c_fsyn_sr1k_1  ();
  extern void c_fsyn_sr1k_4  ();
  extern void c_fsyn_sr1k_10 ();
  extern void c_fsyn_sr1k_24 ();
  extern void c_fsyn_sr1k_56 ();
  extern void c_fsyn_sr4k_10 ();
  
  vhi_addcmodel ("c_sr1k_8a"     , c_sr1k_8a     );
  vhi_addcmodel ("c_sr1k_8b"     , c_sr1k_8b     );
  vhi_addcmodel ("c_sr8k_8a"     , c_sr8k_8a     );
  vhi_addcmodel ("c_fsyn_sr1k_1" , c_fsyn_sr1k_1 );
  vhi_addcmodel ("c_fsyn_sr1k_4" , c_fsyn_sr1k_4 );
  vhi_addcmodel ("c_fsyn_sr1k_10", c_fsyn_sr1k_10);
  vhi_addcmodel ("c_fsyn_sr1k_24", c_fsyn_sr1k_24);
  vhi_addcmodel ("c_fsyn_sr1k_56", c_fsyn_sr1k_56);
  vhi_addcmodel ("c_fsyn_sr4k_10", c_fsyn_sr4k_10);
  }
