/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* ###--------------------------------------------------------------### */
/* file		: vh_util.c						*/
/* date		: Dec 13 1999						*/
/* author	: Pirouz BAZARGAN SABET					*/
/* version	: v3.02							*/
/* description	: This file contains some utility functions :		*/
/*		  vhu_splitname, vhu_alloc  , vhu_avers, vhu_warning,	*/
/*		  vhu_toolbug  , vhu_message, vhu_error			*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include "vh_util.h"

/* ###--------------------------------------------------------------### */
/* function	: vhu_splitname						*/
/* description	: split a name onto an instance name and a signal name	*/
/* called func.	: namalloc						*/
/* ###--------------------------------------------------------------### */

void vhu_splitname (name, ins_name, sig_name)

char  *name    ;
char **ins_name;
char **sig_name;

  {
  unsigned int i   = 0  ;
  unsigned int j   = 0  ;
  char         buf [256];

  for (i=0 ; name [i] != '\0' ; i++)
    {
    if (name [i] == '.')
      j = i + 1;
    }

  if ((j != 0) && (j != i))
    {
    strcpy (buf, name);
    buf [j-1] = '\0';
    *ins_name = namealloc (buf    );
    *sig_name = namealloc (buf + j);
    }
  else
    {
    *ins_name = NULL;
    *sig_name = name;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhu_alloc						*/
/* description	: make memory allocation				*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

char *vhu_alloc (size, mode)

unsigned int size;
char         mode;

  {
  static char         *mem_pt  = NULL;
  static unsigned int  mem_idx = 0   ;
  char                *res_pt  = NULL;

	/* ###------------------------------------------------------### */
	/*    align the size on a multiple of 4				*/
	/* ###------------------------------------------------------### */

  if (size >= VHU_MEMALC)
    res_pt = (char *) mbkalloc (size);
  else
    {
    if (mode == VHU_ALIGN)
      {
      mem_idx = (mem_idx + 3) & 0xfffffffc;
      size    = (size    + 3) & 0xfffffffc;
      }

    if ((mem_pt == NULL) || ((mem_idx + size) > VHU_MEMALC))
      {
      mem_pt  = (char *) mbkalloc (VHU_MEMALC);
      mem_idx = 0;
      }

    res_pt   = mem_pt + mem_idx;
    mem_idx += size   ;
    }

  return (res_pt);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhu_avers						*/
/* description	: return the current version				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char *vhu_avers ()
  {
  return ("v3.02");
  }

/* ###--------------------------------------------------------------### */
/* function	: vhu_warning						*/
/* description	: print a warning message corresponding to the argument	*/
/*		  code							*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhu_warning (code,str1)

int   code;
char *str1;

  {
  (void) fprintf (stderr,"Warning %d : ",code);
  switch (code)
    {
    case 1:
      (void) fprintf (stderr,"cannot find register `%s`\n",str1);
      break;
    case 2:
      (void) fprintf (stderr,"consistency checks will be disabled\n");
      break;
    case 3:
      (void) fprintf (stderr,"signal `%s` never used\n",str1);
      break;
    case 4:
      (void) fprintf (stderr,"cannot restore the mode of port `%s`\n",str1);
      break;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhu_toolbug						*/
/* description	: print a fatal error message corresponding to the	*/
/*		  argument code and exit				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhu_toolbug (code,str1,str2,nbr1)

int   code;
char *str1;
char *str2;
int   nbr1;

  {
  (void) fprintf (stderr,"Fatal error %d executing `%s`: ", code,str1);
  switch (code)
    {
    case 1:
      (void) fprintf (stderr,"unknown operator\n");
      break;
    case 2:
      (void) fprintf (stderr,"cannot create empty atom\n");
      break;
    case 7:
      (void) fprintf (stderr,"unknown type `%c` for IO `%s`\n",nbr1,str2);
      break;
    case 8:
      (void) fprintf (stderr,"illegal signal value : hex `%x`\n",nbr1);
      break;
    case 13 :
      (void) fprintf (stderr,"cannot find model of `%s`\n",str2);
      break;
    case 18:
      (void) fprintf (stderr,"unknown mode `%c` for port `%s`\n",nbr1,str2);
      break;
    case 19:
      (void) fprintf (stderr,"unknown type `%c` for input `%s`\n",nbr1,str2);
      break;
    case 20:
      (void) fprintf (stderr,"`%s` no such locon\n", str2);
      break;
    case 21:
      (void) fprintf (stderr,"illegal direction `%c` for `%s`\n", nbr1, str2);
      break;
    case 22:
      (void) fprintf (stderr,"`%s`: `%c` unknown resolution func.\n",str2,nbr1);
      break;
    default:
      (void) fprintf (stderr, "BUG\n");
    }
  exit (1);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhu_message						*/
/* description	: print a message corresponding to the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhu_message (code, str1, nmb1)

int   code;
char *str1;
int   nmb1;

  {
  switch (code)
    {
    case 0:
      (void) fprintf (stdout,"\nasimut %s\n", vhu_avers());
      break;
    case 1:
      (void) fprintf (stdout,"Usage : asimut [options] [files]\n");
      (void) fprintf (stdout,"      : asimut -h\n");
      break;
    case 2:
      (void) fprintf (stdout,"Initializing ...\n");
      break;
    case 3:
      (void) fprintf (stdout,"Compiling `%s` (Behaviour) ...\n",str1);
      break;
    case 4:
      (void) fprintf (stdout,"Compiling `%s` (Structural) ...\n\n",str1);
      break;
    case 5:
      (void) fprintf (stdout,"Searching `%s` ...\n",str1);
      break;
    case 6:
      (void) fprintf (stdout,"Searching pattern file : `%s` ...\n",str1);
      break;
    case 8:
      (void) fprintf (stdout,"Compiling `%s` (Pattern) ...\n\n",str1);
      break;
    case 9:
      (void) fprintf (stdout,"Linking ...\n");
      break;
    case 10:
      (void) fprintf (stdout,"###----- processing pattern %d -----###\n",nmb1);
      break;
    case 12:
      (void) fprintf (stdout,"Generating the file '%s' ...\n",str1);
      break;
    case 13:
      (void) fprintf (stdout,"of the figure '%s'.\n",str1);
      break;
    case 14:
      (void) fprintf (stdout,"File '%s' has been generated.\n",str1);
      break;
    case 15:
      (void) fprintf (stdout,"Making %s ...\n\n", str1);
      break;
    case 16:
      (void) fprintf (stdout,"Restoring ...\n\n");
      break;
    case 17:
      (void) fprintf (stdout,"Creating the core file ...\n\n");
      break;
    case 18:
      (void) fprintf (stdout, "Flattening the root figure ...\n\n");
      break;
    case 19:
      (void) fprintf (stdout, "stop mark set on the current structure\n");
      break;
    case 20:
      (void) fprintf (stdout, "--- STACK OVERFLOW !! ---\n");
      break;
    case 21:
      (void) fprintf (stdout, "\tParis, France, Europe, Earth, Solar system, Milky Way, ...\n");
      break;
    default:
      (void) fprintf (stderr,"vhu_message : code %d unknown.\n",code);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhu_error						*/
/* description	: print an error message depending on the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int vhu_error (code, str1, str2)

int   code;
char *str1;
char *str2;

  {
  if (code < 200)
    (void)fprintf (stderr,"Error %d :",code);

  switch (code)
    {
    case 100:
      (void) fprintf (stderr,"cannot find `%s`\n",str1);
      break;
    case 102:
      (void) fprintf (stderr,"instance and model of `%s` mismatch\n",str1);
      break;
    case 103:
      (void) fprintf (stderr,"incompatible input/output mode for `%s`\n",str1);
      break;
    case 104:
      (void) fprintf (stderr,"input/output `%s` doesn't exist\n",str1);
      break;
    case 105:
      (void) fprintf (stderr,"Possible combinatory loop. Simulation aborted\n");
      break;
    case 106:
      (void) fprintf (stderr,"Too many errors. Simulation interrupted\n");
      break;
    case 107:
      (void) fprintf (stderr,"Cannot open result file\n");
      break;
    case 109:
      (void) fprintf (stderr,"signal `%s` never assigned\n",str1);
      break;
    case 110:
      (void) fprintf (stderr,"signal `%s` not connected to any port\n",str1);
      break;
    case 111:
      (void) fprintf (stderr,"illegal connections on signal `%s`\n",str1);
      break;
    case 112:
      (void) fprintf (stderr,"drivers conflict on `%s`\n", str1);
      break;
    case 113:
      (void) fprintf (stderr,"expected value and simulation's result mismatch on `%s`\n",str1);
      break;
    case 114:
      (void) fprintf (stderr,"multiple driver on unguarded signal `%s`\n",str1);
      break;
    case 115:
      (void) fprintf (stderr,"save file doesn't match with the description\n");
      break;
    case 116:
      (void) fprintf (stderr,"cannot open file `%s`\n", str1);
      break;
    case 117:
      (void) fprintf (stderr,"instance `%s` doesn't exist or cannot be inspected\n",str1);
      break;
    case 118:
      (void) fprintf (stderr,"command not understood\n");
      break;
    case 119:
      (void) fprintf (stderr,"cannot display VOID pointer\n");
      break;
    case 120:
      (void) fprintf (stderr,"cannot push NULL pointer\n");
      break;
    case 121:
      (void) fprintf (stderr,"driver conflict on `%s` in `%s`\n", str2, str1);
      break;

    case 200:
      (void) fprintf (stderr,"\n	cannot continue further more.\n");
      (void) fprintf (stderr,"\n		Have a nice day...\n");
      break;

    default:
      (void) fprintf (stderr,"unknown error code\n");
      break;
    }

  return (1);
  }
