/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :    placer_dprfile.c

    Date        :    30 09 1994
    Version     :    1.0

    Auteur      :    CHAPUT  jean-paul
                     BEAURIN oliver

    Description :    Read the dpr file for placement step

\*********************************************************************/

#include  <stdio.h>
#include  <string.h>

#include   MUT_H
#include   MLO_H
#include   MPH_H

#include   GFP_H
#include  "placer_dprfile.h"


/* ********************* Internal constants ********************* */

#define        IW_PlCon_SetPhConPower             2
#define        IW_PlCon_SetPhConCTRL              3
#define        IW_PlCon_SetPhConInvalidSlice      5
#define        IW_PlCon_SetPhConInvalidTrack      6

#define        IC_PlCon_ConNameLength          1024
#define        IC_UtFDP_SliceHeight              60  /*         lambdas */
#define        IC_PlCon_TrackNumber              10
#define        IC_PlCon_PowBarWideVdd             7  /* 6/2-1+5 lambdas */
#define        IC_PlCon_PowBarWideVss             8  /* 8/2-1+5 lambdas */
#define        IC_PlCon_TrackPitch                5  /*         lambdas */
#define        IC_PlCon_PowerRefreshWide         10  /*         lambdas */
#define        IC_PlCon_ConTypeData               2
#define        IC_PlCon_ConTypeCTRL               3


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "placer_dprfile" module identifier.        */
	char  SetPhCon_ConName[IC_PlCon_ConNameLength];
							 /* Name of the connector currently
							  * processed by 'PlCon_SetPhCon'.
							  */
	long  SetPhCon_SliceNB; /* Number of slice of the physical figure. */
	} GlobVar;
	

/* ********************* Internal Variables ********************* */


/* ******************** Interface Functions ********************* */


/* extern void  PlCon_Init(void);
 * extern void  PlCon_SetPhCon(      phfig_list *aPhFig,
 *                                    locon_list *aLoCon);
 * extern void  PlCon_SetPhPowFig(   phfig_list *aPhFig,
 *                                    lofig_list *aLoFig);
 * extern void  PlCon_SetPhPowIns(   phfig_list *aPhFig,
 *                                    loins_list *aLoIns);
 * extern void  PlCon_MakePlacement(       char *aFigName);
 */


/* ********************* Internal Functions ********************* */


/*  Function  :  'PlCon_Init'.                                   */
/* -------------------------------------------------------------- */
/*  Initialisation of the module "placer_dprfile". Get the module
 *  identifier, load errors and warnings.
 */

extern void  PlCon_Init()
{
	static long  FirstCall =  TRUE;


	if (FirstCall == TRUE)
	{
		/* Attribution of a module identifier. */
		GlobVar.ModuleId = GfpUT_GetModuleId("placer_dprfile", F_GfpUT_ModuleAdd);

		/* Warnings belonging to the physical placement functions.
		 */
		GfpUT_AddWarng(GlobVar.ModuleId, IW_PlCon_SetPhConPower,
		"[%s]:For terminal \"%s\", power supplies terminals cannot be moved.");
		GfpUT_AddWarng(GlobVar.ModuleId, IW_PlCon_SetPhConCTRL,
		"[%s]:For terminal \"%s\", control terminals cannot be moved.");
		GfpUT_AddWarng(GlobVar.ModuleId, IW_PlCon_SetPhConInvalidSlice,
		"[%s]:For terminal \"%s\", invalid slice %ld (0<=\'slice\'<%ld).");
		GfpUT_AddWarng(GlobVar.ModuleId, IW_PlCon_SetPhConInvalidTrack,
		"[%s]:For terminal \"%s\", invalid track %ld (0<=\'track\'<%ld).");


		FirstCall = FALSE;
	}
}


/*  Function  :  'PlCon_SetPhPowFig'                             */
/* -------------------------------------------------------------- */

extern void  PlCon_SetPhPowFig(aPhFig, aLoFig)
	phfig_list *aPhFig;
	lofig_list *aLoFig;
{
	ulofig_power       *PtPower;
		   phfig_list  *PtFigIns;
		   phins_list  *PtIns;
		   phcon_list  *PtCon;
		   phcon_list **PPtCon;
		   ptype_list  *PtUser;
				 long   XStep;
				 long   XPower;
				 long   XWidth;
				 long   Flag;


	/* Delete all the NORTH and SOUTH power connnectors. */
	for( PPtCon  =  &(aPhFig->PHCON);
		*PPtCon != (phcon_list*)NULL;)
	{
		if (   (   isvdd((*PPtCon)->NAME)
				|| isvss((*PPtCon)->NAME))
			&& (   ((*PPtCon)->ORIENT == NORTH)
				|| ((*PPtCon)->ORIENT == SOUTH)) )
		{ delphcon(aPhFig, *PPtCon); continue; }

		PPtCon  = &((*PPtCon)->NEXT);
	}

	/* Look for the figure power refreshment placement informations. */
	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigPower,
											F_GfpUT_UserGet       );

	/* No power refreshment informations, nothing to do. */
	if (   (PtUser       == (void*)NULL)
		|| (PtUser->DATA == (void*)NULL)) return;

	PtPower = (ulofig_power*)(PtUser->DATA);

	XStep = (aPhFig->XAB2 - aPhFig->XAB1) / (PtPower->NUMB);

	for(XPower  = aPhFig->XAB1 + XStep/2;
		XPower  < aPhFig->XAB2;
		XPower += XStep)
		for(PtIns = aPhFig->PHINS;
			PtIns != (phins_list*)NULL;
			PtIns  =       PtIns->NEXT)
		{
			PtFigIns = getphfig(PtIns->FIGNAME, 'A');
			  XWidth = PtFigIns->XAB2 - PtFigIns->XAB1;

			if (   (XPower >= PtIns->XINS)
				&& (XPower <  PtIns->XINS + XWidth))
			{
				if (XPower < PtIns->XINS+XWidth/2)
				{
					Flag = FALSE;
					for(PtCon  =     aPhFig->PHCON;
						PtCon != (phcon_list*)NULL;
						PtCon  =       PtCon->NEXT)
						if (   (   isvdd(PtCon->NAME)
								|| isvss(PtCon->NAME))
							 &&(PtIns->XINS == PtCon->XCON))
							Flag = TRUE;

					if (Flag == TRUE) break;

					addphcon(aPhFig, NORTH, VDD, PtIns->XINS,
												aPhFig->YAB2,
								  						ALU1,
									   PtPower->WIDE*SCALE_X);
					addphcon(aPhFig, NORTH, VSS, PtIns->XINS,
												aPhFig->YAB2,
								  						ALU1,
									   PtPower->WIDE*SCALE_X);
				}
				else
				{	
					Flag = FALSE;
					for(PtCon  =     aPhFig->PHCON;
						PtCon != (phcon_list*)NULL;
						PtCon  =       PtCon->NEXT)
						if (   (   isvdd(PtCon->NAME)
								|| isvss(PtCon->NAME))
							 &&(PtIns->XINS+XWidth == PtCon->XCON))
							Flag = TRUE;

					if (Flag == TRUE) break;

					addphcon(aPhFig, NORTH, VDD, PtIns->XINS+XWidth,
													   aPhFig->YAB2,
															   ALU1,
											  PtPower->WIDE*SCALE_X);
					addphcon(aPhFig, NORTH, VSS, PtIns->XINS+XWidth,
													   aPhFig->YAB2,
															   ALU1,
											  PtPower->WIDE*SCALE_X);
				}

				break;
			}
		}
}


/*  Function  :  'PlCon_SetPhPowIns'                             */
/* -------------------------------------------------------------- */
/*  Set the power supplies refreshment associated to the instance
 *  "aLoIns" in the figure "aPhFig".
 */

extern void  PlCon_SetPhPowIns(aPhFig, aLoIns)
	phfig_list *aPhFig;  /* Physical figure.              */
	loins_list *aLoIns;  /* Associated logical connector. */
{
	uloins_power      *PtPower;
		   phfig_list *PtPhFig;
		   phins_list *PtPhIns;
		   phcon_list *PtCon;
		   ptype_list *PtUser;
				 long  XPower;
				 long  Flag;


	/* Look for the instance power refreshment placement informations. */
	PtUser = GfpUT_GetUser(&(aLoIns->USER), C_GfpUT_UserLoInsPower,
											F_GfpUT_UserGet       );

	/* No power refreshment informations, nothing to do. */
	if (   (PtUser       == (void*)NULL)
		|| (PtUser->DATA == (void*)NULL)) return;

	PtPower = (uloins_power*)(PtUser->DATA);
	PtPhIns = getphins(aPhFig, aLoIns->INSNAME);
	PtPhFig = getphfig(        aLoIns->FIGNAME, 'A');

	if (PtPower->SIDE & C_UtFDP_SideTypeWest)
	{
		XPower = PtPhIns->XINS;

		Flag = FALSE;
		for(PtCon  =     aPhFig->PHCON;
			PtCon != (phcon_list*)NULL;
			PtCon  =       PtCon->NEXT)
			if (   (   isvdd(PtCon->NAME)
					|| isvss(PtCon->NAME))
				 &&(XPower == PtCon->XCON) )
				{ Flag = TRUE; break; }

		if (Flag == FALSE)
		{
			addphcon(aPhFig, NORTH, VDD, XPower, aPhFig->YAB2, ALU1,
											  PtPower->WIDE*SCALE_X);
			addphcon(aPhFig, NORTH, VSS, XPower, aPhFig->YAB2, ALU1,
											  PtPower->WIDE*SCALE_X);
		}
	}

	if (PtPower->SIDE & C_UtFDP_SideTypeEast)
	{
		XPower = PtPhIns->XINS + (PtPhFig->XAB2 - PtPhFig->XAB1);

		Flag = FALSE;
		for(PtCon  =     aPhFig->PHCON;
			PtCon != (phcon_list*)NULL;
			PtCon  =       PtCon->NEXT)
			if (   (   isvdd(PtCon->NAME)
					|| isvss(PtCon->NAME))
				 &&(XPower == PtCon->XCON) )
				{ Flag = TRUE; break; }

		if (Flag == FALSE)
		{
			addphcon(aPhFig, NORTH, VDD, XPower, aPhFig->YAB2, ALU1,
											  PtPower->WIDE*SCALE_X);
			addphcon(aPhFig, NORTH, VSS, XPower, aPhFig->YAB2, ALU1,
											  PtPower->WIDE*SCALE_X);
		}
	}
}


/*  Function  :  'PlCon_MakePlacement'.                          */
/* -------------------------------------------------------------- */
/*  This function use the physical informations given whith the
 *  netlist to modify the physical figure.
 *      It place the terminals, resize the abutment-box and set
 *  the vertical power refreshment.
 */

extern void  PlCon_MakePlacement(aFigName)
	char *aFigName;  /* Name of the physical and logical figure. */
{
	lofig_list *PtLoFig;
	phfig_list *PtPhFig;
	loins_list *PtLoIns;
	/* Get logical and physical figures. */
	PtLoFig = getlofig(aFigName, 'A');
	PtPhFig = getphfig(aFigName, 'A');

	/* Set power supplies refreshment. The informations about power
	 * supplies refreshment are stored in the 'lofig' user field for
	 * the global frequency.
	 *     An information is also present in the user 'loins' field
	 * for specific power refreshment.
	 */
	PlCon_SetPhPowFig(PtPhFig, PtLoFig);
	UtFDP_DelLoFigPower(       PtLoFig);
	for(PtLoIns  = PtLoFig->LOINS;
		PtLoIns != (loins_list*)NULL;
		PtLoIns  =     PtLoIns->NEXT)
	{
		PlCon_SetPhPowIns(PtPhFig, PtLoIns);
		UtFDP_DelLoInsPower(       PtLoIns);
	}

}
