/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/********************************************************************
*                                                                   *
* Laboratoire MASI CAO-VLSI, UPMC, Copyright 1991 1992 1993 1994    *
*                                                                   *
* Software support Email : cao-vlsi@masi.ibp.fr                     *
*                                                                   *
* Authors : Zouina AKTOUF &                                         *
*           El Arabi RHOMARI &                                      *
*           Jose MARTINS DOS SANTOS                                 *
*                                                                   *
* Supervision & Modifications : Lotfi BEN AMMAR                     *
*                                                                   *
* Restructuration : Olivier BEAURIN                                 *
*                                                                   *
********************************************************************/

#ifndef DPR_H
#define DPR_H

/*---------------------------------------------------------\
                     structures
\---------------------------------------------------------*/

typedef struct AFIGURE 
        {
        struct chain  *LIN;		/* liste des lignes	        	*/
        struct chain  *COL;		/* liste des colonnes	    	*/
        struct chain  *SIG;		/* liste des signaux	    	*/
        struct chain  *CON;		/* liste des connecteurs	    */
        struct chain  *ALLOW;	/* liste des transparences    	*/
        struct chain  *ALIM;	/* liste des ports d'alim	    */
        struct chain  *GRID;	/* noeud bas gauche de la figure*/
        }
        FIGURE;	  /* mother figure */

typedef struct ACOLUMN 
	    {
        char		*NAME;		/* nom indexe de la colonne 	*/
        char		 TYPE;	        /* Left,Right,Channel,Operator 	*/
        long		 X,Y;          	/* coordonnees			*/
        long		 WIDTH;        	/* Largeur 			*/
        long		 S_WIDTH;      	/* Largeur Symbolique		*/
        long     	 HEIGHT;       	/* Hauteur 			*/
        long     	 S_HEIGHT;     	/* Hauteur Symbolique		*/
	    int		 INDEX;		/* position  de la colonne 	*/
        struct chain 	*SIG;       	/* Liste des signaux 		*/
        struct chain 	*DF_CON;      	/* Liste des connecteurs (ph)	*/
        struct chain 	*CT_CON;      	/* Liste des connecteurs (ph)	*/
        struct chain 	*ALIMS;      	/* Liste des connecteurs (ph)	*/
        struct chain 	*INS;        	/* Liste des instances 		*/
        struct chain 	*VIA;        	/* Liste des vias ou contacts 	*/
        struct chain 	*SEG;        	/* Liste des segments 		*/
        struct chain 	*ALLOW;       	/* Liste des Transparences 	*/
        struct chain 	*RDATA;        	/* Zone de Travail 		*/     
        struct chain 	*ALLOWUSED;    	/* Zone de Manoeuvre 		*/
        struct chain    *USER;
        }
        COLUMN;	  /* Modele de colonne */

typedef struct ANWELLREF
      	{
        long 		 X,Y;		/* coordonnees 		 	*/
        long 		 WIDTH;		/* largeur 		 	*/
    	struct chain	*INS;		/* Pointeur sur la cellule      */
    	struct chain    *COL;       	/* Pointeur sur la colonne      */
        }
    	NWELLREF;/* Modele de reference nwell */

typedef struct ACONNECTOR 
    	{
        char		*NAME;         	/* Nom 				*/
        char		 TYPE;       	/* (I)nterne ou (E)xterne 	*/
        long 		 NBPOSS;	/* Nombre de possibilites 	*/
        long 		 MINPOS;	/* Nombre de possibilites 	*/
        long 		 MAXPOS;	/* Nombre de possibilites 	*/
        struct chain    *POS;        	/* liste des possibilites 	*/ 
    	struct chain    *SIG;       	/* Pointeur sur le signal       */ 
    	struct chain	*INS;		/* Pointeur sur la cellule      */
    	struct chain    *COL;       	/* Pointeur sur la colonne      */
    	struct chain    *LIN;       	/* Pointeur sur la ligne        */
        struct chain	*USER;
        }
        CONNECTOR;  /* Modele du Connecteur */

typedef struct APOSSIBILITY
      	{
        long 		 X,Y;		/* coordonnees 		 	*/
        long 		 WIDTH;		/* largeur 		 	*/
        long 		 TRACK;		/* numero de piste 	 	*/
        char		 TYPE;       	/* fixe ou acces-multiple 	*/
    	char             ACCES;		/* segment name             	*/
    	char             LAYER;		/* code du layer              	*/
        CONNECTOR       *CON;           /* Connecteur                   */
        struct chain	*USER;
        }
    	POSSIBILITY;/* Modele de possibilite */

typedef struct ATRACK
        {
        struct ATRACK *NEXT;
        long           Y;
        }
        TRACK;

typedef struct AGABARIT
        {
        char          *ALIM1;
        long           WALIM1,WALIM2;
        long           HEIGHT;
        long           NBTRACK;
        struct ATRACK *TRACK;
        }
        GABARIT;

typedef struct  AACCESS
        {
        struct AACCESS       *NEXT;
        struct AACCESS       *PREV;
        struct APOSSIBILITY  *POS;
        char                  ORIENT;
        struct ASIGNAL       *SIG;
        char                  DONE;
        }
        ACCESS;

typedef struct  ANODE
        {
        struct ANODE   *NEXT;
        struct ANODE   *PREV;
        struct ANODE   *DPR_UP;
        struct ANODE   *DPR_DWN;
        struct AACCESS *DPR_RIGHT;
        struct AACCESS *DPR_LEFT;
        }
        NODE;

typedef struct ALINE 
    	{
        long		 X,Y;          	/* coordonnees			*/
        long		 WIDTH;        	/* Largeur 			*/
        long     	 HEIGHT;       	/* Hauteur 			*/
    	int		 INDEX;		/* position de la ligne (B-S)	*/
        GABARIT	        *GAB;           /* Gabarit de la ligne	        */
        NODE            *GRID;          /* noeud bas gauche de la ligne */
        struct chain 	*SIG;       	/* Liste des signaux 		*/
        struct chain 	*CON;      	/* Liste des connecteurs (Ph) 	*/
        struct chain 	*INS;        	/* Liste des instances 		*/
        struct chain 	*N_WELL;       	/* Liste des Bornes NWell 	*/
        struct chain 	*ALLOW;       	/* Liste des Transparences 	*/
        }
        LINE;	/* Modele de colonne */

typedef struct ASIGNAL 
    	{
        char		*NAME;		/* nom du signal 		*/
        long		 INDEX;		/* identificateur du signal MBK	*/
        char		 TYPE;       	/* (I)nterne ou (E)xterne 	*/
        long             MIN_LIN;       /* index de la ligne Dn         */
        long             HEIGHT;        /* hauteur du signal            */
        long             MIN_COL;       /* index de la colonne left     */
        long             WIDTH;         /* largeur du signal            */
        struct chain	*CON;		/* Liste des connecteurs	*/
        struct chain	*COL;		/* Liste des colonnes 		*/
        struct chain	*LIN;		/* Liste des lignes 		*/
        struct chain	*BCROSS;	/* Zone de Manoeuvre 		*/
        struct chain	*YPRIOR;	/* Zone de Manoeuvre 		*/
        struct chain	*USER;
        }
        SIGNAL;	/* Modele de Signal */  

typedef struct AINSTANCE 
    	{
        char		*INSNAME;	/* instance name     		*/
        char		*FIGNAME;	/* figure name       		*/
        long		 X,Y;		/* coordonnees			*/
        long		 WIDTH;		/* Largeur 			*/
        long		 HEIGHT;       	/* Hauteur 			*/
    	char		 TRANSF;	/* transformation		*/
        struct chain 	*DF_CON;	/* Liste des connecteurs (Ph) 	*/
        struct chain 	*CT_CON;	/* Liste des connecteurs (Ph) 	*/
        struct chain 	*ALIMS; 	/* Liste des connecteurs (Ph) 	*/
        struct chain 	*RAPPELS; 	/* Liste des connecteurs (Ph) 	*/
        struct chain 	*ALLOW; 	/* Liste des Transparences 	*/
        struct chain 	*N_WELL;       	/* Liste des Bornes NWell 	*/
        struct chain 	*USER;
        }
        INSTANCE;	/* Modele d'une instance  */

typedef struct ASEGMENT 
    	{
    	char            *NAME;		/* segment name             	*/
    	long             X1,Y1,X2,Y2;	/* segment coordinates      	*/
    	long             WIDTH;	        /* largeur du segment        	*/
    	char             TYPE;		/* possibles values: V H    	*/
    	char             LAYER;		/* physical layer           	*/
        }
    	SEGMENT;	/* modele de segment  */

typedef struct AINLINE
    	{
        int		Y;		/* Abscisse 			*/
        struct chain	*CON;		/* Liste des ports a connecter  */
        }
        INLINE;		/* Modele de liste de ports alignes */  

typedef struct AVIA 
    	{
        int		 TYPE;		/* contact ou via		*/
    	long		 X,Y;		/* coordonnees			*/
        }
        VIA;		/* Modele de via */      

typedef struct SCR_ConnectorList
    	{
    	struct SCR_ConnectorList       *NextCon; /* Next symbolic terminal */
    	long				ConName; /* Signal index           */
    	long				Mark   ; /* Track indice           */
        struct ACONNECTOR  *CON;
    	}
        ConnectorList;

typedef struct SCR_SegmentList
    	{
    	struct SCR_SegmentList	       *NextSeg;
        char                           *SegName;
    	long				X1Seg   ;
    	long				Y1Seg   ;
    	long				X2Seg   ;
    	long				Y2Seg   ;
    	}
        SegmentList;

typedef struct SCR_ViasList
    	{
    	struct SCR_ViasList	       *NextVia;
    	long				XVia    ;
    	long				YVia    ;
    	}
        ViasList;

typedef struct ACHANNEL
        {
        ConnectorList  *NORTHCON;
        ConnectorList  *SOUTHCON;
        ConnectorList  *WESTCON;
        ConnectorList  *EASTCON;
        }
        CHANNEL ;

typedef struct ANUMTRACK
        {
        struct ANUMTRACK *NEXT ; /* Next numtrack          */
        long              NUM  ; /* Symbolic track index   */
        long              TRACK; /* Horizontal track Y     */
        } 
        NUMTRACK ;


/*---------------------------------------------------------\
                     constantes
\---------------------------------------------------------*/
#define SETACCESS       1803
#define SESAME          1903
#define OPTIPLACE       2003

#define MAXLONG  999999999
#define MINLONG -999999999
#define MAXINT  99999
#define MININT -99999
#define TYPEFIX 'F'
#define MULTIAX 'M'
#define ORIENTX 'X'
#define FTRACK  -1
#define MAXTRACK 10
#define EAST_WEST 'H'
#define TRUE     1
#define FALSE    0

#define WLAYER1            1 
#define WLAYER2            2 
#define WTRAN              2 
#define NWELL_NWELL       12 
#define MIN_NWELL          4 
#define ALU1_ALU1          4 
#define NOTCHE_LAYER1      1 
#define NOTCHE_LAYER3      2 
#define WIDTH_INTER        4
#define VIA_VIA            4 
#define CANAL_BORDER_RIGHT 2 
#define CANAL_BORDER_LEFT  2 
#define CANAL_INTER_TRACK  5 

#define Upcase(c) (((c>='a')&&(c<='z'))? c+'A'-'a':c)
#define dpr_min(x,y) ((x<y)?x:y)
#define dpr_max(x,y) ((x>y)?x:y)

#endif

