/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_strech.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_supply.h"
#include  "strecher_phcut.h"
#include  "strecher_strech.h"

/*---------------------------------------------------------\
                     macros
\---------------------------------------------------------*/
#define  streched_y(y)  ( ((y)>=Ycoupe) ? ((y)+SX_YADD)  \
                                        :  (y)          )

/*---------------------------------------------------------\
                     locals
\---------------------------------------------------------*/
static phfig_list *ptfig_WORK = (phfig_list *) NULL;
static ptype_list *PPHSEGPOL = (ptype_list *) NULL;
static supply_list *psupply;
static long     Ycoupe, NB_pitch, SX_YADD;

/*---------------------------------------------------------\
                     strech_con
\---------------------------------------------------------*/
static void     strech_con()
{
    phcon_list     *pcon;

    for (pcon = ptfig_WORK->PHCON; pcon != (phcon_list *) NULL;
         pcon = pcon->NEXT)
    {
        pcon->YCON = streched_y(pcon->YCON);
    }
}

/*---------------------------------------------------------\
                     strech_via
\---------------------------------------------------------*/
static void     strech_via()
{
    phvia_list     *pvia;


    for (pvia = ptfig_WORK->PHVIA; pvia != (phvia_list *) NULL;
         pvia = pvia->NEXT)
    {
        pvia->YVIA = streched_y(pvia->YVIA);
    }

}

/*---------------------------------------------------------\
                     rename_refcon
\---------------------------------------------------------*/
static char    *rename_refcon(pref)
phref_list     *pref;
{
    char           *pname, new_name[MBK_SZ_NAME];
    long            index;


    if (pref->FIGNAME != namealloc("ref_con"))
        return (pref->NAME);

    pname = getrefconname(pref->NAME, &index);

    if ((psupply->TYPE == C_SUPPLY_VDD)
        && (pref->YREF < Ycoupe))
        return (pref->NAME);
    if ((psupply->TYPE == C_SUPPLY_VSS)
        && (pref->YREF >= Ycoupe))
        return (pref->NAME);

    /* Nouvel index de la reference. */
    sprintf(new_name, "%s_%ld", pname, index + NB_pitch);

    return (namealloc(new_name));
} /* end of rename_refcon */

/*---------------------------------------------------------\
                     strech_ref
\---------------------------------------------------------*/
static void     strech_ref()
{
    phref_list     *pref;


    for (pref = ptfig_WORK->PHREF; pref != (phref_list *) NULL;
         pref = pref->NEXT)
    {
        pref->YREF = streched_y(pref->YREF);
        pref->NAME = rename_refcon(pref);
    }

}

/*---------------------------------------------------------\
                     strech_ins
\---------------------------------------------------------*/
static void     strech_ins()
{
    phins_list     *pins;


    for (pins = ptfig_WORK->PHINS; pins != (phins_list *) NULL;
         pins = pins->NEXT)
    {
        pins->YINS = streched_y(pins->YINS);
    }

}

/*---------------------------------------------------------\
                     strech_nwell
\---------------------------------------------------------*/
static void     strech_nwell()
{
#define  access_nwell  ( (phseg_list*)(pnwell->DATA) )

    ptype_list     *pnwell;
    phseg_list     *newpnwell;


    if ((SX_YADD / SCALE_X) % 2)
        for (pnwell = PPHSEGPOL; pnwell != (ptype_list *) NULL;
             pnwell = pnwell->NEXT)
        {
            /* Segment inferieur. */
            newpnwell = addphseg(ptfig_WORK,
                                 access_nwell->LAYER,
                                 access_nwell->WIDTH + lambda_floor(SX_YADD),
                                 access_nwell->X1,
                                 access_nwell->Y1 + lambda_floor(SX_YADD / 2),
                                 access_nwell->X2,
                                 access_nwell->Y2 + lambda_floor(SX_YADD / 2),
                                 access_nwell->NAME);
            newpnwell->USER = access_nwell->USER;

            /* Segment superieur. */
            newpnwell = addphseg(ptfig_WORK,
                                 access_nwell->LAYER,
                                 access_nwell->WIDTH + lambda_floor(SX_YADD),
                                 access_nwell->X1,
                                 access_nwell->Y1 + lambda_ceil(SX_YADD / 2),
                                 access_nwell->X2,
                                 access_nwell->Y2 + lambda_ceil(SX_YADD / 2),
                                 access_nwell->NAME);
            newpnwell->USER = access_nwell->USER;

            delphseg(ptfig_WORK, access_nwell);
        }
    else
        for (pnwell = PPHSEGPOL; pnwell != (ptype_list *) NULL;
             pnwell = pnwell->NEXT)
        {
            access_nwell->WIDTH = access_nwell->WIDTH + SX_YADD;
            access_nwell->Y1 = access_nwell->Y1 + SX_YADD / 2;
            access_nwell->Y2 = access_nwell->Y2 + SX_YADD / 2;
        }

#undef  access_nwell
}

/*---------------------------------------------------------\
                     strech_seg
\---------------------------------------------------------*/
static void     strech_seg( must_cross )
int must_cross;
{
    phseg_list     *pseg;
    long            lerror;

    lerror = FALSE;

    for (pseg = ptfig_WORK->PHSEG; pseg != (phseg_list *) NULL;
         pseg = pseg->NEXT)
    {
        if ((pseg->LAYER == PTIE) || (pseg->LAYER == NTIE) || (pseg->LAYER == NWELL))
        {
            if (mc_IsHor(pseg->TYPE)
                && (pseg->Y1 - pseg->WIDTH / 2 < Ycoupe)
                && (pseg->Y1 + pseg->WIDTH / 2 >= Ycoupe))
            {
                PPHSEGPOL = addptype(PPHSEGPOL, PTYPE_XSEGPOL, (void *) pseg);
                continue;
            }
        }

        switch (pseg->LAYER)
        {
        case PTRANS:
        case NTRANS:
            if (mc_IsHor(pseg->TYPE)
                && (pseg->Y1 - SX_XTRANS < Ycoupe)
                && (pseg->Y1 + SX_XTRANS >= Ycoupe))
                lerror = TRUE;
            if (mc_IsVer(pseg->TYPE)
                && (pseg->Y1 < Ycoupe)
                && (pseg->Y2 >= Ycoupe))
                lerror = TRUE;
            break;
        case PDIF:
        case NDIF:
            if (mc_IsHor(pseg->TYPE)
                && (pseg->Y1 - pseg->WIDTH / 2 < Ycoupe)
                && (pseg->Y1 + pseg->WIDTH / 2 >= Ycoupe))
                lerror = TRUE;
            if( !must_cross )
            {
                if (mc_IsVer(pseg->TYPE)
                    && (pseg->Y1 < Ycoupe)
                    && (pseg->Y2 >= Ycoupe))
                    lerror = TRUE;
            }
            break;
        }

        pseg->Y1 = streched_y(pseg->Y1);
        pseg->Y2 = streched_y(pseg->Y2);

        if (lerror != FALSE)
        {
            fflush( stdout );
            fprintf( stderr, "*** strecher error *** : Unable to cut transistor.\n" );
            exit( 1 );
        }
    }

}

/*---------------------------------------------------------\
                     strech_ab
\---------------------------------------------------------*/
static void     strech_ab()
{

    ptfig_WORK->YAB2 = ptfig_WORK->YAB2 + SX_YADD;
}

/*---------------------------------------------------------\
                     strech_cell
\---------------------------------------------------------*/
long            strech_cell(a_ptfig_WORK, a_supply, a_Ycoupe, a_NB_pitch, must_cross)
phfig_list     *a_ptfig_WORK;
supply_list    *a_supply;
long            a_Ycoupe;
long            a_NB_pitch;
int             must_cross;
{
    ptfig_WORK = a_ptfig_WORK;
    Ycoupe = a_Ycoupe;
    NB_pitch = a_NB_pitch;
    SX_YADD = NB_pitch * SX_PITCH_TALU2;
    psupply = a_supply;

    PPHSEGPOL = (ptype_list *) NULL;

    strech_ab();
    strech_con();
    strech_seg(must_cross);
    strech_via();
    strech_ref();
    strech_ins();
    strech_nwell();

    freeptype(PPHSEGPOL);

    return (SX_YADD);
}

/*---------------------------------------------------------\
                     strech_supply
\---------------------------------------------------------*/
void            strech_supply(ptfig, ptcut)
phfig_list     *ptfig;
phcut_list     *ptcut;
{
#define  access_supply  ( (supply_list*)(ptuser_supply->DATA) )

    ptype_list     *ptuser_supply;
    phcut_list     *pcut;
    supply_list    *psupply;
    long            y_sum;


    if ((ptuser_supply = get_ls_data(ptfig, C_SUPP_USUPPLY, FALSE))
                           == (ptype_list *) NULL)
        return;

    y_sum = 0;
    for (pcut = ptcut; pcut != (phcut_list *) NULL;
         pcut = pcut->NEXT)
    {
        NB_pitch = pcut->PITCHNB;
        Ycoupe = pcut->YCUT + y_sum;
        SX_YADD = NB_pitch * SX_PITCH_TALU2;

        for (psupply = access_supply; psupply != (supply_list *) NULL;
             psupply = psupply->NEXT)
            psupply->YSUPPLY = streched_y(psupply->YSUPPLY);

        y_sum += SX_YADD;
    }

}
