/*------------------------------------------------------------\
|                                                             |
| Tool    :                   DREAL                           |
|                                                             |
| File    :                 Dialog.c                          |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                  28.03.95                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>
# include <Xm/Xm.h>
# include <Xm/FileSB.h>
# include <Xm/SelectioB.h>
# include <Xm/PushBG.h>

# include MUT_H
# include MPH_H
# include RDS_H
# include RPR_H
# include RFM_H
# include GRD_H
# include GMX_H
# include GTB_H
# include GSB_H
# include GMF_H

# include "GMF_dialog.h"
# include "GMF_file.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          File Dialog                        |
|                                                             |
\------------------------------------------------------------*/

  DrealDialogItem DrealFileOpenDialog =

  {
    "Open File",
    DREAL_DIALOG_FILE,
    (Widget)NULL,
    (void *)CallbackFileOpenOk,
    (XtPointer)NULL,
    (void *)CallbackFileOpenCancel,
    (XtPointer)NULL
  };

  DrealDialogItem DrealFileQuitDialog =

  {
    "Do you really want to quit Dreal ?",
    DREAL_DIALOG_WARNING,
    (Widget)NULL,
    (void *)CallbackFileQuitOk,
    (XtPointer)NULL,
    (void *)CallbackFileQuitCancel,
    (XtPointer)NULL
  };

  DrealDialogItem DrealFileSaveAsDialog =

  {
    "Save As",
    DREAL_DIALOG_PROMPT,
    (Widget)NULL,
    (void *)CallbackFileSaveAsOk,
    (XtPointer)NULL,
    (void *)CallbackFileSaveAsCancel,
    (XtPointer)NULL
  };

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                  DrealBuildFileDialog                       |
|                                                             |
\------------------------------------------------------------*/

void DrealBuildFileDialog()

{
  rdsbegin();

  DrealBuildDialog( DrealMainWindow, &DrealFileSaveAsDialog );
  DrealBuildDialog( DrealMainWindow, &DrealFileOpenDialog   );
  DrealBuildDialog( DrealMainWindow, &DrealFileQuitDialog   );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                  CallbackFileQuitOk                         |
|                                                             |
\------------------------------------------------------------*/

void CallbackFileQuitOk( MyWidget, ClientData, CallData )

    Widget                        MyWidget;
    caddr_t                       ClientData;
    caddr_t                       CallData;
{
  rdsbegin();

  DrealExitDialog();

  if ( DrealHeadUndo != (drealundo *)NULL )
  {
    CallbackFileSaveAs( NULL, NULL, NULL );
    DrealHeadUndo = (drealundo *)0;
  }

  XtCloseDisplay( XtDisplay( XtParent( MyWidget ) ) );

  DrealExitErrorMessage( 0 );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                  CallbackFileSaveAsOk                       |
|                                                             |
\------------------------------------------------------------*/

void CallbackFileSaveAsOk( MyWidget, ClientData, CallData )

    Widget                        MyWidget;
    caddr_t                       ClientData;
    XmSelectionBoxCallbackStruct *CallData;
{
  char *FileName;

  rdsbegin();

  XmStringGetLtoR(  CallData->value,
                    XmSTRING_DEFAULT_CHARSET,
                   &FileName );

  FileName = DrealPostTreatString( FileName );

  XtUnmanageChild( DrealFileSaveAsDialog.WIDGET );

  DrealExitDialog();

  DrealFileSaveAs( FileName );

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                  CallbackFileSaveAsCancel                   |
|                                                             |
\------------------------------------------------------------*/

void CallbackFileSaveAsCancel( MyWidget, ClientData, CallData )

    Widget                        MyWidget;
    caddr_t                       ClientData;
    XmSelectionBoxCallbackStruct *CallData;
{
  rdsbegin();

  XtUnmanageChild( DrealFileSaveAsDialog.WIDGET );

  DrealExitDialog();

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                  CallbackFileQuitCancel                     |
|                                                             |
\------------------------------------------------------------*/

void CallbackFileQuitCancel( MyWidget, ClientData, CallData )

    Widget                        MyWidget;
    caddr_t                       ClientData;
    caddr_t                       CallData;
{
  rdsbegin();

  DrealExitDialog();

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                    CallbackFileOpenOk                       |
|                                                             |
\------------------------------------------------------------*/

void CallbackFileOpenOk( MyWidget, ClientData, FileStruct )

    Widget                            MyWidget;
    caddr_t                           ClientData;
    XmFileSelectionBoxCallbackStruct *FileStruct;
{
  char *FileName;

  rdsbegin();

  XtUnmanageChild( DrealFileOpenDialog.WIDGET );

  DrealExitDialog();

  if ( FileStruct->value != NULL )
  {
    XmStringGetLtoR( FileStruct->value,
                     XmSTRING_DEFAULT_CHARSET,
                     &FileName 
                   );

    FileName = DrealPostTreatString( FileName ); 

    if ( FileName != (char *)NULL )
    {
      if ( DrealHeadUndo != (drealundo *)NULL )
      {
        CallbackFileSaveAs( NULL, NULL, NULL );
      }

      DrealFileOpen( FileName );

      DrealChangeEditMode( DREAL_EDIT_MEASURE, 
                           DrealPromptEditMeasure );
    }
    else
    {
      DrealErrorMessage( DrealMainWindow, "Unable to load this file !" );
    }
  }

  rdsend();
}

/*------------------------------------------------------------\
|                                                             |
|                    CallbackFileOpenCancel                   |
|                                                             |
\------------------------------------------------------------*/

void CallbackFileOpenCancel( MyWidget, ClientData, FileStruct )

    Widget                            MyWidget;
    caddr_t                           ClientData;
    XmFileSelectionBoxCallbackStruct *FileStruct;
{
  rdsbegin();

  XtUnmanageChild( DrealFileOpenDialog.WIDGET );

  DrealExitDialog();

  rdsend();
}
