/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtBase.c,v 1.1.1.1 1998/10/02 15:26:39 alliance Exp $ */
/*                                                                          */
/*          Alliance CAD System 3.2b                                         */
/*             FpGen Package 1.10                                           */
/*                                                                          */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                                 */
/*                                                                          */
/*  Author      :          Jean-Paul CHAPUT                                 */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                                 */
/* ************************************************************************ */
/*  Module      :      "UtBase.c"                                           */


/* ******************************* Includes ******************************* */


#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <malloc.h>
#include  <signal.h>
#include  <unistd.h>

#include  "GenericGN.h"


/* ************************** Internal constants ************************** */


#define   IE_GfpUT_ModuleNameDup           3
#define   IE_GfpUT_ModuleNameVoid          4
#define   IE_GfpUT_ModuleNameUnknow        5
#define   IE_GfpUT_ModuleOverflow          6
#define   IE_GfpUT_ModuleUnknow            7
#define   IE_GfpUT_ErrorDup                8
#define   IE_GfpUT_WarngDup                9
#define   IE_GfpUT_WarngUnknow            10
#define   IE_UtBas_Malloc                 11

#define   IC_UtBas_PrintFBufferSize   (16384+1)  /* 16 Ko. */


/* ********************* Internal Types And Structures ******************** */


/* Structure holding an error. */
typedef struct error {
          long  CODE;  /* Value of the error code.       */
          char *FORM;  /* Format (printf) for the error. */
  struct error *NEXT;  /* Pointer to the next element.   */
  } error_list;

/* Structure holding an warning. */
typedef struct warng {
          long  CODE;  /* Value of the error code.       */
          char *FORM;  /* Format (printf) for the error. */
  struct warng *NEXT;  /* Pointer to the next element.   */
  } warng_list;

/* Structure for self-allocated "printf" functions. */
typedef struct allocstr {
             long  MARK;  /* Mark associated to the string. */
             char *STR;   /* String value.                  */
  struct allocstr *NEXT;  /* Pointer to the next element.   */
  } allocstr_list;


/* ************************** Internal Variables ************************** */


static          long      ModuleNB = 0L;
static          long      ModuleId;
static          char  *TabModule[C_GfpUT_ModuleMaxNB];
static    error_list  *TabError[ C_GfpUT_ModuleMaxNB];
static    warng_list  *TabWarng[ C_GfpUT_ModuleMaxNB];
static          long   ErrorMode   = C_UtBas_ErrorModeExit;
static          char   Dut_StringBuffer[C_SUPP_MaxLenString];
static          long   ArgLc_GetNB[C_GfpUT_ArgLcNB];
static          long   ArgLc_ArgC;
static          char **ArgLc_ArgV;
static          long   PrintF_Mark = 0L;
static allocstr_list  *PrintF_Head = (allocstr_list*)NULL;
static          char   PrintF_Buffer[IC_UtBas_PrintFBufferSize];


/* ************************** External Variables ************************** */


/* ************************* Interface Functions ************************** */


/* Module interface functions. */
/* extern long  GfpUT_GetModuleId(char *aName);
 * extern void  GfpUT_AddError(long  aModuleId, long  aCode, long  aForm);
 * extern void  GfpUT_PrintError(va_alist);
 * extern void  UtBas_SetErrorMode( long aMode );
 * extern void  GfpUT_AddWarng(long  aModuleId, long  aCode, long  aForm);
 * extern void  GfpUT_PrintWarng(va_alist);
 * extern void  UtBas_Init(void);   
 *
 *
 * extern void  GfpUT_InitArgLc(int  aArgC, char *aArgV[]);
 * extern char *GfpUT_GetArgLcByPos(long  aPos, long *aPtGetNB);
 * extern long  GfpUT_GetArgLcByName(char *aName, long *aPtGetNB);
 * extern char *GfpUT_GetArgLcRemain(void);
 *
 *
 * extern long  UtBas_PrintFMark(void);
 * extern long  UtBas_PrintFFree(long aMark);
 * extern char *UtBas_PrintF(va_alist);
 * extern char *UtBas_PrintFAdd( va_alist );
 * extern char *UtBas_vPrintFAdd( char *aStr, char *aFormat, va_list ArgList );
 */


/* ************************** Internal Functions ************************** */


#ifndef  ANSI_PROTO
  static void  ExitError();
#else
  static void  ExitError( void );
#endif


/* ************************* Functions Definitions ************************ */


/*  Function  :  'GfpUT_GetModuleId'.                             */
/* -------------------------------------------------------------- */
/*  Affect an identifier to the module name "aName".
 *      This function ensure that each module have a different
 *  identifier from one another. So, this is useful when filling
 *  the error and warning tables of one module.
 *      The function return the identifier of the new module.
 *      With the flag "aFlags" you can specify the function
 *  behavior:
 *      - C_GfpUT_ModuleAdd:
 *            Add a new module and check if it is The only one
 *            with the name "aName".
 *      - C_GfpUT_ModuleGet:
 *            Return the identifier of an already allocated
 *            module.
 */

extern long  GfpUT_GetModuleId( aName, aFlags )
    char *aName;   /* Name that identify the module.    */
    long  aFlags;  /* Select what operation to perform. */
{
    char *Name;
    long  Id;

    /* Check the "aName" argument.
     */
    if (aName == (char*)NULL)
        GfpUT_PrintError(ModuleId,
                       IE_GfpUT_ModuleNameVoid,
                                 "GetModuleId");

    /* Allocation of the new element.
     */
    Name = (char*)mbkalloc(strlen(aName)+1);
    strcpy(Name, aName);


    /* Search in the table of already declared modules.
     */
    for(Id=0; Id<ModuleNB; Id++)
        if (!strcmp(TabModule[Id],Name))
        {
            if (aFlags & F_GfpUT_ModuleGet)
                return(Id);
            else
                GfpUT_PrintError(ModuleId,
                               IE_GfpUT_ModuleNameDup,
                                        "GetModuleId", Name);
        }

    /* If the module has not been found and his addition
     * to the table is not asked : this is an error.
     */
    if (!(aFlags & F_GfpUT_ModuleAdd))
        GfpUT_PrintError(ModuleId,
                       IE_GfpUT_ModuleNameUnknow,
                                   "GetModuleId", Name);

    /* Addition of the new module.
     */
    if (ModuleNB >= C_GfpUT_ModuleMaxNB)
        GfpUT_PrintError(ModuleId,
                       IE_GfpUT_ModuleOverflow,
                                 "GetModuleId", Name, C_GfpUT_ModuleMaxNB);

    TabModule[ModuleNB] = Name;

    return(ModuleNB++);
}


/*  Function  :  'GfpUT_AddError'.                                */
/* -------------------------------------------------------------- */
/*  Addition of a new error in the internal table.                */

extern void  GfpUT_AddError(aModuleId, aCode, aForm)
    long  aModuleId;  /* Module which the error belong.    */
    long  aCode;      /* Code of the error.                */
    char *aForm;      /* Format to use to print the error. */
{
    error_list *PtErr, *PtNew;

    /* Check if the module exist. */
    if (aModuleId >= ModuleNB)
        GfpUT_PrintError(ModuleId,
                         IE_GfpUT_ModuleUnknow,
                                    "AddError", aModuleId);

    /* Check the error code unicity. */
    for(PtErr  = TabError[aModuleId];
        PtErr != (error_list*)NULL;
        PtErr  = PtErr->NEXT)
        if (PtErr->CODE == aCode)
            GfpUT_PrintError(ModuleId,
                             IE_GfpUT_ErrorDup,
                                    "AddError",
                                         aCode,
                          TabModule[aModuleId]);

    /* Allocation of the new element. */
    PtNew = (error_list*)mbkalloc(sizeof(error_list));
    PtNew->FORM = (char*)mbkalloc(sizeof(char)*strlen(aForm)+1);

    /* Filling the fields of the new element */
           PtNew->CODE = aCode;
    strcpy(PtNew->FORM,  aForm);

    /* Addition of the new element in head of list. */
    PtNew->NEXT = TabError[aModuleId];
                  TabError[aModuleId] = PtNew;
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "GfpUtPrintError()"                                  | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Print error then exit.                                                 |%
 */

extern void  GfpUT_PrintError __FPV(( long aId, ... ))
{
  __KR_C( long  aId; )
          long  aCode;
       va_list  ArgList;
    error_list *PtErr;

  /* Get the two first arguments. */
  __ANSI_C( va_start( ArgList, aId ); )
    __KR_C( va_start( ArgList ); )
    __KR_C( aId   = va_arg( ArgList, long); )
            aCode = va_arg( ArgList, long);
  /* Check if the module exist. */
  if (aId >= ModuleNB) {
    va_end(ArgList);
    GfpUT_PrintError( ModuleId, IE_GfpUT_ModuleUnknow, "PrintError", aId );
  }
  /* Print all the "stdout" remainings messages. */
  fflush(  stdout );
  fprintf( stderr, "Error:*** %s ***:", TabModule[aId] );
  /* Search the error associated with "aId" and "aCode". */
  for(PtErr  = TabError[aId];
      PtErr != (error_list*)NULL; PtErr = PtErr->NEXT) {
    if (aCode == PtErr->CODE) {
      /* Print the error message then exit. */
      vfprintf( stderr, PtErr->FORM, ArgList );
       fprintf( stderr, "\n" );
      va_end( ArgList );
      ExitError();
    }
  } /* End of `PtErr' loop. */
  /* The error has not been found in the table. */
  fprintf(
    stderr, "[%s]:Unknow error code:%ld (in module %s).\n",
    "PrintError", aCode, TabModule[aId] );
  fprintf( stderr, "\n" );
  va_end( ArgList );
  ExitError();
}


/*  Function  :  'UtBas_SetErrorMode'.                                      */
/* ------------------------------------------------------------------------ */
/*  Select the method to use after an error occurs.                         */

extern void  UtBas_SetErrorMode( aMode )
  long  aMode;
{
  switch( aMode )
  {
    case C_UtBas_ErrorModeKill:
    case C_UtBas_ErrorModeExit:

      ErrorMode = aMode;
      break;

    default:

      ErrorMode = C_UtBas_ErrorModeExit;

      fprintf( stderr, "*** UtBase Warning ***:SetErrorMode:" );
      fprintf( stderr, "Invalid error exit mode %ld.\n", aMode );
      fprintf( stderr, "--> Error Mode reset to \"UtBas_ErrorModeExit\".\n" );
  }
}


/*  Function  :  'ExitError'.                                               */
/* ------------------------------------------------------------------------ */
/*  Terminate the program after an error occurs. This is done by sending a
 *  SIGTERM signal or executing an 'exit( 1 )' according to "ErrorMode".
 */

static void  ExitError()
{
  switch( ErrorMode )
  {
    case C_UtBas_ErrorModeKill:

      (void)kill( getpid(), SIGTERM );

      /* If Kill fail, use exit. */
      fprintf( stderr, "*** UtBase Error ***:ExitError:" );
      fprintf( stderr, "\'kill()\' fail, attempt to use \'exit()\'.\n" );

    case C_UtBas_ErrorModeExit:
    default:

      exit( 1 );
  }
}


/*  Function  :  'GfpUT_AddWarng'.                                */
/* -------------------------------------------------------------- */
/*  Addition of a new warning in the internal table.              */

extern void  GfpUT_AddWarng(aModuleId, aCode, aForm)
    long  aModuleId;  /* Module which the warning belong.    */
    long  aCode;      /* Code of the warning.                */
    char *aForm;      /* Format to use to print the warning. */
{
    warng_list *PtWar, *PtNew;

    /* Check if the module exist. */
    if (aModuleId >= ModuleNB)
        GfpUT_PrintError(ModuleId,
                         IE_GfpUT_ModuleUnknow,
                                    "AddWarng", aModuleId);

    /* Check the warning code unicity. */
    for(PtWar  = TabWarng[ModuleId];
        PtWar != (warng_list*)NULL;
        PtWar  = PtWar->NEXT)
        if (PtWar->CODE == aCode)
            GfpUT_PrintError(ModuleId,
                             IE_GfpUT_WarngDup,
                                    "AddWarng",
                                         aCode,
                          TabModule[aModuleId]);

    /* Allocation of the new element. */
    PtNew = (warng_list*)mbkalloc(sizeof(warng_list));
    PtNew->FORM = (char*)mbkalloc(sizeof(char)*strlen(aForm)+1);

    /* Filling the fields of the new element */
           PtNew->CODE = aCode;
    strcpy(PtNew->FORM,  aForm);

    /* Addition of the new element in head of list. */
    PtNew->NEXT = TabWarng[aModuleId];
                  TabWarng[aModuleId] = PtNew;
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "GfpUtPrintWarng()"                                  | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Print warning then exit.                                               |%
 */

extern void  GfpUT_PrintWarng __FPV(( long aId, ... ))
{
  __KR_C( long  aId; )
          long  aCode;
       va_list  ArgList;
    warng_list *PtWar;

  /* Get the two first arguments. */
  __ANSI_C( va_start( ArgList, aId ); )
    __KR_C( va_start( ArgList ); )
    __KR_C( aId   = va_arg( ArgList, long); )
            aCode = va_arg( ArgList, long);
  /* Check if the module exist. */
  if (aId >= ModuleNB) {
    va_end(ArgList);
    GfpUT_PrintError( ModuleId, IE_GfpUT_ModuleUnknow, "PrintWarng", aId );
  }
  /* Print all the "stdout" remainings messages. */
  fflush( stdout);
  fprintf(stderr, "Warning:*** %s ***:", TabModule[aId]);
  /* Search the warng associated with "aId" and "aCode". */
  for(PtWar  = TabWarng[aId];
      PtWar != (warng_list*)NULL; PtWar = PtWar->NEXT)
    if (aCode == PtWar->CODE) {
      /* Print the warng message then exit. */
      vfprintf( stderr, PtWar->FORM, ArgList );
       fprintf( stderr, "\n" );
      va_end( ArgList );
      return;
    }
  /* The warng has not been found in the table. */
  va_end(ArgList);
  fprintf(stderr, "\n\n\n");
  GfpUT_PrintError(
    ModuleId, IE_GfpUT_WarngUnknow, "PrintWarng", aCode, TabModule[aId] );
}


/*  Function  :  'UtBas_Init()'.                                            */
/* ------------------------------------------------------------------------ */
/*  Perform all initialisations neededs to use the functions of the utilies
 *  module :
 *        I - Detect the size of a "long int".
 *       II - Load internal errors table.
 *      III - Load internal warnings table.
 */

extern void  UtBas_Init()
{
  static char  ModuleName[] = "UtBase";
  static long  FirstCall    = TRUE;
         long  Id;

  if (FirstCall == FALSE) return;
      FirstCall  = FALSE;

  /* Initialize the error and warning hash table.
   */
  for(Id=0; Id<C_GfpUT_ModuleMaxNB; Id++)
  {
    TabModule[Id] =       (char*)NULL;
     TabError[Id] = (error_list*)NULL;
     TabWarng[Id] = (warng_list*)NULL;
  }

  /* Force the "ModuleId" to zero, in order to self-load
   * the "UT.c" internal error and warning table.
   */
  TabModule[0] = ModuleName;
     ModuleId  = 0;
     ModuleNB  = 1;


  /* Errors associated to the module identifier allocator.
   */
  GfpUT_AddError(ModuleId, IE_GfpUT_ModuleNameDup,
  "[%s]:Duplicate module name \"%s\".\n");
  GfpUT_AddError(ModuleId, IE_GfpUT_ModuleNameVoid,
  "[%s]:(null) pointer passed as \"aName\" argument.\n");
  GfpUT_AddError(ModuleId, IE_GfpUT_ModuleNameUnknow,
  "[%s]:Unknow module name \"%s\".\n");
  GfpUT_AddError(ModuleId, IE_GfpUT_ModuleOverflow,
  "[%s]:Can\'t add module \"%s\" Table overflow (>%ld).\n");


  /* Errors associated to the errors and warnings functions.
   */
  GfpUT_AddError(ModuleId, IE_GfpUT_ModuleUnknow,
  "[%s]:Unknow module identifier:%ld.\n");
  GfpUT_AddError(ModuleId, IE_GfpUT_ErrorDup,
  "[%s]:Duplicate error code %ld in module \"%s\".\n");
  GfpUT_AddError(ModuleId, IE_GfpUT_WarngDup,
  "[%s]:Duplicate warning code %ld in module \"%s\".\n");
  GfpUT_AddError(ModuleId, IE_GfpUT_WarngUnknow,
  "[%s]:Unknow warning code %ld in module \"%s\".\n");


  /* Errors associated to the self-allocated print functions.
   */
  GfpUT_AddError(ModuleId, IE_UtBas_Malloc,
  "[%s]:Not enought memory.\n");
}


/*  Function  :  'GfpUT_InitArgLc'.                               */
/* -------------------------------------------------------------- */
/*  Give the access to all command line management functions to
 *  the standart 'main' function argument "argc" and "argv".
 *      Initialise module internal variables.
 */

extern void  GfpUT_InitArgLc(aArgC,aArgV)
     int  aArgC;    /* The "argc" argument of the 'main' function. */
    char *aArgV[];  /* The "argv" argument of the 'main' function. */
{
    long  ArgPos;

    ArgLc_ArgV =       aArgV;
    ArgLc_ArgC = (long)aArgC;

    /* Internal variables initialisation. */
    for(ArgPos=0; ArgPos<C_GfpUT_ArgLcNB; ArgPos++)
        ArgLc_GetNB[ArgPos] = 0;
}


/*  Function  :  'GfpUT_GetArgLcByPos'.                           */
/* -------------------------------------------------------------- */
/*  Return the argument at a given position.                      */

extern char *GfpUT_GetArgLcByPos(aPos    ,
                                 aPtGetNB)
    long  aPos;      /* Position of the wanted argument.              */
    long *aPtGetNB;  /* Number of time the argument have been called. */
{
    /* Check for out of range positions. */
    if (   (aPos >= C_GfpUT_ArgLcNB) 
        || (aPos >= ArgLc_ArgC      ))
    {
        *aPtGetNB = 0;

        return((char*)NULL);
    }

    /* Number of time the argument have been called. */
    *aPtGetNB = ++(ArgLc_GetNB[aPos]);

    return(ArgLc_ArgV[aPos]);
}


/*  Function  :  'GfpUT_GetArgLcByName'.                          */
/* -------------------------------------------------------------- */
/*  Look for the argument named "aName" in the command line, and 
 *  return it position.
 */

extern long  GfpUT_GetArgLcByName(aName   ,
                                  aPtGetNB)
    char *aName;     /* Name of the wanted argument.                  */
    long *aPtGetNB;  /* Number of time the argument have been called. */
{
    long  ArgPos;

    for(ArgPos=0; ArgPos<ArgLc_ArgC; ArgPos++)
    {
        /* Check for out of range positions. */
        if (ArgPos >= C_GfpUT_ArgLcNB) break;

        if (!strcmp(aName,ArgLc_ArgV[ArgPos]))
        {
            /* Number of time the argument have been called. */
            *aPtGetNB = ++(ArgLc_GetNB[ArgPos]);

            return(ArgPos);
        }
    }
    *aPtGetNB = 0;

    /* The search have fail. */
    return(C_GfpUT_ArgLcPosUnknow);
}


/*  Function  :  'GfpUT_GetArgLcRemain'.                          */
/* -------------------------------------------------------------- */
/*  Look for the first unaxamined argument of the command line.
 *  If a such argument is not found, a null pointer is returned.
 */

extern char *GfpUT_GetArgLcRemain()
{
    long  ArgPos;

    for(ArgPos=0; ArgPos<ArgLc_ArgC; ArgPos++)
    {
        /* Check for out of range positions. */
        if (ArgPos >= C_GfpUT_ArgLcNB) break;

        /* Look for the first unexamined argument. */
        if (ArgLc_GetNB[ArgPos] == 0)
        {
            /* Number of time the argument have been called. */
            ++(ArgLc_GetNB[ArgPos]);

            return(ArgLc_ArgV[ArgPos]);
        }
    }

    /* The search have fail, no unexamined arguments. */
    return((char*)NULL);
}


/*  Function  :  'UtBas_PrintFMark'.                                        */
/* ------------------------------------------------------------------------ */

extern long  UtBas_PrintFMark()
{
  return(PrintF_Mark++);
}


/*  Function  :  'UtBas_PrintFFree'.                                        */
/* ------------------------------------------------------------------------ */

extern long  UtBas_PrintFFree(aMark)
  long  aMark;
{
  allocstr_list *PtDel;


  while((PrintF_Head != (allocstr_list*)NULL) && (PrintF_Head->MARK > aMark))
  {
    PtDel = PrintF_Head; PrintF_Head = PtDel->NEXT;

    free((char*)(PtDel->STR));
    free((char*) PtDel);
  }

  return(--PrintF_Mark);
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "UtBas_PrintF()"                                     | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern char *UtBas_PrintF __FPV(( char* aFormat, ... ))
{
  allocstr_list *PtNew;
        va_list  ArgList;
  __KR_C(  char *aFormat; )

  /* Get the first argument of the list. */
  __ANSI_C( va_start( ArgList, aFormat ); )
    __KR_C( va_start( ArgList ); )
    __KR_C( aFormat = va_arg(ArgList, char*); )
  /* Print the remainings arguments in a string. */
  vsprintf( PrintF_Buffer, aFormat, ArgList );
  va_end( ArgList );

  PtNew = (allocstr_list*)malloc( sizeof(allocstr_list) );
  if (PtNew == (allocstr_list*)NULL)
    GfpUT_PrintError( ModuleId, IE_UtBas_Malloc, "UtBas_PrintF" );
  PtNew->STR = (char*)malloc( strlen(PrintF_Buffer) + 1 );
  if (PtNew->STR == (char*)NULL)
    GfpUT_PrintError( ModuleId, IE_UtBas_Malloc, "UtBas_PrintF" );
  else
    strcpy( PtNew->STR, PrintF_Buffer );
  PtNew->MARK = PrintF_Mark;
  PtNew->NEXT = PrintF_Head;
  PrintF_Head = PtNew;
  return( PtNew->STR );
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "UtBas_PrintFAdd()"                                  | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern char *UtBas_PrintFAdd __FPV(( char* aStr, ... ))
{
          va_list  ArgList;
  __KR_C(    char *aStr; )
             char *aFormat;

  /* Get the two first arguments. */
  __ANSI_C( va_start( ArgList, aStr ); )
    __KR_C( va_start( ArgList ); )
    __KR_C( aStr    = va_arg( ArgList, char*); )
            aFormat = va_arg( ArgList, char*);
  UtBas_vPrintFAdd( aStr, aFormat, ArgList );
  va_end( ArgList );
}


/*  Function  :  'UtBas_vPrintFAdd'.                                        */
/* ------------------------------------------------------------------------ */

extern char *UtBas_vPrintFAdd( aStr, aFormat, ArgList )
     char *aStr;
     char *aFormat;
  va_list  ArgList;
{
  allocstr_list  *PtNew;
  allocstr_list  *PtDel;
  allocstr_list **PPtAllocStr;


  vsprintf(PrintF_Buffer, aFormat, ArgList);


  if (aStr == (char*)NULL) aStr = "";

  PtNew = (allocstr_list*)malloc(sizeof(allocstr_list));
  if (PtNew == (allocstr_list*)NULL)
    GfpUT_PrintError(ModuleId, IE_UtBas_Malloc, "UtBas_PrintF");

  PtNew->STR = (char*)malloc(strlen(PrintF_Buffer) + strlen(aStr) + 1);
  if (PtNew->STR == (char*)NULL)
    GfpUT_PrintError(ModuleId, IE_UtBas_Malloc, "UtBas_PrintF");
  else
  {
    strcpy(PtNew->STR,         aStr);
    strcpy(PtNew->STR + strlen(aStr), PrintF_Buffer);
  }

  PtNew->MARK = PrintF_Mark;
  PtNew->NEXT = PrintF_Head;
  PrintF_Head = PtNew;

  PPtAllocStr = &PrintF_Head;
  while( *PPtAllocStr != (allocstr_list*)NULL )
  {
    if ((*PPtAllocStr)->STR == aStr)
    {
               PtDel = (*PPtAllocStr);
      (*PPtAllocStr) = (*PPtAllocStr)->NEXT;
  
      free((char*)(PtDel->STR));
      free((char*) PtDel);

      break;
    }
    else
    {
       PPtAllocStr = &((*PPtAllocStr)->NEXT);
    }
  }

  return(PtNew->STR);
}
