/*
 * $Author: ac $
 * $Id: bflvec.c,v 1.1 1999/10/21 12:51:53 ac Exp $
 * $Date: 1999/10/21 12:51:53 $
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <stdlib.h>
#include BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

#define __copy(first, last, Result) { \
    bflptr_t *__first  = (first); \
    bflptr_t *__Result = (Result); \
    while (__first != (last)) \
        *__Result++ = *__first++; \
}

static void
bflvecinsertaux(vec, pos, data)
    bflvec_t *vec;
    bflptr_t *pos;
    bflptr_t data;
{
    bflptr_t *Ptr;

    if (vec->finish != vec->end_of_storage) {
	for (Ptr = vec->finish; Ptr != pos; --Ptr)
	    *Ptr = *(Ptr - 1);
	*pos = data;
	vec->finish++;
    } else {
	bflsize_t Len = bflvecsize(vec) ? 2 * bflvecsize(vec) : 32;
	bflptr_t *Node = bflalloc(Len * sizeof(bflptr_t));

	if (vec->start) {
	    __copy(vec->start, pos, Node);
	    *(Node + (pos - vec->start)) = data;
	    __copy(pos, vec->finish, Node + (pos - vec->start) + 1);
	    free(vec->start);
	} else {
	    *Node = data;
	}
	vec->end_of_storage = Node + Len;
	vec->finish = Node + bflvecsize(vec) + 1;
	vec->start = Node;
    }
}

static void
fixheap(Base, pos, Size, compar)
    bflptr_t *Base;
    bflsize_t pos;
    bflsize_t Size;
    bflcompar_fn compar;
{
    bflsize_t Len;
    bflptr_t Ptr;

    while ((Len = 2 * pos) <= Size) {
	if (Len != Size && compar(Base[Len + 1], Base[Len]) > 0)
	    ++Len;

	if (compar(Base[pos], Base[Len]) >= 0)
	    return;

	Ptr = Base[Len];
	Base[Len] = Base[pos];
	Base[pos] = Ptr;
	pos = Len;
    }
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bflvec_t *
bflvecnew()
{
    bflvec_t *vec;

    vec = (bflvec_t *) bflmemalloc(sizeof(bflvec_t));
    vec->start = 0;
    vec->finish = 0;
    vec->end_of_storage = 0;
    return vec;
}

void
bflvecfree(vec)
    bflvec_t *vec;
{
    if (vec->start)
	free(vec->start);
    bflmemfree(vec, sizeof(bflvec_t));
}

bflvec_t *
bflveccopy(vec)
    bflvec_t *vec;
{
    bflvec_t *copy = bflvecnew();
    bflptr_t *first = bflvecbegin(vec);
    bflptr_t *last = bflvecend(vec);

    while (first != last) {
	bflvecpushback(copy, bflvecdata(first));
	first = bflvecnext(first);
    }
    return copy;
}

bflptr_t *
bflvecinsert(vec, pos, data)
    bflvec_t *vec;
    bflptr_t *pos;
    bflptr_t data;
{
    bflsize_t Size = pos - vec->start;

    if (vec->finish != vec->end_of_storage && pos == vec->finish)
	*vec->finish++ = data;
    else
	bflvecinsertaux(vec, pos, data);
    return vec->start + Size;
}

void
bflvecerase(vec, pos)
    bflvec_t *vec;
    bflptr_t *pos;
{
    while (pos != vec->finish) {
	bflptr_t *next = bflvecnext(pos);

	bflvecdata(pos) = bflvecdata(next);
	pos = next;
    }
    vec->finish--;
}

bflptr_t *
bflvecfind(vec, compar, data)
    bflvec_t *vec;
    bflcompar_fn compar;
    bflptr_t data;
{
    bflptr_t *first = bflvecbegin(vec);
    bflptr_t *last = bflvecend(vec);

    while (first != last) {
	if (compar(bflvecdata(first), data) == 0)
	    return first;
	first = bflvecnext(first);
    }
    return 0;
}

bflptr_t *
bflvecfindptr(vec, data)
    bflvec_t *vec;
    bflptr_t data;
{
    bflptr_t *first = bflvecbegin(vec);
    bflptr_t *last = bflvecend(vec);

    while (first != last) {
	if (bflvecdata(first) == data)
	    return first;
	first = bflvecnext(first);
    }
    return 0;
}

void
bflvecremove(vec, compar, data)
    bflvec_t *vec;
    bflcompar_fn compar;
    bflptr_t data;
{
/* 
 * Watch: Iteration in reverse order not to invalidate the iterator !
 */
    bflptr_t *first = bflvecrbegin(vec);
    bflptr_t *last = bflvecrend(vec);

    while (first != last) {
	if (compar(bflvecdata(first), data) == 0)
	    bflvecerase(vec, first);
	first = bflvecprev(first);
    }
}

void
bflvecsort(vec, compar)
    bflvec_t *vec;
    bflcompar_fn compar;
{
    bflsize_t size;
    bflsize_t i;
    bflptr_t *base;
    bflptr_t tmp;

    base = vec->start - 1;
    size = bflvecsize(vec);

    for (i = size / 2; i > 1; --i)
	fixheap(base, i, size, compar);

    while (size > 1) {
	fixheap(base, 1, size, compar);
	tmp = base[1];
	base[1] = base[size];
	base[size] = tmp;
	size--;
    }
}

bflptr_t *
bflvecsearch(vec, compar, key)
    bflvec_t *vec;
    bflcompar_fn compar;
    bflptr_t key;
{
    int comp;
    bflsize_t index;
    bflsize_t lower = 0;
    bflsize_t upper = bflvecsize(vec);

    while (lower < upper) {
	index = (lower + upper) / 2;
	comp = compar(vec->start[index], key);
	if (comp < 0)
	    upper = index;
	else if (comp > 0)
	    lower = index + 1;
	else
	    return (vec->start + index);
    }
    return 0;
}
