/*
 * $Author: ac $
 * $Id: butcompil.c,v 1.2 1999/10/21 12:51:59 ac Exp $
 * $Date: 1999/10/21 12:51:59 $
 */
/*
 * $Log: butcompil.c,v $
 * Revision 1.2  1999/10/21 12:51:59  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:39  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  BDD_H
#include  ABE_H
#include  BUT_H

/*----------------------------------------------------------------------*/
/*                   char *is_stable( ablexpr *Expr )                   */
/* -------------------------------------------------------------------- */
/*	Verifie qu'une expression est de la forme:			*/
/*		(not (stable "atom"))					*/
/* -------------------------------------------------------------------- */
/* Parametres:								*/
/*	un pointeur sur un abl.						*/
/* Retours:								*/
/*	atom si l'expression matche le pattern, 0 sinon			*/
/*----------------------------------------------------------------------*/

static char *
is_stable( Expr )
    ablexpr	*Expr;
{
    if ( Expr == 0 )
	return 0;

    if ( ABL_ATOM( Expr ) || ABL_OPER( Expr ) != ABL_NOT )
	return 0;
    
    Expr = ABL_CADR( Expr );
    if ( ABL_ATOM( Expr ) || ABL_OPER( Expr ) != ABL_STABLE )
	return 0;
    
    Expr = ABL_CADR( Expr );
    if ( ABL_ATOM( Expr ) )
        return ABL_ATOM_VALUE( Expr );
    
    return 0;
}

/*----------------------------------------------------------------------*/
/*                    char *is_ff_re( ablexpr *Expr)                    */
/* -------------------------------------------------------------------- */
/*	Verifie si l'expression expr represente la condition d'horloge	*/
/* d'un flip-flop sur front montant (rising edge).			*/
/*	Les patterns reconnus sont les suivant:				*/
/*		(and "atom" (not (stable "atom")))			*/
/*		(and (not (stable "atom")) "atom")			*/
/* -------------------------------------------------------------------- */
/* Parametres:								*/
/*	un pointeur sur un abl.						*/
/* Retours:								*/
/*	atom si l'expression matche un des patterns, 0 sinon		*/
/*----------------------------------------------------------------------*/

static char *
is_ff_re( Expr )
    ablexpr	*Expr;
{
    char	*Name;

    if ( ABL_ATOM( Expr ) || ( ABL_OPER( Expr ) != ABL_AND ) )
	return 0;
    
    if ( ABL_ATOM( ABL_CADR( Expr ) ) ) {
	
	Name = ABL_ATOM_VALUE(ABL_CADR(Expr));
	
	if ( is_stable( ABL_CAR( ABL_CDDR( Expr ) ) ) == Name )
	    return Name;
	else
	    return 0;
    }
    
    if ( ABL_ATOM( ABL_CAR( ABL_CDDR( Expr ) ) ) ) {
	
	Name = ABL_ATOM_VALUE( ABL_CAR( ABL_CDDR( Expr ) ) );
	
	if ( is_stable( ABL_CADR( Expr ) ) == Name )
	    return Name;
	else
	    return 0;
    }
    return 0;
}

/*----------------------------------------------------------------------*/
/*                    char *is_ff_fe( ablexpr *Expr)                    */
/* -------------------------------------------------------------------- */
/*	Verifie si l'expression expr represente la condition d'horloge	*/
/* d'un flip-flop sur front descendant (falling edge).			*/
/*	Les patterns reconnus sont les suivant:				*/
/*		(and (not "atom") (not (stable "atom")))		*/
/*		(and (not (stable "atom")) (not "atom"))		*/
/* -------------------------------------------------------------------- */
/* Parametres:								*/
/*	un pointeur sur un abl.						*/
/* Retours:								*/
/*	atom si l'expression matche un des patterns, 0 sinon		*/
/*----------------------------------------------------------------------*/

static char *is_ff_fe( Expr )
    ablexpr	*Expr;
{
    char	*Name;
    
    if ( ABL_ATOM( Expr ) || ( ABL_OPER( Expr ) != ABL_AND ) )
	return 0;
    
    if ( ( ABL_ATOM( ABL_CADR( Expr ) )            ) ||
	 ( ABL_ATOM( ABL_CAR( ABL_CDDR( Expr ) ) ) ) )
	return 0;
    
    if ((ABL_OPER(ABL_CADR(Expr)) != ABL_NOT         ) ||
	(ABL_OPER(ABL_CAR(ABL_CDDR(Expr))) != ABL_NOT))
	return 0;
    
    if (ABL_ATOM(ABL_CADR(ABL_CADR(Expr)))) {
	
	Name = ABL_ATOM_VALUE( ABL_CADR( ABL_CADR( Expr ) ) );
	
	if ( is_stable( ABL_CAR( ABL_CDDR( Expr ) ) ) == Name )
	    return Name;
	else
	    return 0;
    }
    
    if ( ABL_ATOM( ABL_CADR( ABL_CAR( ABL_CDDR( Expr ) ) ) ) ) {
	
	Name = ABL_ATOM_VALUE( ABL_CADR( ABL_CAR( ABL_CDDR( Expr ) ) ) );
	
	if ( is_stable( ABL_CADR( Expr ) ) == Name )
	    return Name;
	else
	    return 0;
    }
    return 0;
}

/*----------------------------------------------------------------------*/
/*                  char *is_latch_ah( ablexpr *Expr )                  */
/* -------------------------------------------------------------------- */
/*	Verifie si l'expression expr represente la condition d'horloge	*/
/* d'un latch actif a un (active high).					*/
/*	Les patterns reconnus sont les suivant:				*/
/*		("atom")						*/
/* -------------------------------------------------------------------- */
/* Parametres:								*/
/*	un pointeur sur un abl.						*/
/* Retours:								*/
/*	atom si l'expression matche un des patterns, 0 sinon		*/
/*----------------------------------------------------------------------*/

static char *
is_latch_ah( Expr )
    ablexpr	*Expr;
{
    if ( ABL_ATOM( Expr ) )
	return ABL_ATOM_VALUE( Expr );
    else
	return 0;
}

/*----------------------------------------------------------------------*/
/*                  char *is_latch_al( ablexpr *Expr )                  */
/* -------------------------------------------------------------------- */
/*	Verifie si l'expression expr represente la condition d'horloge	*/
/* d'un latch actif a zero (active low).				*/
/*	Les patterns reconnus sont les suivant:				*/
/*		(not "atom")						*/
/* -------------------------------------------------------------------- */
/* Parametres:								*/
/*	un pointeur sur un abl.						*/
/* Retours:								*/
/*	atom si l'expression matche un des patterns, 0 sinon		*/
/*----------------------------------------------------------------------*/

static char *
is_latch_al( Expr )
    ablexpr	*Expr;
{
    if ( ABL_ATOM( Expr ) || ABL_OPER( Expr ) != ABL_NOT )
	return 0;
    
    Expr = ABL_CADR( Expr );
    if ( ABL_ATOM( Expr ) )
	return ABL_ATOM_VALUE( Expr );
    else
	return 0;
}

/*----------------------------------------------------------------------*/
/*      void get_reg_inputs( RegName, Expr, DataExpr, EnableExpr )      */
/* -------------------------------------------------------------------- */
/*	Recheche si Expr represente une entree multiplexee de registre.	*/
/*	Les patterns reconnus sont les suivant:				*/
/*	(or (and "RegName" EnableExpr)(and (not EnableExpr) DataExpr))	*/
/*	(or (and "RegName" EnableExpr)(and DataExpr (not EnableExpr)))	*/
/*	(or (and EnableExpr "RegName")(and (not EnableExpr) DataExpr))	*/
/*	(or (and EnableExpr "RegName")(and DataExpr (not EnableExpr)))	*/
/*	(or (and (not EnableExpr) DataExpr)(and "RegName" EnableExpr))	*/
/*	(or (and DataExpr (not EnableExpr))(and "RegName" EnableExpr))	*/
/*	(or (and (not EnableExpr) DataExpr)(and EnableExpr "RegName"))	*/
/* 	(or (and DataExpr (not EnableExpr))(and EnableExpr "RegName"))	*/
/* -------------------------------------------------------------------- */
/* Parametres:								*/
/*	RegName	: Le nom du registre.					*/
/*	Expr	: L'expression associee					*/
/* Retours:								*/
/*	DataExpr et EnableExpr par adresse.				*/
/*----------------------------------------------------------------------*/


static void get_reg_inputs( RegName, Expr, DataExpr, EnableExpr )
    char	 *RegName;
    ablexpr	 *Expr;
    ablexpr	**DataExpr;
    ablexpr	**EnableExpr;
{
    ablexpr	 *Expr1;
    ablexpr	 *Expr2;
    
    if ( ABL_ATOM( Expr ) || ABL_OPER( Expr ) != ABL_OR ) {
	*DataExpr = Expr;
	*EnableExpr = 0;
	return;
    }
    
    /* (or Expr1 Expr2) */
    Expr1 = ABL_CADR( Expr );
    Expr2 = ABL_CAR( ABL_CDDR( Expr ) );
    
    if ( ABL_ATOM( Expr1 ) || ( ABL_OPER( Expr1 ) != ABL_AND ) ||
	 ABL_ATOM( Expr2 ) || ( ABL_OPER( Expr2 ) != ABL_AND ) ) {
	*DataExpr = Expr;
	*EnableExpr = 0;
	return;
    }

    if ( ( ABL_ATOM( ABL_CADR( Expr1 ) )                     ) &&
	 ( ABL_ATOM_VALUE( ABL_CADR( Expr1 ) ) == RegName ) ) {
	    /*
	     * Expr1 = (and "RegName" EnableExpr)
	     */
	*EnableExpr = ABL_CAR( ABL_CDDR( Expr1 ) );
	if ( isablnotexpr( *EnableExpr, ABL_CADR( Expr2 ) ) ) {
    /*
     * Expr = (or (and "RegName" EnableExpr)(and (not EnableExpr) DataExpr))
     */
	    *EnableExpr = ABL_CADR( Expr2 );
	    *DataExpr = ABL_CAR( ABL_CDDR( Expr2 ) );
	    return;
	} else if ( isablnotexpr( *EnableExpr, ABL_CAR( ABL_CDDR( Expr2 ) ) ) ) {
    /*
     * Expr = (or (and "RegName" EnableExpr)(and DataExpr (not EnableExpr)))
     */
	    *EnableExpr = ABL_CAR( ABL_CDDR( Expr2 ) );
	    *DataExpr = ABL_CADR( Expr2 );
	    return;
	}
    } else if ((ABL_ATOM(ABL_CAR(ABL_CDDR(Expr1)))                     ) &&
	       (ABL_ATOM_VALUE(ABL_CAR(ABL_CDDR(Expr1))) == RegName)) {
	    /*
	     * Expr1 = (and EnableExpr "RegName")
	     */
	*EnableExpr = ABL_CADR(Expr1);
	if (isablnotexpr(*EnableExpr, ABL_CADR(Expr2))) {
    /*
     * Expr = (or (and EnableExpr "RegName") (and (not EnableExpr) DataExpr))
     */
	    *EnableExpr = ABL_CADR(Expr2);
	    *DataExpr = ABL_CAR(ABL_CDDR(Expr2));
	    return;
	} else if (isablnotexpr(*EnableExpr, ABL_CAR(ABL_CDDR(Expr2)))) {
    /*
     * Expr = (or (and EnableExpr "RegName") (and DataExpr (not EnableExpr)))
     */
	    *EnableExpr = ABL_CAR(ABL_CDDR(Expr2));
	    *DataExpr = ABL_CADR(Expr2);
	    return;
	}
    } else if ((ABL_ATOM(ABL_CADR(Expr2))                     ) &&
	       (ABL_ATOM_VALUE(ABL_CADR(Expr2)) == RegName)) {
	    /*
	     * Expr2 = (and "RegName" EnableExpr)
	     */
	*EnableExpr = ABL_CAR(ABL_CDDR(Expr2));
	if (isablnotexpr(*EnableExpr, ABL_CADR(Expr1))) {
    /*
     * Expr = (or (and (not EnableExpr) DataExpr)(and "RegName" EnableExpr))
     */
	    *EnableExpr = ABL_CADR(Expr1);
	    *DataExpr = ABL_CAR(ABL_CDDR(Expr1));
	    return;
	} else if (isablnotexpr(*EnableExpr, ABL_CAR(ABL_CDDR(Expr1)))) {
    /*
     * Expr = (or (and DataExpr (not EnableExpr))(and "RegName" EnableExpr))
     */
	    *EnableExpr = ABL_CAR(ABL_CDDR(Expr1));
	    *DataExpr = ABL_CADR(Expr1);
	    return;
	}
    } else if ((ABL_ATOM(ABL_CAR(ABL_CDDR(Expr2)))                     ) &&
	       (ABL_ATOM_VALUE(ABL_CAR(ABL_CDDR(Expr2))) == RegName)) {
	    /*
	     * Expr2 = (and EnableExpr "RegName")
	     */
	*EnableExpr = ABL_CADR(Expr2);
	if (isablnotexpr(*EnableExpr, ABL_CADR(Expr1))) {
    /*
     * Expr = (or (and (not EnableExpr) DataExpr)(and EnableExpr "RegName"))
     */
	    *EnableExpr = ABL_CADR(Expr1);
	    *DataExpr = ABL_CAR(ABL_CDDR(Expr1));
	    return;
	} else if (isablnotexpr(*EnableExpr, ABL_CAR(ABL_CDDR(Expr1)))) {
    /*
     * Expr = (or (and DataExpr (not EnableExpr))(and EnableExpr "RegName"))
     */
	    *EnableExpr = ABL_CAR(ABL_CDDR(Expr1));
	    *DataExpr = ABL_CADR(Expr1);
	    return;
	}
    }
    
    *DataExpr = Expr;
    *EnableExpr = 0;
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

void
but_compil( BehFigure )
    befig_list	*BehFigure;
{
    bereg_list	*BehReg;
    biabl_list	*BiAbl;
    char	*ClkName;
    ablexpr	*Expr;
    ablexpr	*DataExpr;
    ablexpr	*EnableExpr;

    /*
     * beh_frebeaux( FpSystem->BEFIG->BEDLY );
     * beh_frebemsg( FpSystem->BEFIG->BEMSG );
     * beh_frebegen( FpSystem->BEFIG->BEGEN );
     * beh_frebepgm( FpSystem->BEFIG->BEPGM );
     */

    printf( "   - Compiling ...\n" );

    for ( BehReg = BehFigure->BEREG; BehReg != 0; BehReg = BehReg->NEXT ) {
	
	for ( BiAbl = BehReg->BIABL; BiAbl != 0; BiAbl = BiAbl->NEXT ) {
	
	    Expr = BiAbl->CNDABL;
	    if (        ( ClkName = is_ff_re( Expr ) ) != 0 ) {
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_TYPE, BUT_FF_RE );
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_NAME, ClkName );
	    } else if ( ( ClkName = is_ff_fe( Expr ) ) != 0 ) {
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_TYPE, BUT_FF_FE );
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_NAME, ClkName );
	    } else if ( ( ClkName = is_latch_ah( Expr ) ) != 0 ) {
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_TYPE, BUT_LATCH_AH );
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_NAME, ClkName );
	    } else if ( ( ClkName = is_latch_al( Expr ) ) != 0 ) {
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_TYPE, BUT_LATCH_AL );
		BiAbl->USER = addptype( BiAbl->USER, BUT_CLOCK_NAME, ClkName );
	    } else 
		bflerror( 1, "error: invalid block condition on register `%s'", BehReg->NAME );
	    
	    Expr = BiAbl->VALABL;
	    get_reg_inputs( BehReg->NAME, Expr, &DataExpr, &EnableExpr );
	    BiAbl->VALABL = DataExpr;
	    BiAbl->CNDABL = EnableExpr;
	}
    }
}

