/*
 * $Author: ac $
 * $Id: butsweep.c,v 1.3 1999/10/21 12:52:00 ac Exp $
 * $Date: 1999/10/21 12:52:00 $
 */
/*
 * $Log: butsweep.c,v $
 * Revision 1.3  1999/10/21 12:52:00  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.2  1998/11/16 17:46:35  ac
 * 	Ajout de -D$(ALLIANCE_OS) dans les defines des Makefiles.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:39  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  BDD_H
#include  ABE_H
#include  BUT_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/* # define DEBUG */

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*           static void but_sweep_abl(befig_list *BehFigure)           */
/* -------------------------------------------------------------------- */
/*	Suppression de tous les auxiliaires de la befig dont le fanin	*/
/* est vide (i.e. toutes les constantes). On utilise pour cela l'ABL.	*/
/*----------------------------------------------------------------------*/

static void
but_sweep_abl(befig_list *BehFigure)
{
	beaux_list *BehAux = BehFigure->BEAUX;
	
	while ( BehAux != NULL ) {
		beaux_list *BehNext = BehAux->NEXT;
		chain_list *Expr    = BehAux->ABL;

		if ( Expr == (ablexpr *)0 ) {
# ifdef DEBUG
			bflwarning("but_sweep_abl: node `%s' has no abl !", BehAux->NAME);
# endif
			BehFigure->BEAUX = beh_delbeaux(BehFigure->BEAUX, BehAux, 'Y');
		} else if ( ABL_ATOM( Expr ) ) {
			/*
			 * Si l'Abl de BehAux est atomique, alors on substitue
			 * toutes les occurences de cette variable dans Figure.
			 */
# ifdef DEBUG
			bflwarning("but_sweep_abl: sweeping node `%s' with `%s'", BehAux->NAME, ABL_ATOM_VALUE( Expr ) );
# endif
			but_substname( BehFigure, BehAux->NAME, ABL_ATOM_VALUE( Expr ) );
			BehFigure->BEAUX = beh_delbeaux(BehFigure->BEAUX, BehAux, 'Y');
		}

		BehAux = BehNext;
	}
}

/*----------------------------------------------------------------------*/
/*           static void but_sweep_bdd(befig_list *BehFigure)           */
/* -------------------------------------------------------------------- */
/*	Suppression de tous les auxiliaires de la befig dont le fanin	*/
/* est vide (i.e. toutes les constantes). On utilise pour cela les BDD.	*/
/*									*/
/* Cette facon de faire est plus efficace, mais aussi plus complexe. En */
/* effet, on doit mettre a jours tous les bdds du bddcircuit associe a	*/
/* la befig, supprimer l'auxiliaire aussi bien dans la befig que dans	*/
/* le circuit.								*/
/*----------------------------------------------------------------------*/

#define BDD_CONST(Node)     ((Node)->INDEX < 2)
#define BDD_ATOM(Node)      (BDD_CONST((Node)->LOW) && BDD_CONST((Node)->HIGH))
#define BDD_SWEEP(Node)     (BDD_CONST(Node) ? 1 : BDD_ATOM(Node))

static void
but_sweep_bdd(befig_list *BehFigure)
{
	beaux_list *BehAux = BehFigure->BEAUX;
	
	while ( BehAux != NULL ) {
		beaux_list *BehNext = BehAux->NEXT;
		
		if ( BehAux->NODE == (bddnode *)0 ) {
# ifdef DEBUG
			bflwarning("but_sweep_bdd: node `%s' has no bdd !", BehAux->NAME);
# endif
		} else if ( BDD_SWEEP( BehAux->NODE ) ) {
# ifdef DEBUG
			bflwarning("but_sweep_abl: sweeping node `%s' with index `%d'", BehAux->NAME, BehAux->NODE->INDEX );
# endif
			/* sweepbddcircuitout( BehFigure->CIRCUI, BehAux->NAME ); */
			BehFigure->BEAUX = beh_delbeaux( BehFigure->BEAUX, BehAux, 'Y' );
		}
		
		BehAux = BehNext;
	}
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                 void but_sweep(befig_list *BehFigure)                */
/* -------------------------------------------------------------------- */
/*	Suppression de tous les auxiliaires de la befig dont le fanin	*/
/* est vide (i.e. toutes les constantes)				*/
/*----------------------------------------------------------------------*/

void
but_sweep(befig_list *BehFigure, int Flag)
{
#ifdef DEBUG
	beaux_list *BehAux;
	
	if ( BehFigure == NULL )
		bflerror( 1, "but_sweep: null figure" );

	printf( "sweeping figure %s\n", BehFigure->NAME );

	printf("before sweeping, BehFigure->BEAUX =\n");
	for (BehAux = BehFigure->BEAUX; BehAux != NULL; BehAux = BehAux->NEXT)
		printf("\t%s\n", BehAux->NAME);
#endif

	switch ( Flag ) {
	case BUT_SWEEP_ABL:
		but_sweep_abl( BehFigure );
		break;
		
	case BUT_SWEEP_BDD:
		but_sweep_bdd( BehFigure );
		break;

	default:
		bflwarning("but_sweep: invalid flag");
	}
#ifdef DEBUG
	printf("after sweeping, BehFigure->BEAUX =\n");
	for (BehAux = BehFigure->BEAUX; BehAux != NULL; BehAux = BehAux->NEXT)
		printf("\t%s\n", BehAux->NAME);
#endif
}

