/*
 * $Author: ac $
 * $Id: fpm_optim.c,v 1.2 1999/10/21 12:52:06 ac Exp $
 * $Date: 1999/10/21 12:52:06 $
 */
/*
 * $Log: fpm_optim.c,v $
 * Revision 1.2  1999/10/21 12:52:06  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:41  alliance
 * fpmap import
 *
 * Revision 1.3  1998/02/11 17:34:36  ac
 * fpmap version 0.62
 *
 * Revision 1.2  1998/01/06 11:56:42  ac
 * Added .lax parameter file support.
 *
 * Revision 1.1  1997/11/14 11:49:46  ac
 * Initial revision
 *
 */

/* A faire: fpm_init_cost n'est plus indipensable dans fpmsift. On peut se 
 * contenter de le lancer une seule fois. Le probleme est juste de calculer
 * CurrCost = BestCost dans fpmsift comme la somme des couts de toutes les
 * sorties du Network.
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  BDD_H
#include  ABE_H
#include "fpm_network.h"
#include "fpm_user.h"
#include "fpm_optim.h"

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static int		 FpMode;
static fpm_cost_fn	 FpCost;
static fpm_clear_fn	 FpClear;

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

static long
fpmcost(fpm_network_t * Network, fpm_expr_t * Expr)
{
	long Count = 0;
	long Depth = 0;

	(*FpCost) (Expr, &Count, &Depth);
	return ((Count * (5 - FpMode)) + (Depth * FpMode));
}

/* Compute the cost of the entire network... for stat !
 */
static void
fpm_network_cost(fpm_network_t *Network, long *Count, long *Depth)
{
	bflmapnode_t *First;
	bflmapnode_t *Last;
	fpm_expr_t *Expr;
	long LocalDepth;

	*Count = 0;
	*Depth = 0;

	(*FpClear) ((fpm_atom_t *) 0);

	First = bflmapbegin(Network->MAP_EXPR);
	Last = bflmapend(Network->MAP_EXPR);
	while (First != Last) {
		Expr = (fpm_expr_t *) bflmapdata(First);
		(*FpCost) (Expr, Count, &LocalDepth);
		if (LocalDepth > *Depth)
			*Depth = LocalDepth;
		First = bflmapnext(First);
	}
}

/* Initialize the cost field for all the exprs of the network.
 */
static long
fpm_init_cost(fpm_network_t * Network)
{
	bflmapnode_t *First;
	bflmapnode_t *Last;
	fpm_expr_t *Expr;
	long TotalCost;

	First = bflmapbegin(Network->MAP_EXPR);
	Last = bflmapend(Network->MAP_EXPR);
	TotalCost = 0;

	FpClear((fpm_atom_t *) 0);

	while (First != Last) {
		Expr = (fpm_expr_t *) bflmapdata(First);
		Expr->COST = fpmcost(Network, Expr);
		TotalCost += Expr->COST;
		First = bflmapnext(First);
	}
	return TotalCost;
}

/* Computes the cost of all the Exprs whose depend on Atom.
 */
static long
fpm_delta_cost(fpm_network_t * Network, fpm_atom_t * Atom)
{
	chain_list *Chain;
	fpm_expr_t *Expr;
	long DeltaCost;
	long OldCost;

	FpClear(Atom);

	DeltaCost = 0;
	for (Chain = Atom->BDD_EXPR; Chain != NULL; Chain = Chain->NEXT) {
		Expr = (fpm_expr_t *) Chain->DATA;
		OldCost = Expr->COST;
		Expr->COST = fpmcost(Network, Expr);
		DeltaCost += Expr->COST - OldCost;
	}

	return DeltaCost;
}

void
fpmsift(fpm_network_t * Network, fpm_atom_t * Atom)
{
	long CurrSize;
	long MaxSize;
	bddvar StartPos;
	bddvar CurrPos;
	bddvar BestPos;
	long CurrCost;
	long BestCost;
	bddindex NumberIndex;
	bddvar NumberVar;

	setbddlocalsystem(Network->BDD_SYSTEM);

	NumberIndex = Network->BDD_SYSTEM->NUMBER_INDEX;
	NumberVar = Network->BDD_SYSTEM->NUMBER_VAR;

	StartPos = CurrPos = BestPos = fpm_atom_var(Network, Atom);
	BestCost = CurrCost = fpm_init_cost(Network);

#ifdef DEBUG
	printf("sifting %s (%d)\n", Atom->NAME, StartPos);
	printf("  Initial cost: %ld\n", BestCost);
#endif
	CurrSize = Network->BDD_SYSTEM->NUMBER_NODE;
	MaxSize = 2.0 * CurrSize;

	while (CurrPos < NumberVar - 1 && CurrSize < MaxSize) {
		if (swapbddvar(BddLocalSystem, CurrPos)) {
			CurrCost += fpm_delta_cost(Network, Atom);
		}
		CurrPos += 1;
		CurrSize = BddLocalSystem->NUMBER_NODE;

		if (CurrCost < BestCost) {
			BestPos = CurrPos;
			BestCost = CurrCost;
#ifdef DEBUG
			printf("  Best cost: %ld (%d)\n", BestCost, BestPos);
#endif
		}
	}

	while (CurrPos != StartPos) {
		CurrPos -= 1;
		swapbddvar(BddLocalSystem, CurrPos);
	}

	while (CurrPos > 0 && CurrSize < MaxSize) {
		CurrPos -= 1;

		if (swapbddvar(BddLocalSystem, CurrPos))
			CurrCost += fpm_delta_cost(Network, Atom);

		CurrSize = BddLocalSystem->NUMBER_NODE;

		if (CurrCost < BestCost) {
			BestPos = CurrPos;
			BestCost = CurrCost;
#ifdef DEBUG
			printf("  Best cost: %ld (%d)\n", BestCost, BestPos);
#endif
		}
	}

	while (CurrPos != BestPos) {
		swapbddvar(BddLocalSystem, CurrPos);
		CurrPos += 1;
	}

	/* Met Atom a sa position optimale... Met a jour le cout...
	 */
	CurrCost += fpm_delta_cost(Network, Atom);
#ifdef DEBUG
	printf("  Final cost: %ld:%ld (%d)\n", BestCost, CurrCost, BestPos);
#endif
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

void
fpm_optim()
{
	long Count;
	long Depth;
	long Level;
	fpm_network_t *Network;

	FpMode  = FpmUser->MODE;
	FpCost  = FpmUser->ARCHI_COST;
	FpClear = FpmUser->ARCHI_CLEAR;
	Network = FpmUser->NETWORK;

	printf("\n * Running the Fpga optimizer on `%s' ...\n", FpmUser->FILE_VHDL);

	setbddlocalsystem(Network->BDD_SYSTEM);

	(*FpmUser->ARCHI_INIT)();

	fpm_network_cost(Network, &Count, &Depth);
	garbagebddsystem(Network->BDD_SYSTEM);

#ifdef DEBUG
	fpm_network_view(Network);
#endif
	printf("   + Initial cost\n");
	printf("      number of CLB -> %ld\n", Count);
	printf("      depth in CLB  -> %ld\n", Depth);
	printf("      number of BDD -> %ld\n", Network->BDD_SYSTEM->NUMBER_NODE);

	for (Level = 0; Level < FpmUser->LEVEL; Level++) {
		bflmapnode_t *First;
		bflmapnode_t *Last;
		fpm_atom_t *Atom;

		garbagebddsystem(BddLocalSystem);
		fpm_init_cost(Network);
#ifndef DEBUG
		printf("     o %ld [", Level + 1);
#endif

		First = bflmapbegin(Network->MAP_ATOM);
		Last = bflmapend(Network->MAP_ATOM);
		while (First != Last) {
			Atom = (fpm_atom_t *) bflmapdata(First);
#ifndef DEBUG
			putchar('.');
			fflush(stdout);
#endif
			fpmsift(Network, Atom);
			First = bflmapnext(First);
		}

#ifndef DEBUG
		printf("]\n");
#endif
	}

	fpm_network_cost(Network, &Count, &Depth);

	(*FpmUser->ARCHI_DONE)();

	garbagebddsystem(Network->BDD_SYSTEM);
	printf("   + Final cost\n");
	printf("      number of CLB -> %ld\n", Count);
	printf("      depth in CLB  -> %ld\n", Depth);
	printf("      number of BDD -> %ld\n", Network->BDD_SYSTEM->NUMBER_NODE);
}
