%{
# include <stdio.h>
# include  MUT_H
# include  BFL_H
# include  LAX_H

# define	FLAG_MODE		(1 << 0)
# define	FLAG_LEVEL		(1 << 1)
# define	FLAG_DELAY		(1 << 2)
# define	FLAG_EARLY		(1 << 3)
# define	FLAG_SIGNAL		(1 << 4)
# define	FLAG_TRANSN		(1 << 5)
# define	FLAG_TRANSP		(1 << 6)
# define	FLAG_SLOPE		(1 << 7)
# define	FLAG_CAPI		(1 << 8)
# define	FLAG_CAPO		(1 << 9)
# define	FLAG_IMPI		(1 << 10)
# define	FLAG_BUFI		(1 << 11)
# define	FLAG_CLOCK		(1 << 12)

extern	FILE		*yyin;
extern	char		*yyfilename;
extern	int		 yylineno;
extern	char		*yytext;

static	lax_param_t	*yyparam;
static	int		 yyflag;

extern	void		 yyerror	__P(( char *str ));
extern	void		 yywarning	__P(( char *str ));
extern	int		 yylex		__P(( void ));
static	char		*laxnameindex	__P(( char *Name, int Index ));

%}

%union {
    long	 Long;
    char	*String;
    ptype_list	*PType;
    chain_list	*Chain;
};

%token	<Long>		T_Long
%token	<String>	T_Identifier
%token			T_MODE
%token			T_LEVEL
%token			T_DELAY
%token			T_EARLY     
%token			T_SIGNAL
%token			T_TRANSN     
%token			T_TRANSP     
%token			T_SLOPE     
%token			T_CAPI      
%token			T_CAPO      
%token			T_IMPI      
%token			T_BUFI      
%token			T_CLOCK

%type	<Chain>		simple_list
%type	<PType>		double_list
%type	<Long>		.index.

%%

file
	: /* empty */
	| line file
	;

line
	: T_MODE	'{' T_Long '}'
	{
	    if ( yyflag & FLAG_MODE )
		yywarning( "redefinition of mode" );
	    if ( $3 < 0 || $3 > 4 )
		yyerror( "mode out of range (0 to 4)" );
	    
	    yyparam->mode = $3;
	    yyflag |= FLAG_MODE;
	}
	| T_LEVEL	'{' T_Long '}'
	{
	    if ( yyflag & FLAG_LEVEL )
		yywarning( "redefinition of level" );
	    if ( $3 < 0 || $3 > 5 )
		yyerror( "level out of range (0 to 5)" );
	    
	    yyparam->level = $3;
	    yyflag |= FLAG_LEVEL;
	}
	| T_DELAY	'{' double_list '}'
	{
	    if ( yyflag & FLAG_DELAY )
		yywarning( "redefinition of delayed inputs" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->delay_pi = $3;
	    yyflag |= FLAG_DELAY;
	}
	| T_EARLY	'{' simple_list '}'
	{
	    if ( yyflag & FLAG_EARLY )
		yywarning( "redefinition of early outputs" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->early_po = $3;
	    yyflag |= FLAG_EARLY;
	}
	| T_SIGNAL	'{' simple_list '}'
	{
	    if ( yyflag & FLAG_SIGNAL )
		yywarning( "redefinition of intermediate signals" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->saved_aux = $3;
	    yyflag |= FLAG_SIGNAL;
	}
	| T_TRANSN	'{' T_Long '}'
	{
	    if ( yyflag & FLAG_TRANSN )
		yywarning( "redefinition of N transistors" );
	    if ( $3 < 0 || $3 > 6 )
	        yyerror( "N transistors out of range ( 0 to 6)" );
	    
	    yyparam->num_trans_n = $3;
	    yyflag |= FLAG_TRANSN;
	}
	| T_TRANSP	'{' T_Long '}'
	{
	    if ( yyflag & FLAG_TRANSP )
		yywarning( "redefinition of P transistors" );
	    if ( $3 < 0 || $3 > 6)
	        yyerror( "P transistors out of range (0 to 6)" );
	    
	    yyparam->num_trans_p = $3;
	    yyflag |= FLAG_TRANSP;
	}
	| T_SLOPE	'{' T_Long '}'
	{
	    if ( yyflag & FLAG_SLOPE )
		yywarning( "redefinition of slope time" );
	    if ( $3 < 0 || $3 > 10000 )
	        yyerror( "slope time out of range (0 to 10000)" );
	    
	    yyparam->max_slope_time = $3;
	    yyflag |= FLAG_SLOPE;
	}
	| T_CAPI	'{' double_list '}'
	{
	    if ( yyflag & FLAG_CAPI )
		yywarning( "redefinition of inputs capacitances" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->capa_pi = $3;
	    yyflag |= FLAG_CAPI;
	}
	| T_CAPO	'{' double_list '}'
	{
	    if ( yyflag & FLAG_CAPO )
		yywarning( "redefinition of outputs capacitances" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->capa_po = $3;
	    yyflag |= FLAG_CAPO;
	}
	| T_IMPI	'{' double_list '}'
	{
	    if ( yyflag & FLAG_IMPI )
		yywarning( "redefinition of imputs impedances" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->imp_pi = $3;
	    yyflag |= FLAG_IMPI;
	}
	| T_BUFI	'{' double_list '}'
	{
	    if ( yyflag & FLAG_BUFI )
		yywarning( "redefinition of buffered inputs" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->buff_pi = $3;
	    yyflag |= FLAG_BUFI;
	}
	| T_CLOCK	'{' simple_list '}'
	{
	    if ( yyflag & FLAG_CLOCK )
		yywarning( "redefinition of clocks" );
	    if ( $3 == 0 )
	        yywarning( "declaration empty" );
	    
	    yyparam->clock = $3;
	    yyflag |= FLAG_CLOCK;
	}
	;

simple_list
	: /* empty */
	{
	    $$ =  0;
	}
	| T_Identifier .index. ';' simple_list
	{
	    if ( $2 < 0 )
	        $$ = addchain( $4, (void *)$1 );
	    else
	        $$ = addchain( $4, (void *)laxnameindex( $1, $2 ) );
	}
	;

double_list
	: /* empty */
	{
	    $$ = 0;
	}
	| T_Identifier .index. ':' T_Long ';' double_list
	{
	    if ( $2 < 0 )
	        $$ = addptype( $6, $4, $1 );
	    else
	    	$$ = addptype( $6, $4, laxnameindex( $1, $2 ) );
	}
	;

.index.
	: /* empty */		{ $$ = -1; }
	| '(' T_Long ')'	{ $$ = $2; }
	;
	
%%

char	*yyfilename;

void yyerror( str )
    char *str;
{
    fflush( stdout );
    fprintf( stdout, "   - %s.lax:%d: error: %s\n", yyfilename, yylineno, str );
    exit(1);
}

void yywarning( str )
    char *str;
{
    fflush( stdout );
    fprintf( stdout, "   - %s.lax:%d: warning: %s\n", yyfilename, yylineno, str );
}

static char *laxnameindex( Name, Index )
    char *Name;
    int Index;
{
    char Buf[BUFSIZ];

    sprintf( Buf, "%s %d", Name, Index );
    return namealloc( Buf );
}

lax_param_t *laxloadparam( Filename )
    char *Filename;
{
    int Res;
    
    printf( " * Loading parameter file '%s.lax' ...\n", Filename );

    if ( ( yyin = mbkfopen( Filename, "lax", "r" ) ) == NULL )
	bflsyserror( 1, "can't open file %s.lax", Filename );

    yyparam = laxnewparam();
    yyfilename = Filename;
    yyflag = 0;

    Res = yyparse();
    fclose( yyin );
    if ( Res == 0 )
        return yyparam;
    else
    	return 0;
}

