/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * Purpose : system dependant functions
 * Date    : 06/03/92
 * Author  : Frederic Petrot <Frederic.Petrot@lip6.fr>
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1997,98
 */

#ident  "$Id: mbk_sys.c,v 1.20 2000/01/17 16:10:47 gregoire Exp $";

#include <stdio.h>
#include <unistd.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <string.h>
#include <fcntl.h>

#include MUT_H
#include "mbk_sys.h"

static char filename[BUFSIZ];
char MBKFOPEN_NAME[BUFSIZ];
char MBKFOPEN_FILTER;
endchld_list *HEAD_END_CHLD=NULL;

long mbkalloc_stat = 0; /* statistics on maximun allocated memory */

void *mbkalloc(nbytes)
unsigned int nbytes;
{
void 	*pt;

	mbkalloc_stat += nbytes;
	if (!(pt = malloc(nbytes))) {
		(void)fflush(stdout);
  		(void)fprintf(stderr,"*** mbk error ***\n");
  		(void)fprintf(stderr,"fatal mbkalloc error : not enough memory\n");
  		EXIT(1);
  	}
	return pt;
}

void *mbkrealloc(pt, nbytes)
void 	*pt;
unsigned int nbytes;
{

	mbkalloc_stat += nbytes;
	if (!(pt = realloc(pt, nbytes))) {
		(void)fflush(stdout);
  		(void)fprintf(stderr,"*** mbk error ***\n");
  		(void)fprintf(stderr,"fatal mbkrealloc error : not enough memory\n");
  		EXIT(1);
  	}
	return pt;
}

void mbkfree(ptr)
void *ptr;
{
#ifdef lint
	(void)free((char *)ptr);
#else
	(void)free(ptr); /* lint makes a mistake here : expects (char *) */
#endif
}

/* file opening :
   defines the strategy used for searching and opening file in the
   mbk environement. */
FILE* mbkfopen( name, extension, mode )
char *name;
char *extension;
char *mode;
{
  FILE *ptf;

  ptf = mbkfopentrace( name, extension, mode );
  if(( TRACE_FILE == 1 ) || (( TRACE_FILE == 3 ) && (ptf != NULL )))
  {
    fflush( stdout );
    fprintf( stderr,
             "*** mbkfopen( \"%s\", \"%s\", \"%s\" ) -> %s\n",
             name,
             extension ? extension : "",
             mode,
             ptf ? MBKFOPEN_NAME : "FAILED"
           );
  }
  return( ptf );
}

void mbkfopen_infos( in, filename, name, extension, mode, filter ) 
FILE *in;
char *filename;
char *name;
char *extension;
char *mode;
char  filter;
{
  if( in )
  {
    strcpy( MBKFOPEN_NAME, filename );
    MBKFOPEN_FILTER = filter;
  }

  if( TRACE_FILE == 2 )
  {
    /* La methode de Czo : y parait qu'on peut debugger avec ca... */
    fprintf( stdout,
             "--- mbk --- mbkfopen %s : %s, (%s,%s,%s)\n",
             in ? "ok" : "FAILED",
             filename,
             name,
             extension ? extension : "",
             mode
           );
  }
}

FILE *mbkfopentrace(name, extension, mode)
char *name, *extension, *mode;
{
FILE *in=NULL;
FILE *insav;
int i;


if (!CATA_LIB || !WORK_LIB)
    mbkenv(); /* not done yet */


  if( *mode == 'w' )
  {
    if( OUT_FILTER && FILTER_SFX )
    {
      if (extension)
        (void)sprintf( filename,
                       "%s/%s.%s%s",
                       WORK_LIB,
                       name,
                       extension,
                       FILTER_SFX
                     );
      else
        (void)sprintf(filename, "%s/%s%s", WORK_LIB, name, FILTER_SFX );

      strcpy( MBKFOPEN_NAME, filename );
      in = mbkpopen( filename, OUT_FILTER, 'w' );
      mbkfopen_infos( in, filename, name, extension, mode, YES );
      if (extension)
        (void)sprintf( filename,
                       "%s/%s.%s",
                       WORK_LIB,
                       name,
                       extension
                     );
      else
        (void)sprintf(filename, "%s/%s", WORK_LIB, name );
       if((insav = fopen(filename,READ_TEXT)) != NULL)
         {
          fflush( stdout );
          fprintf( stderr,
             "*** mbk warning : file %s already exist in the drive directory\n", filename) ;
          fprintf( stderr, "it will be deleted\n" ) ;
          fflush( stderr );
          unlink(filename) ;
          fclose(insav) ;
         }
      return ( in );
    }
    else
    {
      if (extension)
        (void)sprintf( filename,
                       "%s/%s.%s",
                       WORK_LIB,
                       name,
                       extension
                     );
      else
        (void)sprintf(filename, "%s/%s", WORK_LIB, name );
      
      strcpy( MBKFOPEN_NAME, filename );
      in = fopen( filename, WRITE_TEXT ) ;
      mbkfopen_infos( in, filename, name, extension, mode, NO );
      return ( in );
    }
  }
  else
  if( *mode == 'r' )
  {
    if (extension) /* if extension is null, no dot is required */
      (void)sprintf( filename,
                     "%s/%s.%s",
                     WORK_LIB,
                     name,
                     extension
                   );
    else
      (void)sprintf(filename, "%s/%s", WORK_LIB, name);
  
    in = fopen( filename, READ_TEXT );
    mbkfopen_infos( in, filename, name, extension, mode, NO );
    if( in )
    {
      if( FILTER_SFX && IN_FILTER )
      {
        if (extension) /* if extension is null, no dot is required */
          (void)sprintf( filename,
                         "%s/%s.%s%s",
                         WORK_LIB,
                         name,
                         extension,
                         FILTER_SFX
                       );
        else
          (void)sprintf(filename, "%s/%s%s", WORK_LIB, name , FILTER_SFX);
        if((insav = fopen(filename,READ_TEXT)) != NULL)
         {
            fflush( stdout );
            fprintf( stderr,
               "*** mbk warning : file %s also exist it will be ignored\n", filename) ;
            fflush( stderr );
            fclose(insav) ;
         }
      }
      return in;
    }

    if( FILTER_SFX && IN_FILTER )
    {
      if (extension) /* if extension is null, no dot is required */
        (void)sprintf( filename,
                       "%s/%s.%s%s",
                       WORK_LIB,
                       name,
                       extension,
                       FILTER_SFX
                     );
      else
        (void)sprintf(filename, "%s/%s%s", WORK_LIB, name, FILTER_SFX );
  
      in = mbkpopen( filename, IN_FILTER, 'r' );
      mbkfopen_infos( in, filename, name, extension, mode, YES );
      if( in )
        return in ;
    }
  
    i=0;

    while (CATA_LIB[i])
    {
      if (extension)
        (void)sprintf( filename,
                       "%s/%s.%s",
                       CATA_LIB[i],
                       name,
                       extension
                     );
      else
        (void)sprintf( filename,
                       "%s/%s",
                       CATA_LIB[i],
                       name
                     );
  
      in = fopen( filename, READ_TEXT );
      mbkfopen_infos( in, filename, name, extension, mode, NO );
      if( in )
      {
        if( FILTER_SFX && IN_FILTER )
        {
          if (extension) /* if extension is null, no dot is required */
            (void)sprintf( filename,
                           "%s/%s.%s%s",
                           CATA_LIB[i],
                           name,
                           extension ,
                           FILTER_SFX
                         );
          else
            (void)sprintf(filename, "%s/%s%s", CATA_LIB[i], name ,FILTER_SFX);
          if((insav = fopen(filename,READ_TEXT)) != NULL)
           {
              fflush( stdout );
              fprintf( stderr,
               "*** mbk warning : file %s also exist it will be ignored\n", filename) ;
              fflush( stderr );
              fclose(insav) ;
             }
        }
        return( in );
      }

      if( FILTER_SFX && IN_FILTER )
      {
        if (extension) /* if extension is null, no dot is required */
          (void)sprintf( filename,
                         "%s/%s.%s%s",
                         CATA_LIB[i],
                         name,
                         extension,
                         FILTER_SFX
                       );
        else
          (void)sprintf(filename, "%s/%s%s", CATA_LIB[i], name, FILTER_SFX );
    
        in = mbkpopen( filename, IN_FILTER, 'r' );
        mbkfopen_infos( in, filename, name, extension, mode, YES );
        if( in )
         {
          return in ;
         }
      }

      i++;
    } 
    return( NULL );
  }
  else
  {
    (void)fflush(stdout);
    (void)fprintf(stderr, "*** mbk error ***\n");
    (void)fprintf(stderr, "unknown file opening mode %s\n", mode);
    EXIT(1);
  }

  return NULL; /* never reached */
}

/* unlink :
   ensures that only files in the working library may be erased. */
int mbkunlink(name, extension)
char *name, *extension;
{
	if (!CATA_LIB || !WORK_LIB)
		mbkenv(); /* not done yet */

	if (extension) /* if extension is null, no dot is required */
		(void)sprintf(filename, "%s/%s.%s", WORK_LIB, name, extension);
	else
		(void)sprintf(filename, "%s/%s", WORK_LIB, name);

	return unlink(filename);
}
	
/* filepath :
   find the complete path of file from mbkfopen point of view. */
char *filepath(name, extension)
char *name, *extension;
{
FILE *in;
int i ;

  if (!CATA_LIB || !WORK_LIB)
    mbkenv(); /* not done yet */

  if (extension) /* if extension is null, no dot is required */
    (void)sprintf(filename, "%s/%s.%s", WORK_LIB, name, extension );
  else
    (void)sprintf(filename, "%s/%s", WORK_LIB, name );
  
  if ((in = fopen(filename, READ_TEXT))) {
    (void)fclose(in);
    return filename;
  }

  if( FILTER_SFX && IN_FILTER )
  {
    if (extension) /* if extension is null, no dot is required */
      (void)sprintf(filename, "%s/%s.%s%s", WORK_LIB, name, extension, FILTER_SFX );
    else
      (void)sprintf(filename, "%s/%s%s", WORK_LIB, name, FILTER_SFX );
    
    if ((in = fopen(filename, READ_TEXT))) {
      (void)fclose(in);
      return filename;
    }
  }


  i = 0;
  while (CATA_LIB[i]) {
    if (extension) /* if extension is null, no dot is required */
      (void)sprintf(filename, "%s/%s.%s", CATA_LIB[i], name, extension );
    else
      (void)sprintf(filename, "%s/%s", CATA_LIB[i], name );
    
    if ((in = fopen(filename, READ_TEXT))) {
      (void)fclose(in);
      return filename;
    }
  
    if( FILTER_SFX && IN_FILTER )
    {
      if (extension) /* if extension is null, no dot is required */
        (void)sprintf(filename, "%s/%s.%s%s", CATA_LIB[i], name, extension, FILTER_SFX );
      else
        (void)sprintf(filename, "%s/%s%s", CATA_LIB[i], name, FILTER_SFX );
      
      if ((in = fopen(filename, READ_TEXT))) {
        (void)fclose(in);
        return filename;
      }
    }
    i++;
  }
  return NULL;
}

#if 0
/* process state :
   this gives some information on allocated memory, time spend, and so on. */
void mbkps()
{
#include <sys/time.h>
#include <sys/resource.h>
struct rusage rusage;
static int pagesize;
extern long mbkalloc_stat;
static int times;

	if (!times)
		pagesize = getpagesize() / 1024;
	(void)getrusage(RUSAGE_SELF, &rusage);
	(void)fprintf(stdout, "mbk stats          : %d call\n", ++times);
	(void)fprintf(stdout, "user     time      : %ld sec\n",
									rusage.ru_utime.tv_sec);
	(void)fprintf(stdout, "system   time      : %d sec\n",
									rusage.ru_stime.tv_usec);
	(void)fprintf(stdout, "real     size      : %ld k\n",
									rusage.ru_maxrss * pagesize);
	(void)fprintf(stdout, "mbkalloc peak size : %ld k\n",
									mbkalloc_stat / 1024);
	(void)fprintf(stdout, "mallinfo size      : %ld k\n",
									mallinfo().arena / 1024);
}
#endif

/* mbkpopen : open a pipe with a filter, return a FILE*
 * Gregoire.Avot
 */

FILE* mbkpopen( char *nom, char *filter, char mode )
{
  int ptf[2];
  static int fic;
  int pid;
  FILE *file;
  char **argv;

  if( mode == 'w' )
    fic = open( nom,
                O_WRONLY|O_CREAT|O_TRUNC,
                S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH
              );
  else
    fic = open( nom, O_RDONLY );

  if( fic == -1 )
    return( NULL );

  if( pipe( ptf ) == -1 )
  {
    fflush( stdout );
    fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" )\n", nom );
    perror( "pipe() ");
    EXIT(1);
  }
  
  pid = vfork();
  if( pid == -1 )
  {
    fflush( stdout );
    fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" )\n", nom );
    perror( "vfork() ");
    EXIT(1);
  }
  
  if( pid == 0 )
  {
    /* fils */
    fflush( stdout );

    argv = decompfilter( filter );
    if( argv == NULL )
    {
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
      fprintf( stderr, "Bad filter.\n" );
      EXIT(1);
    }

    if( mode == 'w' )
    {
      if( dup2( ptf[0], 0 ) == -1 )
      {
        fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
        perror( "dup2( ptf[0], 0 ) ");
        EXIT(1);
      }
          
      if( dup2( fic, 1 ) == -1 )
      {
        fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
        perror( "dup2( fic, 1 ) ");
        EXIT(1);
      }
    }
    else
    {
      if( dup2( ptf[1], 1 ) == -1 )
      {
        fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
        perror( "dup2( ptf[1], 1 ) ");
        EXIT(1);
      }
          
      if( dup2( fic, 0 ) == -1 )
      {
        fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
        perror( "dup2( fic, 0 ) ");
        EXIT(1);
      }
    }

    if( close( fic ) == -1 )
    {
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
      perror( "close( fic ) ");
      EXIT(1);
    }
    
    if( close( ptf[1] ) == -1 )
    {
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
      perror( "close( ptf[1] ) ");
      EXIT(1);
    }

    if( close( ptf[0] ) == -1 )
    {
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
      perror( "close( ptf[0] ) ");
      EXIT(1);
    }
    
    if( execvp( argv[0], &(argv[1]) ) == -1 )
    {
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) child\n", nom );
      perror( "execvp " );
      EXIT(1);
    }
  }

  close( fic );

  if( mode == 'w' )
  {  
    if( close( ptf[0] ) == -1 )
    {
      fflush( stdout );
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) parent\n", nom );
      perror( "close( ptf[0] ) ");
      EXIT(1);
    }
    
    file = fdopen( ptf[1], "w" );
    if( file == NULL )
    {
      fflush( stdout );
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) parent\n", nom );
      perror( "fdopen( ptf[1], \"w\" ) ");
      EXIT(1);
    }
  }
  else
  {  
    if( close( ptf[1] ) == -1 )
    {
      fflush( stdout );
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) parent\n", nom );
      perror( "close( ptf[1] ) ");
      EXIT(1);
    }
    
    file = fdopen( ptf[0], "r" );
    if( file == NULL )
    {
      fflush( stdout );
      fprintf( stderr, "*** Fatal error in mbkpopen( \"%s\" ) parent\n", nom );
      perror( "fdopen( ptf[0], \"r\" ) ");
      EXIT(1);
    }
  }

  return( file );
}

char **decompfilter( char *filter )
{
  int         pos;
  int         i;
  char       *pt;
  char       *fullpath;
  char       *argv0;
  char       *argvn;
  chain_list *head;
  chain_list *scan;
  int         nbarg;
  char      **ret;

  /* remove trailing space */
  for( pos = 0 ; filter[pos] == ' ' && filter[pos] != 0  ; pos++ );
  if( filter[pos] == 0 )
    return( NULL );

  /* The full path to binairie */
  for( i = pos ; filter[i] != ' ' && filter[i] != 0 ; i++ );
  fullpath = (char*)mbkalloc( sizeof(char) * ( i-pos+1 ) );
  strncpy( fullpath, filter+pos, i-pos );
  fullpath[i-pos] = 0;
  pos = i;

  /* The binairie himself */
  pt = strrchr( fullpath, '/' );
  argv0 = (char*)mbkalloc( sizeof( char ) * (strlen( fullpath ) + 1) );
  if( pt == NULL )
    strcpy( argv0, fullpath );
  else
    strcpy( argv0, pt+sizeof(char) );

  /* The arguments */
  head = NULL;
  nbarg=2;

  do
  {
    for( ; filter[pos] == ' ' && filter[pos] != 0  ; pos++ );
    if( filter[pos] == 0 )
      break;

    for( i = pos ; filter[i] != ' ' && filter[i] != 0 ; i++ );
    argvn = (char*)mbkalloc( sizeof( char ) * ( i-pos+1 ) );
    strncpy( argvn, filter+pos, i-pos );
    argvn[i-pos] = 0;
    pos = i;
    head = addchain( head, argvn );
    nbarg++;
  }
  while( 1 );

  head = reverse( head );

  ret = (char**)mbkalloc( sizeof(char*) * (nbarg+1) );

  ret[0] = fullpath;
  ret[1] = argv0;

  for( scan = head, pos=2 ; scan ; scan = scan->NEXT, pos++ )
    ret[pos] = scan->DATA;
  ret[pos] = NULL;

  freechain( head );

  return( ret );
}

void mbkackchld( int sig )
{
  int           pid;
  int           status;
  endchld_list *new;
  sigset_t ens;
  struct sigaction sgct;
  

  if( sig != SIGCHLD )
  {
    fflush( stdout );
    fprintf( stderr,
             "*** mbk warning : call of mbkackchld with signal %d\n",
             (int)SIGCHLD
           );
    return;
  }
  
  pid           = wait( &status );
  new           = (endchld_list*)mbkalloc( sizeof( endchld_list ));
  new->NEXT     = HEAD_END_CHLD;
  HEAD_END_CHLD = new;
  new->PID      = pid;
  new->STATUS   = status;
 
  #ifdef SunOS 
    signal( SIGCHLD, mbkackchld );
  #else
    sigemptyset( &ens );
    sgct.sa_handler  = mbkackchld;
    sgct.sa_mask     = ens;
    sgct.sa_flags    = SA_RESTART;
    sigaction( SIGCHLD, &sgct , NULL );
  #endif
}
