/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ident  "$Id: rcn_lo_util.c,v 1.3 1999/08/08 20:52:46 gregoire Exp $";

/*******************************************************************************
* rcn    : acces functions to logical rc network structures                    *
*                                                                              *
* version : 1.02                                                               *
* date   : 07/06/96                                                            *
*******************************************************************************/

/*******************************************************************************
* Grgoire AVOT                                                                *
* Laboratoire LIP6 - Dpartement ASIM                                          *
* Universite Paris 6 - Pierre et Marie Curie                                   *
* 4, Place Jussieu                                                             *
* Paris Cedex 05.                                                              *
*******************************************************************************/

#include MUT_H
#include MLO_H
#include RCN_H
#include "rcn_lo.h"

typedef struct s_todo
{
  struct s_todo       *next;
  lonode_list         *ptnode;
  chain_list          *head_wire;
} todolist ;

todolist*        breakloop2 __P(( losig_list*,
                                  lonode_list*,
                                  lowire_list*,
                                  todolist*,
                                  chain_list**
                               ));

void connexe_rec __P(( losig_list*,
                       lowire_list*,
                       lonode_list*,
                       int
                    ));

/******************************************************************************/

void		duplorcnet( ptsig_dest, ptsig_src )
losig_list	*ptsig_dest;
losig_list	*ptsig_src;
{
  lowire_list	*scanlowire;
  lowire_list	*newwire;

  if( ptsig_dest->PRCN )
  {
    fflush( stdout );
    fprintf( stderr,
       "*** MBK error in duplorcnet *** : ptsig_dest have already a rc datas.\n"
           );
    EXIT(1);
  }
  
  if( !ptsig_src->PRCN )
    return;

  addlorcnet( ptsig_dest );
  
  /* Capacit du signal */
  ptsig_dest->PRCN->CAPA = ptsig_src->PRCN->CAPA;
  
  /* Wire constituant le signal */
  for( scanlowire = ptsig_src->PRCN->PWIRE ;
       scanlowire ;
       scanlowire = scanlowire->NEXT
     )
  {

    if( RCN_GETFLAG( scanlowire->FLAG, RCN_FLAG_DEL ) )
      continue;
    newwire = heaplowire();
    
    newwire->NEXT  = ptsig_dest->PRCN->PWIRE;
    newwire->RESI  = scanlowire->RESI;
    newwire->CAPA  = scanlowire->CAPA;
    newwire->X     = scanlowire->X;
    newwire->Y     = scanlowire->Y;
    newwire->DX    = scanlowire->DX;
    newwire->DY    = scanlowire->DY;
    newwire->NODE1 = scanlowire->NODE1;
    newwire->NODE2 = scanlowire->NODE2;
    newwire->LAYER = scanlowire->LAYER;
    newwire->FLAG  = scanlowire->FLAG;
    newwire->USER  = NULL;
      
    ptsig_dest->PRCN->PWIRE = newwire;
  }

  /* CTC sur le signal : non gere */
  ptsig_dest->PRCN->PCTC = NULL;

  /* NBNODE */
  ptsig_dest->PRCN->NBNODE = ptsig_src->PRCN->NBNODE ;

  /* USER */
  ptsig_dest->PRCN->USER = NULL;
}

/******************************************************************************/

todolist*        breakloop2( ptsig, ptnode, curwire, headtodo, cleanwire )
losig_list      *ptsig;
lonode_list     *ptnode;
lowire_list     *curwire;
todolist        *headtodo;
chain_list      **cleanwire;
{
  lowire_list   *tmpwire;
  chain_list    *scanchain;
  todolist      *newtodo;

  if( curwire )
  {
    if( RCN_GETFLAG( curwire->FLAG, RCN_FLAG_PASS ) )
    {
      fflush( stdout );
      fprintf( stderr, "*** rcn error ***\n" );
      fprintf( stderr, "Internal error in function breakloop2().\n" );
      EXIT(1);
    }

    RCN_SETFLAG( curwire->FLAG, RCN_FLAG_PASS );
  }
  
  do
  {
    /* On trouve une boucle */
    if( RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_PASS ) )
    {
      tmpwire = addlowire( ptsig,
                           curwire->LAYER,
                           curwire->FLAG,
                           curwire->RESI,
                           curwire->CAPA,
                           curwire->X,
                           curwire->Y,
                           curwire->DX,
                           curwire->DY,
                           curwire->NODE1,
                           addlonode( ptsig, NULL )
                         );
      *cleanwire = addchain( *cleanwire, curwire );
      RCN_SETFLAG( tmpwire->FLAG, RCN_FLAG_PASS );
      RCN_SETFLAG( curwire->FLAG, RCN_FLAG_PASS );
    }
    else
    {
      RCN_SETFLAG( ptnode->FLAG, RCN_FLAG_PASS );
    }
   
    /* Recherche un wire non parcouru sur le noeud courant */
    for( scanchain = ptnode->WIRELIST ;
         scanchain ;
         scanchain = scanchain->NEXT
       )
    {
      curwire = (lowire_list*)(scanchain->DATA);
      if( ! RCN_GETFLAG( curwire->FLAG, RCN_FLAG_PASS ) )
        break;
    }

    if( ! scanchain )
      break; /* on quitte le while */
    
    /* Ajoute les autre wire non parcouru sur le noeud dans une
     * structure todo */
    
    newtodo = NULL;

    if( scanchain )
    {
      for( scanchain = scanchain->NEXT ;
           scanchain ;
           scanchain = scanchain->NEXT
         )
      {
        tmpwire = (lowire_list*)(scanchain->DATA);
  
        if( ! RCN_GETFLAG( tmpwire->FLAG, RCN_FLAG_PASS ) )
        {
          if( !newtodo )
          {
            newtodo            = (todolist*)(mbkalloc( sizeof(todolist) ) );
            newtodo->next      = headtodo ;
            headtodo           = newtodo;
            newtodo->ptnode    = ptnode;
            newtodo->head_wire = NULL;
          }
          newtodo->head_wire = addchain( newtodo->head_wire, tmpwire );
        }
      }
    }

    ptnode = getlonode( ptsig,
                        ( curwire->NODE1 == ptnode->INDEX ) ? curwire->NODE2 :
                                                              curwire->NODE1
                      );
    
    RCN_SETFLAG( curwire->FLAG, RCN_FLAG_PASS );
  }
  while(1); /* c'est ben vrai! */

  return( headtodo );
}

int              breakloop( ptsig )
losig_list      *ptsig;
{
  
  todolist              *headtodo ;  
  todolist              *curtodo ;  
  todolist              *scantodo ;  
  todolist              *newtodo ;  
  lowire_list           *firstwire;
  lowire_list           *curwire;
  lonode_list           *ptnode;
  chain_list            *cleanwire;
  chain_list            *scanchain;
  
  if( !ptsig->PRCN || !gettabnode( ptsig->PRCN ) )
  {
    fflush( stdout );
    fprintf( stderr, "*** rcn error *** : breakloop() impossible.\n" );
    fprintf( stderr, "No rc data on signal %ld.\n", ptsig->INDEX );
    EXIT( 1 );
  }
  
  clearallwireflag( ptsig->PRCN, RCN_FLAG_PASS );
  cleanwire = NULL;
  
  /* Boucle pour parcourir les differentes composantes connexes */

  for( firstwire  = ptsig->PRCN->PWIRE ;
       firstwire ;
       firstwire = firstwire->NEXT
     )
  {
    if( RCN_GETFLAG( firstwire->FLAG, RCN_FLAG_DEL ) )
      continue;

    ptnode = getlonode( ptsig, firstwire->NODE1 );
    
    if( ! RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_PASS ) )
    {
      /* Corps d'une des composante connexe du reseau */
     
      headtodo = breakloop2( ptsig, ptnode, NULL, NULL, &cleanwire );

      while( headtodo )
      {
        newtodo = NULL;
        
        for( scantodo = headtodo ; scantodo ; scantodo = scantodo->next )
        {
          for( scanchain = scantodo->head_wire ;
               scanchain ;
               scanchain = scanchain->NEXT
             )
          {
          
            curwire = (lowire_list*)( scanchain->DATA );
          
            if( ! RCN_GETFLAG( curwire->FLAG, RCN_FLAG_PASS ) )
            {
               
              newtodo = breakloop2(  ptsig,
                                     getlonode( ptsig,
                                                ( scantodo->ptnode->INDEX ==
                                                              curwire->NODE1 ) ?
                                                                curwire->NODE2 :
                                                                curwire->NODE1
                                             ),
                                      curwire,
                                      newtodo,
                                     &cleanwire
                                   );
            }
          }
        }
        
        while( headtodo )
        {
          curtodo = headtodo->next;
          freechain( headtodo->head_wire );
          mbkfree( headtodo );
          headtodo = curtodo;
        }

        headtodo = newtodo;
      }
    }
  }

  for( scanchain = cleanwire ; scanchain ; scanchain = scanchain->NEXT )
  {
    curwire = (lowire_list*)( scanchain->DATA );
    dellowire( ptsig, curwire->NODE1, curwire->NODE2 );    
  }

  freechain ( cleanwire );

  return(1);
}

int reduce_rcn( ptsig )
losig_list *ptsig;
{
  lowire_list *scanwire;
  lonode_list *ptnode;
  chain_list  *todo;
  chain_list  *doing;
  chain_list  *scanctc;
  loctc_list  *ptctc;
  float        resi;
  int          node1;
  int          node2;
  lowire_list *wire1;
  lowire_list *wire2;
  int          another;

  if( !gettabnode( ptsig->PRCN ) )
  {
    fflush( stdout );
    fprintf( stderr, "*** Reduce RCN error\n" );
    fprintf( stderr, "Node table not built.\n" );
    EXIT(1);
  }

  do
  {
    another = 0;

    todo = NULL;
    ptsig->PRCN->CAPA = 0.0;
  
    for( scanwire = ptsig->PRCN->PWIRE ; scanwire ; scanwire = scanwire->NEXT )
    {
      if( RCN_GETFLAG( scanwire->FLAG, RCN_FLAG_DEL ) )
        continue;
  
      ptsig->PRCN->CAPA = ptsig->PRCN->CAPA + scanwire->CAPA;
      scanwire->CAPA = 0.0;
      ptnode = getlonode( ptsig, scanwire->NODE1 );
      if(   RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_TWO )   &&
          ! RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_LOCON ) &&
          ! RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_PASS )
        )
      {
        todo = addchain( todo, ptnode );
        RCN_SETFLAG( ptnode->FLAG, RCN_FLAG_PASS );
      }
      ptnode = getlonode( ptsig, scanwire->NODE2 );
      if(   RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_TWO )   &&
          ! RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_LOCON ) &&
          ! RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_PASS )
        )
      {
        todo = addchain( todo, ptnode );
        RCN_SETFLAG( ptnode->FLAG, RCN_FLAG_PASS );
      }
    }
  
    for( scanctc = ptsig->PRCN->PCTC ; scanctc ; scanctc = scanctc->NEXT )
    {
      ptctc = (loctc_list*)( scanctc->DATA );
      ptsig->PRCN->CAPA = ptsig->PRCN->CAPA + ptctc->CAPA;
    }
  
    for( doing = todo ; doing ; doing = doing->NEXT )
    {
      ptnode = (lonode_list*)( doing->DATA );
      if( ptnode->CTCLIST )
        continue;
 
      another = 1; 
      wire1 = (lowire_list*)( ptnode->WIRELIST->DATA );
      wire2 = (lowire_list*)( ptnode->WIRELIST->NEXT->DATA );
  
      if( ptnode->WIRELIST->NEXT->NEXT )
      {
        fprintf( stderr, "Anomalie.\n" );
      }
  
      resi = wire1->RESI + wire2->RESI ;
      node1 = ( wire1->NODE1 == ptnode->INDEX ) ? wire1->NODE2 : wire1->NODE1 ;
      node2 = ( wire2->NODE1 == ptnode->INDEX ) ? wire2->NODE2 : wire2->NODE1 ;
     
      setdellowire( ptsig, wire1 ); 
      setdellowire( ptsig, wire2 ); 
      wire1 = NULL;
      wire2 = NULL;
      
      scanwire = getlowire( ptsig, node1, node2 );
      if( scanwire )
        addpararesiwire( scanwire, resi );
      else
        addlowire( ptsig,
                   RCN_WIRE_UNKNOW,
                   0,
                   resi,
                   0.0,
                   0l,
                   0l,
                   0l,
                   0l,
                   node1,
                   node2
                 );
    }
    freechain( todo );
  
    mergedellowire( ptsig );
  }
  while( another );

  return(1);
}

void connexe_rec( losig_list *ptsig,
                  lowire_list *from,
                  lonode_list *ptnode,
                  int index
                )
{
  ptype_list  *ptl;
  chain_list  *scanwire;
  lonode_list *next;

  /* Le noeud par lequel on arrive est deja numerote */

  /* Propagation sur les branches droites */
  while( RCN_GETFLAG( ptnode->FLAG, RCN_FLAG_TWO ) )
  {
    if( (lowire_list*)(ptnode->WIRELIST->DATA) == from )
      from = (lowire_list*)(ptnode->WIRELIST->NEXT->DATA);
    else
      from = (lowire_list*)(ptnode->WIRELIST->DATA);
    
    if( from->NODE1 == ptnode->INDEX )
      ptnode = getlonode( ptsig, from->NODE2 );
    else
      ptnode = getlonode( ptsig, from->NODE1 );

    ptl = getptype( ptnode->USER, CONNEXE );
    if( (int)(ptl->DATA) != 0 )
    {
      if( (int)(ptl->DATA) != index )
      {
        fflush( stdout );
        fprintf( stderr, "RCN ERROR in connexe()\n" );
        fprintf( stderr, "index differs (1).\n" );
        exit( 1 );
      }
      else
        return;
    }
    else
      ptl->DATA = (void*)(index);
  }

  /* On arrive soit sur un noeud terminal, soit sur un embranchement */
  for( scanwire = ptnode->WIRELIST ; scanwire ; scanwire = scanwire->NEXT )
  {
    from = ( lowire_list* )( scanwire->DATA );
    if( from->NODE1 == ptnode->INDEX )
      next = getlonode( ptsig, from->NODE2 );
    else
      next = getlonode( ptsig, from->NODE1 );

    ptl = getptype( next->USER, CONNEXE );
    if( (int)(ptl->DATA) == 0 )
    {
      ptl->DATA = (void*)(index);
      connexe_rec( ptsig, from, next, index ); 
    }
    else
    {
      if( (int)(ptl->DATA) != index )
      {
        fflush( stdout );
        fprintf( stderr, "RCN ERROR in connexe()\n" );
        fprintf( stderr, "index differs (2).\n" );
        exit( 1 );
      }
    }
  }

  return;
}

int connexe( losig_list *ptsig )
{
  lowire_list     *scanlowire;
  ptype_list      *ptl;
  lonode_list     *ptnode;
  int              index;

  if( !gettabnode( ptsig->PRCN ) )
  {
    fflush( stdout );
    fprintf( stderr, "*** Reduce RCN error\n" );
    fprintf( stderr, "Node table not built.\n" );
    EXIT(1);
  }

  for( scanlowire = ptsig->PRCN->PWIRE ;
       scanlowire ;
       scanlowire = scanlowire->NEXT
     )
  {
    ptnode = getlonode( ptsig, scanlowire->NODE1 );
    ptl = getptype( ptnode->USER, CONNEXE );
    if( !ptl )
      ptnode->USER = addptype( ptnode->USER, CONNEXE, (void*)0 );
    else
      ptl->DATA = (void*)0;

    ptnode = getlonode( ptsig, scanlowire->NODE2 );
    ptl = getptype( ptnode->USER, CONNEXE );
    if( !ptl )
      ptnode->USER = addptype( ptnode->USER, CONNEXE, (void*)0 );
    else
      ptl->DATA = (void*)0;
  }

  index = 0;

  for( scanlowire = ptsig->PRCN->PWIRE ;
       scanlowire ;
       scanlowire = scanlowire->NEXT
     )
  {
    ptnode = getlonode( ptsig, scanlowire->NODE1 );
    ptl = getptype( ptnode->USER, CONNEXE );
    
    if( (int)(ptl->DATA) != 0 )
      continue;
 
    index++;

    ptl->DATA = (void*)index;
    connexe_rec( ptsig, scanlowire, ptnode, index );

    ptnode = getlonode( ptsig, scanlowire->NODE2 );
    ptl = getptype( ptnode->USER, CONNEXE );
    ptl->DATA = (void*)index;
    connexe_rec( ptsig, scanlowire, ptnode, index );

  }

  return( index );
}
