/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - CMOS Cells Compiler
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident  "$Id: c4_lofig.c,v 1.1.1.1 1998/10/02 15:25:07 alliance Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MLO_H
#include MLU_H
#include MPH_H
#include MGN_H

void
lofigNot (name, lc_io, out, in)
     char *name;
     chain_list *lc_io;
     char *out;
     char *in;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 10, 1, in, "vdd", out);
  LOTRS (TRANSN, 5, 1, in, "vss", out);
  SAVE_LOFIG ();
}

void
lofigNotP (name, lc_io, out, in)
     char *name;
     chain_list *lc_io;
     char *out;
     char *in;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 20, 1, in, out, "vdd");
  LOTRS (TRANSN, 10, 1, in, out, "vss");
  SAVE_LOFIG ();
}


void
lofigBuf (name, lc_io, out, in)
     char *name;
     chain_list *lc_io;
     char *out;
     char *in;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 13, 1, in, "mbk_sig2", "vdd");
  LOTRS (TRANSP, 20, 1, "mbk_sig2", out, "vdd");
  LOTRS (TRANSP, 20, 1, "mbk_sig2", "vdd", out);
  LOTRS (TRANSN, 8, 1, in, "mbk_sig2", "vss");
  LOTRS (TRANSN, 10, 1, "mbk_sig2", out, "vss");
  LOTRS (TRANSN, 10, 1, "mbk_sig2", "vss", out);
  SAVE_LOFIG ();
}

void
lofigHigh (name, lc_io, out)
     char *name;
     chain_list *lc_io;
     char *out;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 13, 1, "vss", "vdd", out);
  SAVE_LOFIG ();
}

void
lofigLow (name, lc_io, out)
     char *name;
     chain_list *lc_io;
     char *out;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSN, 10, 1, "vdd", "vss", out);
  SAVE_LOFIG ();
}

void
lofigBus (name, lc_io, out, i1, i2)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 20, 1, "mbk_sig3", "vdd", out);
  LOTRS (TRANSP, 20, 1, "mbk_sig3", out, "vdd");
  LOTRS (TRANSP, 12, 1, i1, "mbk_sig3", "vdd");
  LOTRS (TRANSP, 12, 1, i2, "vdd", "mbk_sig3");
  LOTRS (TRANSP, 12, 1, i2, "mbk_sig2", "vdd");
  LOTRS (TRANSP, 12, 1, "mbk_sig2", "mbk_sig1", "mbk_sig3");
  LOTRS (TRANSN, 6, 1, i1, "mbk_sig1", "vss");
  LOTRS (TRANSN, 6, 1, i2, "mbk_sig1", "mbk_sig3");
  LOTRS (TRANSN, 6, 1, i2, "mbk_sig2", "vss");
  LOTRS (TRANSN, 6, 1, "mbk_sig2", "vss", "mbk_sig1");
  LOTRS (TRANSN, 10, 1, "mbk_sig1", "vss", out);
  LOTRS (TRANSN, 10, 1, "mbk_sig1", out, "vss");
  SAVE_LOFIG ();
}

void
lofigTie (name, lc_io)
     char *name;
     chain_list *lc_io;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  SAVE_LOFIG ();
}

	/* LATCH1 & LATCH1N : i1 = data   i2 = commande
	   i3 = point memorisant       */

void
lofigLatch1 (name, lc_io, out, i1, i2, i3)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2, *i3;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 1, 3, "mbk_sig5", "vdd", i3);
  LOTRS (TRANSP, 4, 1, i3, "vdd", "mbk_sig5");
  LOTRS (TRANSP, 20, 1, "mbk_sig5", out, "vdd");
  LOTRS (TRANSN, 2, 6, "mbk_sig5", "vss", i3);
  LOTRS (TRANSN, 5, 1, i2, i1, i3);
  LOTRS (TRANSN, 4, 1, i3, "vss", "mbk_sig5");
  LOTRS (TRANSN, 10, 1, "mbk_sig5", out, "vss");
  SAVE_LOFIG ();
}

void
lofigLatch1N (name, lc_io, out, i1, i2, i3)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2, *i3;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 1, 3, "mbk_sig1", "vdd", i3);
  LOTRS (TRANSP, 2, 1, i3, "vdd", "mbk_sig1");
  LOTRS (TRANSP, 10, 1, i3, out, "vdd");
  LOTRS (TRANSN, 1, 6, "mbk_sig1", "vss", i3);
  LOTRS (TRANSN, 5, 1, i2, i1, i3);
  LOTRS (TRANSN, 2, 1, i3, "vss", "mbk_sig1");
  LOTRS (TRANSN, 9, 1, i3, out, "vss");
  SAVE_LOFIG ();
}

	/* LATCH2 & LATCH2N : i1 = data   i2 = commande
	   i3 = data   i4 = commande
	   i5 = point memorisant    */

void
lofigLatch2 (name, lc_io, out, i1, i2, i3, i4, i5)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2, *i3, *i4, *i5;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 1, 2, "mbk_sig5", "vdd", i5);
  LOTRS (TRANSP, 4, 1, i5, "vdd", "mbk_sig5");
  LOTRS (TRANSP, 20, 1, "mbk_sig5", out, "vdd");
  LOTRS (TRANSN, 1, 6, "mbk_sig5", "vss", i5);
  LOTRS (TRANSN, 5, 1, i4, i5, i3);
  LOTRS (TRANSN, 5, 1, i2, i1, i5);
  LOTRS (TRANSN, 4, 1, i5, "vss", "mbk_sig5");
  LOTRS (TRANSN, 10, 1, "mbk_sig5", out, "vss");
  SAVE_LOFIG ();
}

void
lofigLatch2N (name, lc_io, out, i1, i2, i3, i4, i5)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2, *i3, *i4, *i5;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 20, 1, i5, out, "vdd");
  LOTRS (TRANSP, 2, 1, i5, "vdd", "mbk_sig2");
  LOTRS (TRANSP, 1, 3, "mbk_sig2", "vdd", i5);
  LOTRS (TRANSN, 20, 1, i5, out, "vss");
  LOTRS (TRANSN, 2, 1, i5, "vss", "mbk_sig2");
  LOTRS (TRANSN, 5, 1, i2, i1, i5);
  LOTRS (TRANSN, 5, 1, i4, i5, i3);
  LOTRS (TRANSN, 1, 6, "mbk_sig2", "vss", i5);
  SAVE_LOFIG ();
}

	/* BD1 i1 = data   i2 = commande
	   i3 = 1er point memorisant  (master)
	   i4 = 2eme point memorisant (slave)      */

void
lofigBd1 (name, lc_io, out, i1, i2, i3, i4)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2, *i3, *i4;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      printf ("con %s\n", (char *) lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  printf ("i1=%s, i2=%s, i3=%s, i4=%s\n", i1, i2, i3, i4);
  LOTRS (TRANSP, 10, 1, "mbk_sig1", "vdd", "mbk_sig8");
  LOTRS (TRANSP, 4, 1, i3, "mbk_sig1", "vdd");
  LOTRS (TRANSP, 15, 1, i2, "vdd", "mbk_sig10");
  LOTRS (TRANSP, 1, 3, "mbk_sig11", "vdd", i4);
  LOTRS (TRANSP, 4, 1, i4, "mbk_sig11", "vdd");
  LOTRS (TRANSP, 20, 1, i4, "vdd", out);
  LOTRS (TRANSP, 1, 3, "mbk_sig1", "vdd", i3);
  LOTRS (TRANSP, 10, 1, i1, "vdd", "mbk_sig4");
  LOTRS (TRANSN, 2, 6, "mbk_sig1", i3, "vss");
  LOTRS (TRANSN, 5, 1, "mbk_sig1", "vss", "mbk_sig8");
  LOTRS (TRANSN, 5, 1, "mbk_sig10", "mbk_sig8", i4);
  LOTRS (TRANSN, 2, 6, "mbk_sig11", i4, "vss");
  LOTRS (TRANSN, 4, 1, i3, "mbk_sig1", "vss");
  LOTRS (TRANSN, 5, 1, i2, "mbk_sig4", i3);
  LOTRS (TRANSN, 8, 1, i2, "vss", "mbk_sig10");
  LOTRS (TRANSN, 4, 1, i4, "mbk_sig11", "vss");
  LOTRS (TRANSN, 10, 1, i4, "vss", out);
  LOTRS (TRANSN, 5, 1, i1, "vss", "mbk_sig4");
  SAVE_LOFIG ();
}

	/* BDN  i1 = data   i2 = commande
	   i3 = 1er point memorisant  (master)
	   i4 = 2eme point memorisant (slave)      */
void
lofigBd1n (name, lc_io, out, i1, i2, i3, i4)
     char *name;
     chain_list *lc_io;
     char *out;
     char *i1, *i2, *i3, *i4;
{
  DEF_LOFIG (name);
  while (lc_io)
    {
      LOCON (lc_io->DATA, UNKNOWN, lc_io->DATA);
      lc_io = lc_io->NEXT;
    }
  LOTRS (TRANSP, 10, 1, "mbk_sig1", "vdd", "mbk_sig8");
  LOTRS (TRANSP, 4, 1, i3, "mbk_sig1", "vdd");
  LOTRS (TRANSP, 15, 1, i2, "vdd", "mbk_sig10");
  LOTRS (TRANSP, 1, 3, "mbk_sig11", "vdd", i4);
  LOTRS (TRANSP, 4, 1, i4, "mbk_sig11", "vdd");
  LOTRS (TRANSP, 20, 1, "mbk_sig11", "vdd", out);
  LOTRS (TRANSP, 1, 3, "mbk_sig1", "vdd", i3);
  LOTRS (TRANSP, 10, 1, i1, "vdd", "mbk_sig4");
  LOTRS (TRANSN, 2, 6, "mbk_sig1", i3, "vss");
  LOTRS (TRANSN, 5, 1, "mbk_sig1", "vss", "mbk_sig8");
  LOTRS (TRANSN, 5, 1, "mbk_sig10", "mbk_sig8", i4);
  LOTRS (TRANSN, 2, 6, "mbk_sig11", i4, "vss");
  LOTRS (TRANSN, 4, 1, i3, "mbk_sig1", "vss");
  LOTRS (TRANSN, 5, 1, i2, "mbk_sig4", i3);
  LOTRS (TRANSN, 8, 1, i2, "vss", "mbk_sig10");
  LOTRS (TRANSN, 4, 1, i4, "mbk_sig11", "vss");
  LOTRS (TRANSN, 10, 1, "mbk_sig11", "vss", out);
  LOTRS (TRANSN, 5, 1, i1, "vss", "mbk_sig4");
  SAVE_LOFIG ();
}
