/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: sl_dontcare.c,v 1.1.1.1 1998/10/02 15:25:11 alliance Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H

	/* fonction de simplification par DC */

/*------------------------------------------------------------------------------
depthNodeBdd : calcul la profondeur d'un BDD
------------------------------------------------------
parametres 	 : un pNode
-------------------------------------------------------
return 		 : int
------------------------------------------------------------------------------*/
int
depthNodeBdd (pt)
     pNode pt;
{
  int depthlow, depthhigh;

  if (pt->index < 2)
    return 0;
  depthlow = depthNodeBdd (pt->low);
  depthhigh = depthNodeBdd (pt->high);
  if (depthlow > depthhigh)
    return 1 + depthlow;
  else
    return 1 + depthhigh;
}

/*------------------------------------------------------------------------------
costDepthTab : calcul le cout en nombre de litteraux 
-------------------------------------------------------
parametres 	 : une table de pNode
-------------------------------------------------------
return 		 : int
------------------------------------------------------------------------------*/
int
costDepthTab (tab, len)
     pNode *tab;
     int len;
{
  int i;
  int count = 0;
  int depth;

  markAllBdd (0);

  for (i = 0; i < len; i++)
    {
      depth = depthNodeBdd (tab[i]);
      if (count < depth)
	count = depth;
    }
  markAllBdd (0);
  return (count);
}


/*------------------------------------------------------------------------------
costTab : calcul le cout en nombre de litteraux 
-------------------------------------------------------
parametres 	 : une table de pNode
-------------------------------------------------------
return 		 : int
------------------------------------------------------------------------------*/
int
costTab (tab, len, func)
     pNode *tab;
     int len;
     int (*func) ();
{
  int i;
  int count = 0;

  markAllBdd (0);

  for (i = 0; i < len; i++)
    {
      count = count + (*func) (tab[i]);
    }
  markAllBdd (0);
  return (count);
}

/*------------------------------------------------------------------------------
sigmaDFBeh     : calcul des DC provenant des entrees 
-------------------------------------------------------
parametres 	 : une befig
-------------------------------------------------------
return 		 : void
------------------------------------------------------------------------------*/
pNode
sigmaDFBeh (support, Fon, beh, dejaTraite)
     chain_list *support;
     pNode Fon;
     befig_list *beh;
     pTH dejaTraite;
{
  pNode ret = zero;
  chain_list *supp = support;
  beaux_list *aux;
  pNode obs;

  /* parcours du support */

  while (supp)
    {
      pNode deja = (pNode) searchTH (dejaTraite, (char *) supp->DATA);


      if (deja != (pNode) EMPTYTH)	/* noeud deja traite */
	{
	  aux = beh->BEAUX;
	  while (aux && strcmp (aux->NAME, supp->DATA))
	    {
	      aux = aux->NEXT;
	    }

	  /* calcul de la fonction d'observabilite */

	  obs = applyBinBdd (XOR,
			     constraintBdd (Fon, (pNode) aux->NODE),
			   constraintBdd (Fon, notBdd ((pNode) aux->NODE)));

	  /* ajout a la fonction de retour */

	  ret = applyBinBdd (OR, ret, applyBinBdd (AND, deja, obs));
	}
      else
	{
	  pNode reportDC;

	  aux = beh->BEAUX;
	  while (aux && strcmp (aux->NAME, supp->DATA))
	    {
	      aux = aux->NEXT;
	    }

	  if (aux)
	    {

	      if (searchExpr (aux->ABL, "'d'"))
		{
		  chain_list *expr;

		  if (!aux->NODE)
		    aux->NODE = ablToBddCct (beh->CIRCUI, aux->ABL);

		  expr = substExpr (aux->ABL, "'d'", createAtom ("'1'"));

		  deja = applyBinBdd (AND, notBdd (aux->NODE),
				      ablToBddCct (beh->CIRCUI, expr));
		  freeExpr (expr);
		}
	      else
		deja = zero;

/*
   printf("DC local : ");
   displayExpr(bddToAblCct(beh->CIRCUI,deja));
 */

	      /* report des DC provenant du support */

	      reportDC = sigmaDFBeh (supportChain_listExpr (aux->ABL), deja,
				     beh, dejaTraite);
	      if (reportDC != zero)
		deja = applyBinBdd (OR, deja, reportDC);

/*
   printf("DC en fn des entrees : ");
   displayExpr(bddToAblCct(beh->CIRCUI,deja));
 */

	      /* calcul de la fonction d'observabilite */

	      obs = applyBinBdd (XOR,
				 constraintBdd (Fon, (pNode) aux->NODE),
			   constraintBdd (Fon, notBdd ((pNode) aux->NODE)));

	      /* ajout a la fonction de retour */

	      ret = applyBinBdd (OR, ret,
				 applyBinBdd (AND, deja, obs));

/*
   printf("Observabilite : ");
   displayExpr(bddToAblCct(beh->CIRCUI,ret));
 */
	      addTH (dejaTraite, (char *)supp->DATA, (int)deja);
	    }
	}

/*
   printf("NAME = %s\n",supp->DATA);
   displayExpr(bddToAblCct(beh->CIRCUI,ret));
 */

      supp = supp->NEXT;
    }


  freechain (support);
  return ret;
}


/*------------------------------------------------------------------------------
CalculDCBeh      : calcule les DC d'une befig 
-------------------------------------------------------
parametres 	 : une befig, un tableau de on, de dc et un compteur
-------------------------------------------------------
return 		 : void
------------------------------------------------------------------------------*/
void
calculDCBeh (beh, pON, pDC)
     befig_list *beh;
     pNode *pON, *pDC;
{
  beout_list *out;
  beaux_list *aux;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  binode_list *binode;
  int i = 0;

  /* ajout dans la couche circuit des BDD des auxiliaires */

  aux = beh->BEAUX;
  while (aux)
    {
      if (aux->NODE)
	addOutputCct (beh->CIRCUI, aux->NAME, aux->NODE);
      aux = aux->NEXT;
    }

  /* traitement des sorties primaires */

  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	{
	  pON[i] = out->NODE;

	  /* recherche des DC locaux */

	  if (searchExpr (out->ABL, "'d'"))
	    {
	      chain_list *expr;

	      expr = substExpr (out->ABL, "'d'", createAtom ("'1'"));

	      pDC[i] = applyBinBdd (AND, notBdd (pON[i]),
				    ablToBddCct (beh->CIRCUI, expr));
	      freeExpr (expr);
	    }
	  else
	    pDC[i] = zero;

	  /* report des DC provenant du support */
/*
   reportDC = sigmaDFBeh(supportChain_listExpr(out->ABL),out->NODE,
   beh,dejaTraite);
   if (reportDC != zero)
   pDC[i] = applyBinBdd(OR,pDC[i],reportDC);
 */

	  i++;
	}
      out = out->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      binode = reg->BINODE;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      pON[i] = binode->VALNODE;

	      /* recherche des DC locaux */

	      if (searchExpr (biabl->VALABL, "'d'"))
		{
		  chain_list *expr;

		  expr = substExpr (biabl->VALABL, "'d'", createAtom ("'1'"));

		  pDC[i] = applyBinBdd (AND, notBdd (pON[i]),
					ablToBddCct (beh->CIRCUI, expr));
		  freeExpr (expr);
		}
	      else
		pDC[i] = zero;

	      /* report des DC provenant du support */
/*
   reportDC = sigmaDFBeh(supportChain_listExpr(biabl->VALABL),
   binode->VALNODE,beh,dejaTraite);

   if (reportDC != zero)
   pDC[i] = applyBinBdd(OR,pDC[i],reportDC);
 */

	      i++;
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      binode = bus->BINODE;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      pON[i] = binode->VALNODE;

	      /* recherche des DC locaux */

	      if (searchExpr (biabl->VALABL, "'d'"))
		{
		  chain_list *expr;

		  expr = substExpr (biabl->VALABL, "'d'", createAtom ("'1'"));

		  pDC[i] = applyBinBdd (AND, notBdd (pON[i]),
					ablToBddCct (beh->CIRCUI, expr));
		  freeExpr (expr);
		}
	      else
		pDC[i] = zero;

	      /* report des DC provenant du support */

/*
   reportDC = sigmaDFBeh(supportChain_listExpr(biabl->VALABL),
   binode->VALNODE,beh,dejaTraite);

   if (reportDC != zero)
   pDC[i] = applyBinBdd(OR,pDC[i],reportDC);
 */

	      i++;
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      binode = bux->BINODE;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      pON[i] = binode->VALNODE;

	      /* recherche des DC locaux */

	      if (searchExpr (biabl->VALABL, "'d'"))
		{
		  chain_list *expr;

		  expr = substExpr (biabl->VALABL, "'d'", createAtom ("'1'"));

		  pDC[i] = applyBinBdd (AND, notBdd (pON[i]),
					ablToBddCct (beh->CIRCUI, expr));
		  freeExpr (expr);
		}
	      else
		pDC[i] = zero;

	      /* report des DC provenant du support */

/*
   reportDC = sigmaDFBeh(supportChain_listExpr(biabl->VALABL),
   binode->VALNODE,beh,dejaTraite);

   if (reportDC != zero)
   pDC[i] = applyBinBdd(OR,pDC[i],reportDC);
 */

	      i++;
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      bux = bux->NEXT;
    }
}


/*------------------------------------------------------------------------------
numberBddBeh     : calcule le nombre de BDD (out + reg + bux + bus)
-------------------------------------------------------
parametres 	 : une befig
-------------------------------------------------------
return 		 : int
------------------------------------------------------------------------------*/
int
numberBddBeh (beh)
     befig_list *beh;
{
  beout_list *out;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  int numberGraph = 0;

  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	{
	  numberGraph++;
	}
      out = out->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      numberGraph++;
	    }
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      numberGraph++;
	    }
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      numberGraph++;
	    }
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }
  return numberGraph;
}

/*------------------------------------------------------------------------------
traitementDCBeh     : procedure de prise en compte des Don't care
-------------------------------------------------------
parametres 	 : une befig
-------------------------------------------------------
return 		 : void
------------------------------------------------------------------------------*/
void
traitementDCBeh (beh, trace, func, level)
     befig_list *beh;
     int trace;
     int (*func) ();
     int level;
{
  beout_list *out;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  binode_list *binode;
  pNode *pDC;
  pNode *pON;
  int numberGraph = 0;
  int i, j;
  pNode simplifPlusDcZeroBdd ();
  pNode restrictBdd ();
  pNode cnstBdd ();
  pNode cliqueBdd ();
  pNode jacoBdd ();
  pTH dejaTraite;
  pElemTH pEl;
  int coutInitial;
  int coutFinal;
  int boucle;

  int DCFUN;

  /* dejaTraite sert a la re-ecriture des expressions */

  dejaTraite = createTH (100);

  /* calcul du nombre de graphes */

  printf ("Running Don't Care Minimization\n");

  numberGraph = numberBddBeh (beh);;

  pDC = (pNode *) mbkalloc (numberGraph * sizeof (pNode));
  pON = (pNode *) mbkalloc (numberGraph * sizeof (pNode));

  /* mise a jour des tables */


  if (trace)
    printf ("Computing the Don't Care of each signal...\n");


  calculDCBeh (beh, pON, pDC);

  coutInitial = costTab (pON, numberGraph, func);

  if (trace)
    printf ("depth max = %d init = %d\n\n", costDepthTab (pON, numberGraph), coutInitial);

  if (trace)
    printf ("Applying the Don't Care simplification function...\n");


  DCFUN = 2;

  if (1)
    {
      boucle = level;

      for (j = 1; j <= boucle; j++)
	{
	  for (i = 0; i < numberGraph; i++)
	    {
	      /* le i-eme graphe possede un DC */

	      if (pDC[i] != zero)
		{
		  pNode pInter;
		  pInter = pON[i];

		  /* simplification de pON[i] */

		  if (DCFUN == 0)
		    pON[i] = restrictBdd (pON[i], notBdd (pDC[i]));
		  if (DCFUN == 1)
		    pON[i] = simplifPlusDcZeroBdd (pON[i], pDC[i]);
		  if (DCFUN == 2)
		    pON[i] = simplifDcZeroBdd (pON[i], pDC[i]);
		  if (DCFUN == 3)
		    pON[i] = cnstBdd (pON[i], notBdd (pDC[i]));
		  if (DCFUN == 4)
		    pON[i] = simplifDcOneBdd (applyBinBdd (OR, pDC[i], pON[i]), pDC[i]);

		  /* minimisation successive */

		  if (level == 2)
		    {
		      coutFinal = costTab (pON, numberGraph, func);


		      if (coutInitial < coutFinal)	/* pas d'amelioration */
			{
			  pON[i] = pInter;
			}
		      else
			/* on enleve au DC la partie utilisee dans ON */
			{
			  if (trace)
			    {
			      printf ("cost(%s) %d --> ", beh->NAME, coutInitial);
			      printf ("%d\n", coutFinal);
			    }
			  pDC[i] = applyBinBdd (AND, pDC[i], notBdd (pON[i]));
			  coutInitial = coutFinal;
			}
		    }
		}
	    }
	}
    }

  if (0 || (level == 1))
    {
      coutFinal = costTab (pON, numberGraph, func);
      if (trace)
	{
	  printf ("cost(%s) %d --> ", beh->NAME, coutInitial);
	  printf ("%d\n", coutFinal);
	  printf ("depth max = %d\n\n", costDepthTab (pON, numberGraph));
	}
    }
  /* regeneration des equations simplifiees */

  /* on vide la table des sorties de la couches circuit */

  pEl = ((beh->CIRCUI)->pTO)->pElem;

  for (i = 0; i < ((beh->CIRCUI)->pTO)->length; i++)
    {
      pEl->value = EMPTYTH;
      pEl++;
    }

  i = 0;

  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	{
	  if (out->NODE != pON[i])
	    {
	      out->NODE = pON[i];
	      freeExpr (out->ABL);
	      out->ABL = bddToAblCct (beh->CIRCUI, pON[i]);
	    }
	  i++;
	}
      out = out->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      binode = reg->BINODE;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      if (binode->VALNODE != pON[i])
		{
		  binode->VALNODE = pON[i];
		  freeExpr (biabl->VALABL);
		  biabl->VALABL = bddToAblCct (beh->CIRCUI, pON[i]);
		}
	      i++;
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      binode = bus->BINODE;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      if (binode->VALNODE != pON[i])
		{
		  binode->VALNODE = pON[i];
		  freeExpr (biabl->VALABL);
		  biabl->VALABL = bddToAblCct (beh->CIRCUI, pON[i]);
		}
	      i++;
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      binode = bux->BINODE;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      if (binode->VALNODE != pON[i])
		{
		  binode->VALNODE = pON[i];
		  freeExpr (biabl->VALABL);
		  biabl->VALABL = bddToAblCct (beh->CIRCUI, pON[i]);
		}
	      i++;
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      bux = bux->NEXT;
    }


  mbkfree (pON);
  mbkfree (pDC);
  destroyTH (dejaTraite);
}
