/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: sl_system.c,v 1.1.1.1 1998/10/02 15:25:12 alliance Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"


/*----------------------------------------------------------------------------
verifArgs	: verifie les arguments de la ligne de commande.
------------------------------------------------------------------------------
retour		: un entier a 1 si OK, 1 sinon.
------------------------------------------------------------------------------*/

int
verifArgs (sl, argc, argv)
     sl_system *sl;
     int argc;
     char *argv[];
{
  /* logic file_input file_output */

  if (argc == 3)
    {
      if (argv[1][0] == '-')
	{
	  if (strcmp (argv[1], "-o") && strcmp (argv[1], "-l") &&
	      strcmp (argv[1], "-s") && strcmp (argv[1], "-c"))
	    {
	      printf ("\nError : bad input file name '%s'\n\n", argv[1]);
	    }
	  else
	    printf ("\nError :  missed output file name\n\n");
	  return (0);
	}
      argv[1] = (char *) genNAME ("%s.vbe", argv[1]);
      return (1);
    }

  /* avec option */

  if (argc > 1 && argv[1][0] == '-')
    {
      if (strcmp (argv[1], "-o") && strcmp (argv[1], "-l") && strcmp (argv[1], "-s") &&
	  strcmp (argv[1], "-c"))
	{
	  printf ("\nError : bad option '%s'\n\n", argv[1]);
	  return (0);
	}
    }

  /* interactif avec un fichier de parametres */

  if (argc == 4 && strcmp (argv[1], "-o") && strcmp (argv[1], "-l") &&
      strcmp (argv[1], "-s") && strcmp (argv[1], "-c"))
    {
      argv[1] = (char *) genNAME ("%s.vbe", argv[1]);
      argv[3] = (char *) genNAME ("%s.lax", argv[3]);
      return (1);
    }				/* fin appel version interactive */


  if (argc < 4 || argc > 5)
  {
      printf ("\nBad command line\n");
      printf ("Command line args :  <input_file> <output_file> [<param_file>]\n");
      printf ("Please, refer to manual.\n");
      return (0);
  }
  
  argv[2] = (char *) genNAME ("%s.vbe", argv[2]);

  /* avec fichier de parametres */

  if (argc > 4)
    argv[4] = (char *) genNAME ("%s.lax", argv[4]);

  return (1);
}


/*----------------------------------------------------------------------------
majSystem	: met a jour les champs du systeme 
------------------------------------------------------------------------------
retour		: rien 
-----------------------------------------------------------------------------*/
void
majSystem (sl)
     sl_system *sl;
{
  sl->file_lib = WORK_LIB;

  if (!existDirectory (sl->file_lib))
    {
      printf ("\nError : MBK_WORK_LIB -> the directory '%s' doesn't exist\n",
	      sl->file_lib);
      exit (-1);
    }

  if (!getenv ("MBK_TARGET_LIB"))
    strcpy (sl->cata_lib, ".");
  else
    sl->cata_lib = getenv ("MBK_TARGET_LIB");

  if (!existDirectory (sl->cata_lib))
    {
      printf ("\nError : MBK_TARGET_LIB -> the directory '%s' doesn't exist\n",
	      sl->cata_lib);
      exit (-1);
    }

  if (!getenv ("MBK_CATA_LIB"))
    strcpy (sl->cata_save, ".");
  else
    sl->cata_save = getenv ("MBK_CATA_LIB");

  if (!getenv ("MBK_C4_LIB"))
    strcpy (sl->c4_lib, ".");
  else
    sl->c4_lib = getenv ("MBK_C4_LIB");

  sl->in_lo = IN_LO;
  sl->out_lo = OUT_LO;

  sl->out_ph = OUT_PH;

  if (getenv ("MBK_TRACE") == NULL || !strcmp (getenv ("MBK_TRACE"), "0"))
    sl->trace = 0;
  else if (!strcmp (getenv ("MBK_TRACE"), "1"))
    sl->trace = 1;
  else if (!strcmp (getenv ("MBK_TRACE"), "2"))
    sl->trace = 2;
  else if (!strcmp (getenv ("MBK_TRACE"), "3"))
    sl->trace = 3;
  else
    sl->trace = 0;


/* appel aux fonctions d'ordonnancement - 0 rien - 1 ABL - 2 ABL + BDD */
    sl->makeOrder = 2;

    if (sl->makeOrder >= 1)
	sl->order = 1;
/* on cree des VI lors de l'optimisation */
    sl->auxOF = 1;

}

/*----------------------------------------------------------------------------
addInfoSystem	: ajoute les infos de process et de fichiers a traiter
------------------------------------------------------------------------------
retour		: rien. 
------------------------------------------------------------------------------*/

void
addInfoSystem (sl, mode, argc, argv, version)
     sl_system *sl;
     int mode;
     int argc;
     char *argv[];
     char *version;
{
  char *carint;

  strcpy (sl->version, version);
  sl->mode = mode;

  /* version batch */

  if (sl->mode == 1)
    {
      /* on recupere le nom du fichier d'entree */

      strcpy (sl->file, argv[2]);
      if (!strcmp (sl->file_lib, "."))
	strcpy (sl->fileVhdl, argv[2]);
      else
	{
	  strcpy (sl->fileVhdl, carint = (char *) genNAME ("%s/%s", sl->file_lib, argv[2]));
	  mbkfree (carint);
	}

      /* on recupere le nom du fichier de sortie */

      strcpy (sl->fileOut, argv[3]);

      /* parametres */

      sl->par = initializeParam (2, 0, 2, NULL, NULL, NULL, 4, 4);
      if (argc == 5)
	{
	  strcpy (sl->filePar, argv[4]);
	  if (!strcmp (sl->file_lib, "."))
	    strcpy (sl->fileParam, argv[4]);
	  else
	    {
	      strcpy (sl->fileParam, carint = (char *) genNAME ("%s/%s", sl->file_lib, argv[4]));
	      mbkfree (carint);
	    }
	  sl->par = loadParam (sl->fileParam);
	  if (sl->par == NULL)
	    {
	      exit (-1);
	    }
	}
      else
	{
	  if (!strcmp (sl->file_lib, "."))
	    strcpy (sl->fileParam, "default.lax");
	  else
	    {
	      strcpy (sl->fileParam, carint = (char *) genNAME ("%s/default.lax", sl->file_lib));
	      mbkfree (carint);
	    }
	}
      sl->option = *(argv[1] + 1);
    }
  else
    /* mode interactif 0 ou 2 */
    {

      /* on recupere le nom du fichier d'entree */

      strcpy (sl->file, argv[1]);
      if (!strcmp (sl->file_lib, "."))
	strcpy (sl->fileVhdl, argv[1]);
      else
	{
	  strcpy (sl->fileVhdl, carint = (char *) genNAME ("%s/%s", sl->file_lib, argv[1]));
	  mbkfree (carint);
	}
      /* on recupere le nom du fichier de sortie */

      strcpy (sl->fileOut, argv[2]);

      sl->par = initializeParam (2, 0, 2, NULL, NULL, NULL, 4, 4);
      if (sl->mode == 0)
	{
	  strcpy (sl->filePar, "default.lax");
	  if (!strcmp (sl->file_lib, "."))
	    strcpy (sl->fileParam, "default.lax");
	  else
	    {
	      strcpy (sl->fileParam, carint = (char *) genNAME ("%s/default.lax", sl->file_lib));
	      mbkfree (carint);
	    }
	}
      else
	/* mode 2 */
	{
	  strcpy (sl->filePar, argv[3]);
	  if (!strcmp (sl->file_lib, "."))
	    strcpy (sl->fileParam, argv[3]);
	  else
	    {
	      strcpy (sl->fileParam, carint = (char *) genNAME ("%s/%s", sl->file_lib, argv[3]));
	      mbkfree (carint);
	    }
	  sl->par = loadParam (sl->fileParam);
	  if (sl->par == NULL)
	    {
	      exit (-1);
	    }
	}
      sl->option = '\0';
    }
  /* mise a jour des variables de MBK */

  put_CATA_LIB ((char *) genNAME ("%s:%s", SL_CATA_LIB, SL_CCCC_LIB));
  strcpy (IN_LO, OUT_LO);
}

/*----------------------------------------------------------------------------
displaySystem   : affiche les arguments de la synthese 
------------------------------------------------------------------------------
retour		: void 
------------------------------------------------------------------------------*/

void
displaySystem (sys)
     sl_system *sys;
{
  if (SL_MODE == 1)		/* mode batch */
    {
      printf ("================================  Environment  ================================\n");
      printf ("MBK_WORK_LIB        = %s\n", SL_FILE_LIB);
      printf ("MBK_CATA_LIB        = %s\n", SL_CATA_SAVE);
      if (SL_OPTION == 's' || SL_OPTION == 'c')
	{
	  if (SL_OPTION == 's')
	    {
	      printf ("MBK_TARGET_LIB      = %s\n", SL_CATA_LIB);
	      printf ("MBK_IN_LO           = %s\n", SL_IN_LO);
	    }
	  else
	    {
	      printf ("MBK_C4_LIB          = %s\n", SL_CCCC_LIB);
	      printf ("MBK_OUT_PH          = %s\n", SL_OUT_PH);
	    }
	  printf ("MBK_OUT_LO          = %s\n", SL_OUT_LO);
	}
      if (SL_TRACE > 0)
	printf ("MBK_TRACE           = %d\n", SL_TRACE);
      printf ("=======================  Files, Options and Parameters  =======================\n");
      printf ("VHDL file           = %s\n", SL_FILE);
      if (SL_OPTION == 'o' || SL_OPTION == 'l')
	printf ("output file         = %s.vbe\n", SL_FILEOUT);
      else
	printf ("output file         = %s.%s\n", SL_FILEOUT, SL_OUT_LO);
      printf ("Parameter file      = %s\n", SL_PARAM);
      printf ("Mode                = ");
      switch (SL_OPTION)
	{
	case 'o':
	  printf ("Global optimization\n");
	  break;
	case 'l':
	  printf ("Local optimization\n");
	  break;
	case 's':
	  printf ("Mapping standard cell\n");
	  break;
	case 'c':
	  printf ("Mapping with compiler cells C4\n");
	  break;
	}
      printf ("Optimization mode   = ");
      switch (SL_OPTIM)
	{
	case 0:
	  printf ("100%% area optimization\n");
	  break;
	case 1:
	  printf ("70%% area - 30%% delay optimization\n");
	  break;
	case 2:
	  printf ("50%% area - 50%% delay optimization\n");
	  break;
	case 3:
	  printf ("30%% area - 70%% delay optimization\n");
	  break;
	case 4:
	  printf ("100%% delay optimization\n");
	  break;
	}
      printf ("Optimization level  = %d\n", SL_LEVEL);
      if (SL_OPTION == 'c')
	{
	  printf ("Number of serial transistors on graph P = %d\n", SL_NUMTRANSP);
	  printf ("Number of serial transistors on graph N = %d\n", SL_NUMTRANSN);
	}
      printf ("===============================================================================\n");
      printf ("\n");
    }
}
