/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Bdd                           |
|                                                             |
| File    :                  bddcheck.c                       |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H

# include <stdio.h>
# include "bddcheck.h"
# include "bdderror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Check Bdd Variable                     |
|                                                             |
\------------------------------------------------------------*/

int checkbddvar( BddSystem, Variable, Severity )

  bddsystem *BddSystem;
  bddvar     Variable;
  int        Severity;
{
  setbddlocalsystem( BddSystem ); 

  if ( Variable >= BddLocalSystem->NUMBER_VAR )
  {
    if ( Severity )
    {
      bdderror( BDD_BAD_VAR_ERROR, Variable );
    }

    return( 0 );
  }

  return( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                       Check Bdd Index                       |
|                                                             |
\------------------------------------------------------------*/

int checkbddindex( BddSystem, Index, Severity )

  bddsystem *BddSystem;
  bddindex   Index;
  int        Severity;
{
  setbddlocalsystem( BddSystem ); 

  if ( ( Index <  BDD_INDEX_MIN                ) ||
       ( Index >= BddLocalSystem->NUMBER_INDEX ) )
  {
    if ( Severity )
    {
      bdderror( BDD_BAD_INDEX_ERROR, Index );
    }

    return( 0 );
  }

  return( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                       Check Bdd Operator                    |
|                                                             |
\------------------------------------------------------------*/

int checkbddoper( BddSystem, Oper, Severity )

  bddsystem *BddSystem;
  long       Oper;
  int        Severity;
{
  setbddlocalsystem( BddSystem ); 

  if ( ( Oper < 0         ) ||
       ( Oper >= ABL_NXOR ) )
  {
    if ( Severity )
    {
      bdderror( BDD_OPERATOR_ERROR, Oper );
    }

    return( 0 );
  }

  return( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                      Check Bdd Assoc                        |
|                                                             |
\------------------------------------------------------------*/

int checkbddassoc( BddSystem, BddAssoc, Severity )

  bddsystem *BddSystem;
  bddassoc  *BddAssoc;
  int        Severity;
{
  setbddlocalsystem( BddSystem );

  if ( BddLocalSystem->ASSOC_RESET )
  {
    resetbddhopertable( BddLocalSystem->HASH_OPER );
  }

  if ( ( BddAssoc->FIRST == BDD_MAX_VAR ) &&
       ( BddAssoc->LAST  == BDD_MAX_VAR ) ) 
  {
    return( 2 );
  }

  if ( BddAssoc->FIRST >= BDD_MAX_VAR )
  {
    if ( Severity )
    {
      bdderror( BDD_BAD_ASSOC_ERROR, BddAssoc->FIRST );
    }

    return( 0 );
  }

  if ( BddAssoc->LAST >= BDD_MAX_VAR )
  {
    if ( Severity )
    {
      bdderror( BDD_BAD_ASSOC_ERROR, BddAssoc->LAST );
    }

    return( 0 );
  }

  return( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                       Check Bdd Max Node                    |
|                                                             |
\------------------------------------------------------------*/

int checkbddmaxnode( BddSystem, Severity )

  bddsystem *BddSystem;
  int        Severity;
{
  long ReorderLimit;
  long FirstNumberNode;

  setbddlocalsystem( BddSystem );

  if ( BddLocalSystem->REORDER )
  {
    if ( ( BddLocalSystem->NUMBER_NODE  > BddLocalSystem->MAX_NODE      ) ||
         ( BddLocalSystem->NUMBER_NODE  > BddLocalSystem->REORDER_LIMIT ) )
    {
      FirstNumberNode = BddLocalSystem->NUMBER_NODE;

      BddLocalSystem->REORDER_FUNC( (bddsystem *)0 );

      bddwarning( BDD_DYNAMIC_REORDER_WARNING, 
                  FirstNumberNode, BddLocalSystem->NUMBER_NODE );

      ReorderLimit = ( BddLocalSystem->NUMBER_NODE   * 
                       BddLocalSystem->REORDER_RATIO ) / 100;

      if ( ReorderLimit < BddLocalSystem->REORDER_LOW )
      {
        BddLocalSystem->REORDER_LIMIT = BddLocalSystem->REORDER_LOW;
      }
      else
      {
        BddLocalSystem->REORDER_LIMIT = ReorderLimit;
      }
    }
  }

  if ( BddLocalSystem->NUMBER_NODE > BddLocalSystem->MAX_NODE )
  {
    if ( Severity )
    {
      bdderror( BDD_MAX_NODE_ERROR, BddLocalSystem->NUMBER_NODE );
    }

    return( 0 );
  }

  return( 1 );
}
