/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Bdd                           |
|                                                             |
| File    :                  bddsatisfy.c                     |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H

# include <stdio.h>
# include "bddsatisfy.h"
# include "bddassoc.h"
# include "bddapply.h"
# include "bdderror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                  Local Satisfy Bdd Node                     |
|                                                             |
\------------------------------------------------------------*/

static bddnode *loc_satisfybdd( BddNode )

  bddnode  *BddNode;
{
  if ( BddNode->INDEX >= BDD_INDEX_MIN )
  {
    if ( BddNode->HIGH == BddLocalSystem->ZERO )
    {
      BddNode = addbddnode( (bddsystem *)0, BddNode->INDEX,
                            BddLocalSystem->ZERO,
                            decbddrefext( loc_satisfybdd( BddNode->LOW ) ) );
    }
    else
    {
      BddNode = addbddnode( (bddsystem *)0, BddNode->INDEX,
                            decbddrefext( loc_satisfybdd( BddNode->HIGH ) ),
                            BddLocalSystem->ZERO );
    }
  }

  return( BddNode );
}

/*------------------------------------------------------------\
|                                                             |
|                    Satisfy Bdd Node                         |
|                                                             |
\------------------------------------------------------------*/

bddnode *satisfybddnode( BddSystem, BddNode )

  bddsystem *BddSystem;
  bddnode   *BddNode;
{
  setbddlocalsystem( BddSystem );

  if ( BddNode->INDEX == BDD_INDEX_ZERO )
  {
    bdderror( BDD_BAD_INDEX_ERROR, BDD_INDEX_ZERO );
  }

  BddNode = loc_satisfybdd( BddNode );

  checkbddmaxnode( (bddsystem *)0, 1 );

  return( BddNode );
}

/*------------------------------------------------------------\
|                                                             |
|                  Local Satisfy Bdd Node Assoc               |
|                                                             |
\------------------------------------------------------------*/

static bddnode *loc_satisfybddassoc( BddNode )

  bddnode  *BddNode;
{
  bddnode *NodeVar;
  bddvar   Variable;

  if ( BddNode->INDEX >= BDD_INDEX_MIN )
  {
    if ( BddNode->HIGH == BddLocalSystem->ZERO )
    {
      BddNode = addbddnode( (bddsystem *)0, BddNode->INDEX,
                            BddLocalSystem->ZERO,
                            decbddrefext( loc_satisfybddassoc( BddNode->LOW ) ) );
    }
    else
    {
      BddNode = addbddnode( (bddsystem *)0, BddNode->INDEX,
                            decbddrefext( loc_satisfybddassoc( BddNode->HIGH ) ),
                            BddLocalSystem->ZERO );
    }

    if ( BddNode->INDEX >= BDD_INDEX_MIN )
    {
      Variable = BddLocalSystem->INDEX_TO_VAR[ BddNode->INDEX ];

      if ( ( Variable >= BddLocalAssoc->FIRST ) &&
           ( Variable <= BddLocalAssoc->LAST  ) )
      {
        NodeVar = BddLocalSystem->VAR_NODE[ BddNode->INDEX ];
        NodeVar->FLAG |= BDD_FLAG_NUM_MASK;
      }
    }
  }

  return( BddNode );
}

/*------------------------------------------------------------\
|                                                             |
|                    Satisfy Bdd Node                         |
|                                                             |
\------------------------------------------------------------*/

bddnode *satisfybddnodeassoc( BddSystem, BddNode, BddAssoc )

  bddsystem *BddSystem;
  bddnode   *BddNode;
  bddassoc  *BddAssoc;
{
  bddnode *BddMore;
  bddnode *NodeVar;
  long     Variable;
  bddindex Index;

  if ( BddNode->INDEX == BDD_INDEX_ZERO )
  {
    bdderror( BDD_BAD_INDEX_ERROR, BDD_INDEX_ZERO );
  }

  if ( checkbddassoc( BddSystem, BddAssoc, 1 ) == 2 )
  {
    return( incbddrefext( BddNode ) );
  }

  BddLocalAssoc     = BddAssoc;
  BddLocalAssocNode = BddAssoc->ASSOC_NODE;

  BddNode = loc_satisfybddassoc( BddNode );
  BddMore = BddLocalSystem->ONE;

  for ( Variable  = (long)BddLocalAssoc->LAST;
        Variable >= (long)BddLocalAssoc->FIRST;
        Variable  = Variable - 1 )
  {
    if ( BddLocalAssocNode[ Variable ] != (bddnode *)0 )
    {
      Index   = BddLocalSystem->VAR_TO_INDEX[ Variable ];
      NodeVar = BddLocalSystem->VAR_NODE[ Index ];

      if ( NodeVar->FLAG & BDD_FLAG_NUM_MASK )
      {
        NodeVar->FLAG &= ~BDD_FLAG_NUM_MASK;
      }
      else
      {
        BddMore = loc_applybdd( ABL_AND, decbddrefext( BddMore ), NodeVar );
      }
    }
  }

  BddNode = loc_applybdd( ABL_AND,
                          decbddrefext( BddNode ),
                          decbddrefext( BddMore ) );
       
  checkbddmaxnode( (bddsystem *)0, 1 );

  return( BddNode );
}
