/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Fsm                           |
|                                                             |
| File    :                  fsmsearch.c                      |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include FSM_H

# include <stdio.h>
# include "fsmsearch.h"
# include "fsmerror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Fsm Search Functions                   |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Fsm Search Figure                    |
|                                                             |
\------------------------------------------------------------*/

fsmfig_list *searchfsmfig( Name )

   char *Name;
{
  fsmfig_list *Figure;

  Name = namealloc( Name );

  for ( Figure  = HEAD_FSMFIG;
        Figure != (fsmfig_list *)0;
        Figure  = Figure->NEXT )
  {
    if ( Figure->NAME == Name ) break;
  }

  return( Figure );
}

/*------------------------------------------------------------\
|                                                             |
|                        Fsm Search State                     |
|                                                             |
\------------------------------------------------------------*/

fsmstate_list *searchfsmstate( Figure, Name )

   fsmfig_list *Figure;
   char        *Name;
{
  authelem  *Element;

  Name    = namealloc( Name );
  Element = searchauthelem( Figure->HASH_STATE, Name );

  if ( Element != (authelem *)0 )
  {
    return( (fsmstate_list *)( Element->VALUE ) );
  }

  return( (fsmstate_list *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                        Fsm Search Output                    |
|                                                             |
\------------------------------------------------------------*/

fsmout_list *searchfsmout( Figure, Name )

   fsmfig_list *Figure;
   char        *Name;
{
  authelem  *Element;

  Name    = namealloc( Name );
  Element = searchauthelem( Figure->HASH_OUT, Name );

  if ( Element != (authelem *)0 )
  {
    return( (fsmout_list *)( Element->VALUE ) );
  }

  return( (fsmout_list *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                        Fsm Search Input                     |
|                                                             |
\------------------------------------------------------------*/

fsmin_list *searchfsmin( Figure, Name )

   fsmfig_list *Figure;
   char        *Name;
{
  authelem  *Element;

  Name    = namealloc( Name );
  Element = searchauthelem( Figure->HASH_IN, Name );

  if ( Element != (authelem *)0 )
  {
    return( (fsmin_list *)( Element->VALUE ) );
  }

  return( (fsmin_list *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                         Fsm Search Port                     |
|                                                             |
\------------------------------------------------------------*/

fsmport_list *searchfsmport( Figure, Name )

   fsmfig_list   *Figure;
   char          *Name;
{
  authelem  *Element;

  Name    = namealloc( Name );
  Element = searchauthelem( Figure->HASH_PORT, Name );

  if ( Element != (authelem *)0 )
  {
    return( (fsmport_list *)( Element->VALUE ) );
  }

  return( (fsmport_list *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                     Fsm Search Transition                   |
|                                                             |
\------------------------------------------------------------*/

fsmtrans_list *searchfsmtrans( Figure, StateFrom, StateTo )

   fsmfig_list   *Figure;
   fsmstate_list *StateFrom;
   fsmstate_list *StateTo;
{
  auth2elem  *Element;

  Element = searchauth2elem( Figure->HASH_TRANS, (char *)StateFrom, (char *)StateTo );

  if ( Element != (auth2elem *)0 )
  {
    return( (fsmtrans_list *)( Element->VALUE ) );
  }

  return( (fsmtrans_list *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                     Fsm Search Stack                        |
|                                                             |
\------------------------------------------------------------*/

fsmstack_list *searchfsmstack( Control, Current, Return )

   char           Control;
   fsmstate_list *Current;
   fsmstate_list *Return;
{
  fsmstack_list *Stack;
  chain_list    *ScanChain;

  for ( ScanChain  = Current->STACK;
        ScanChain != (chain_list *)0;
        ScanChain  = ScanChain->NEXT )
  {
    Stack = (fsmstack_list *)( ScanChain->DATA );

    if ( ( Stack->CTRL   == Control ) &&
         ( Stack->RETURN == Return  ) ) return( Stack );
  }

  return( (fsmstack_list *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                     Fsm Search Local output                 |
|                                                             |
\------------------------------------------------------------*/

fsmlocout_list *searchfsmlocout( State, Out )

   fsmstate_list *State;
   fsmout_list   *Out;
{
  fsmlocout_list *Locout;

  for ( Locout  = State->LOCOUT;
        Locout != (fsmlocout_list *)0;
        Locout  = Locout->NEXT )
  {
    if ( Locout->OUT == Out ) break;
  }

  return( Locout );
}
