/* ###--------------------------------------------------------------### */
/*                                                                      */
/* file     : fbl_drive.c                                               */
/* date     : Nouv 05 1992                                              */
/* version  : v100                                                      */
/* author   : VUONG H.N., M.HANAFI                                      */
/* description  : This file contains VHDL drivers :                     */
/*        vhdlsavefbfig()                                               */
/*                                                                      */
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include MUT_H
#include AUT_H
#include ABL_H
#include BDD_H
#include FVH_H
#include "fbl_utype.h"
#include "fbl_utdef.h"
#include "fbl_drive.h"

/* ###--------------------------------------------------------------### */
/* function : vhdlsavefbfig                     */
/* description  : print out a text file containing a data-flow VHDL */
/*        description                       */
/* called func. : vhdlablname, mbkalloc, fbl_message, fbl_error  , */
/*        fbl_toolbug , getptype, reverse    , fbl_printablfile */
/* ###--------------------------------------------------------------### */

void vhdlsavefbfig (pthedfbfig, FileName, synopsys )

fbfig_list  *pthedfbfig;
char        *FileName;
int          synopsys;
{
  char         *mode;
  char         *type_mark;
  int           nrlabel = 0;
  FILE         *fd;
  time_t        clock;
  int           left,right;
  char         *name;
  char         *bus;

  struct fbgen *ptgeneric = 0;  /* current ptype pnt (generic)  */
  struct ptype *ptptype   = 0;  /* current ptype pointer (USER) */
  struct fbreg *ptfbreg   = 0;  /* current BEREG pointer    */
  struct fbmsg *ptfbmsg   = 0;  /* current BEMSG pointer    */
  struct fbout *ptfbout   = 0;  /* current BEOUT pointer    */
  struct fbbus *ptfbbus   = 0;  /* current BEBUS pointer    */
  struct fbaux *ptfbaux   = 0;  /* current BEAUX pointer    */
  struct fbbux *ptfbbux   = 0;  /* current BEBUX pointer    */
  struct fbpor *ptfbpor   = 0;  /* correctly ordered port list  */
  struct biabl *ptbiabl   = 0;  /* current BIABL pointer    */

  if (pthedfbfig == 0)
    fbl_toolbug (10,"fbl_decomp",0,0);

  fbl_message (11,0,0);

    /* ###------------------------------------------------------### */
    /*    Opening result file                   */
    /* ###------------------------------------------------------### */

  if ((fd = mbkfopen( FileName, ( synopsys ) ? "vhd" : "vbe",WRITE_TEXT)) == 0)
    {
    fbl_error (107,0);
    autexit (1);
    }

  fbl_message (12,pthedfbfig->NAME,0);
  fbl_message (13,pthedfbfig->NAME,0);

  time (&clock);
  (void) fprintf (fd,"-- VHDL data flow description generated from `%s`\n",
                  pthedfbfig->NAME);
  (void) fprintf (fd,"--\t\tdate : %s\n\n",ctime(&clock));

    /* ###------------------------------------------------------### */
    /*    Entity declaration                    */
    /* ###------------------------------------------------------### */

  (void) fprintf (fd,"-- Entity Declaration\n\n");
  (void) fprintf (fd,"ENTITY %s IS\n",vhdlablname(pthedfbfig->NAME));

    /* ###------------------------------------------------------### */
    /*    Generic declaration                   */
    /* ###------------------------------------------------------### */

  ptptype = getptype (pthedfbfig->USER,FBH_GENERIC);
  if (ptptype != 0)
    {
    ptptype->DATA = (void *)reverse (ptptype->DATA);
    ptgeneric     = (struct fbgen *)ptptype->DATA;
    (void) fprintf (fd,"  GENERIC (\n");
    while (ptgeneric != 0)
      {
      (void) fprintf (fd,"    CONSTANT %s : NATURAL := %ld",
                     vhdlablname(ptgeneric->NAME),*((long *)ptgeneric->VALUE));
      if (ptgeneric->NEXT != 0)
        (void)fprintf(fd,";\t-- %s\n",(char *)ptgeneric->NAME);
      else
        (void)fprintf(fd,"\t-- %s\n",(char *)ptgeneric->NAME);

      ptgeneric = ptgeneric->NEXT;
      }
    (void) fprintf (fd,"  );\n");
    ptptype->DATA = (void *)reverse (ptptype->DATA);
    }

    /* ###------------------------------------------------------### */
    /*    Port declaration                      */
    /* ###------------------------------------------------------### */

  ptfbpor = pthedfbfig->BEPOR;
  if (ptfbpor != 0)
    {
    (void) fprintf (fd,"  PORT (\n");
    pthedfbfig->BEPOR = (struct fbpor *)reverse ( (chain_list *)pthedfbfig->BEPOR);

    ptfbpor = pthedfbfig->BEPOR;
    while (ptfbpor != 0)
      {
      switch (ptfbpor->DIRECTION)
        {
        case 'I':
          mode = namealloc("IN"); 
          break;
        case 'O':
        case 'Z':
          mode = namealloc("OUT"); 
          break;
        case 'B':
        case 'T':
          mode = namealloc("INOUT"); 
          break;
        default :
          {
          fbl_error (69,0);
          }
        }
      ptfbpor = (fbpor_list *)fbl_vectnam(ptfbpor,&left,&right,&name,0);

      if(left != -1)
        {
        switch (ptfbpor->TYPE)
          {
            case FBH_TYPE_BIT_VEC :
            type_mark = namealloc("BIT_VECTOR"); 
            bus = "";
            break;
            case FBH_TYPE_WOR_VEC :
            type_mark = namealloc("WOR_VECTOR"); 
            bus = namealloc("BUS");
            break;
            case FBH_TYPE_MUX_VEC :
            type_mark = namealloc("MUX_VECTOR"); 
            bus = namealloc("BUS");
            break;
          default :
            {
            fbl_error (68,0);
            }
          }
        (void)fprintf(fd,"  %s : %s %s(%d %s %d) %s",vhdlablname(name), mode,
                      type_mark, left, (left>=right)?"DOWNTO":"TO",right,bus);
        }
      else
        {
        switch (ptfbpor->TYPE)
          {
            case FBH_TYPE_BIT :
            type_mark = "BIT"; break;
            case FBH_TYPE_WOR_BIT :
            type_mark = "WOR_BIT BUS"; break;
            case FBH_TYPE_MUX_BIT :
            type_mark = "MUX_BIT BUS"; break;
          default :
            {
            fbl_error (68,0);
            }
          }

        (void) fprintf (fd,"  %s : %s %s",vhdlablname(name),
                                          mode,type_mark);
        }

      if (ptfbpor->NEXT != 0)
        (void) fprintf (fd,";\t-- %s\n",name);
      else
        (void) fprintf (fd,"\t-- %s\n  );\n",name);

      ptfbpor = ptfbpor->NEXT;
      }

    pthedfbfig->BEPOR = (struct fbpor *)reverse ( (chain_list *)pthedfbfig->BEPOR);
    }

  (void) fprintf (fd,"END %s;\n\n\n",vhdlablname(pthedfbfig->NAME));

    /* ###------------------------------------------------------### */
    /*    Architecture declaration                  */
    /* ###------------------------------------------------------### */

  (void) fprintf (fd,"-- Architecture Declaration\n\n");
  (void) fprintf (fd,"ARCHITECTURE VBE OF %s IS\n",
                  vhdlablname(pthedfbfig->NAME));

    /* ###------------------------------------------------------### */
        /*  Treatment of the BEREG list                 */
    /* ###------------------------------------------------------### */

  pthedfbfig->BEREG = (struct fbreg *)reverse( (chain_list *)pthedfbfig->BEREG);
  ptfbreg = pthedfbfig->BEREG;

  while (ptfbreg != 0)
  {
      ptfbreg = (fbreg_list *)fbl_vectnam(ptfbreg,&left,&right,&name,2);
      if(left != -1)
      {
        if ( synopsys )
        (void)fprintf(fd,"  SIGNAL %s : BIT_VECTOR(%d %s %d);\t-- %s\n",
                      vhdlablname(name),left,(left>=right)?"DOWNTO":"TO",
                      right,name);
        else
        (void)fprintf(fd,"  SIGNAL %s : REG_VECTOR(%d %s %d) REGISTER;\t-- %s\n",
                      vhdlablname(name),left,(left>=right)?"DOWNTO":"TO",
                      right,name);
      }
      else
      {
        if ( synopsys )
        (void) fprintf (fd,"  SIGNAL %s : BIT;\t-- %s\n",
                      vhdlablname(name),name);
        else
        (void) fprintf (fd,"  SIGNAL %s : REG_BIT REGISTER;\t-- %s\n",
                      vhdlablname(name),name);
      }
      ptfbreg = ptfbreg->NEXT;
  }

  pthedfbfig->BEREG = (struct fbreg *)reverse( (chain_list *)pthedfbfig->BEREG);

    /* ###------------------------------------------------------### */
        /*  Treatment of the BEBUX list                 */
    /* ###------------------------------------------------------### */

  pthedfbfig->BEBUX = (struct fbbux *)reverse( (chain_list *)pthedfbfig->BEBUX);
  ptfbbux = pthedfbfig->BEBUX;
  while (ptfbbux != 0)
    {
    ptfbbux = (fbbux_list *)fbl_vectnam(ptfbbux,&left,&right,&name,1);
    if(left != -1)
      {
      switch (ptfbbux->TYPE)
        {
          case FBH_TYPE_WOR_VEC :
          type_mark = namealloc("WOR_VECTOR"); 
          break;
          case FBH_TYPE_MUX_VEC :
          type_mark = namealloc("MUX_VECTOR"); 
          break;
        }
      (void)fprintf(fd,"  SIGNAL %s : %s(%d %s %d) BUS;\t-- %s\n",
                    vhdlablname(name),type_mark,left,(left>=right)?"DOWNTO":"TO",
                    right,name);
      }
    else
      {
      switch (ptfbbux->TYPE)
        {
          case FBH_TYPE_WOR_BIT :
          type_mark = namealloc("WOR_BIT"); 
          break;
          case FBH_TYPE_MUX_BIT :
          type_mark = namealloc("MUX_BIT"); 
          break;
        }
      (void) fprintf (fd,"  SIGNAL %s : %s BUS;\t\t-- %s\n",vhdlablname(name),
                      type_mark,name);
      }
    ptfbbux = ptfbbux->NEXT;
    }
  pthedfbfig->BEBUX = (struct fbbux *)reverse( (chain_list *)pthedfbfig->BEBUX);

  pthedfbfig->BEAUX = (struct fbaux *)reverse( (chain_list *)pthedfbfig->BEAUX);
  ptfbaux = pthedfbfig->BEAUX;
  while (ptfbaux != 0)
    {
      if ( ( ptfbaux->ABL != (chain_list *)0 ) ||
           ( IsFbhAssignByFsm( ptfbaux )     ) )
      {
    ptfbaux = (fbaux_list *)fbl_vectnam(ptfbaux,&left,&right,&name,3);
    if(left != -1)
      {
      (void)fprintf(fd,"  SIGNAL %s : BIT_VECTOR(%d %s %d);\t-- %s\n",
                    vhdlablname(name),left,(left>=right)?"DOWNTO":"TO",
                    right,name);
      }
    else
      {
      (void) fprintf (fd,"  SIGNAL %s : BIT;\t\t-- %s\n",
                    vhdlablname(name),name);
      }
      }
    ptfbaux = ptfbaux->NEXT;
    }
  pthedfbfig->BEAUX = (struct fbaux *)reverse( (chain_list *)pthedfbfig->BEAUX);

  (void) fprintf (fd,"\nBEGIN\n");

    /* ###------------------------------------------------------### */
    /*    Print out a concurrent assert statement for each BEMSG    */
    /* ###------------------------------------------------------### */

  ptfbmsg = pthedfbfig->BEMSG;
  while (ptfbmsg != 0)
    {
    if (ptfbmsg->LABEL != 0)
      (void)fprintf(fd,"  %s :",ptfbmsg->LABEL);

    (void) fprintf (fd,"  ASSERT (" );
    fbl_printablfile(fd, ptfbmsg->ABL );
    (void) fprintf (fd," = '1')\n" );

    if (ptfbmsg->MESSAGE != 0)
      (void) fprintf (fd,"    REPORT \"%s\"\n",ptfbmsg->MESSAGE);

    if (ptfbmsg->LEVEL == 'W')
      (void) fprintf (fd,"    SEVERITY WARNING;");
    else
      (void) fprintf (fd,"    SEVERITY ERROR;");

    (void) fprintf (fd,"\n\n");
    ptfbmsg = ptfbmsg->NEXT;
    }

    /* ###------------------------------------------------------### */
    /*    Print out a concurrent signal assignment for each BEAUX   */
    /* ###------------------------------------------------------### */

  ptfbaux = pthedfbfig->BEAUX;

  while (ptfbaux != 0)
  {
    if (ptfbaux->ABL != 0)
      {
      (void) fprintf (fd,"  %s <= ",fbl_vectorize(ptfbaux->NAME));
      fbl_printablfile( fd, ptfbaux->ABL);
      (void) fprintf (fd,";\n");
      }
    /* LUDO
    else
    {
      fbl_error (40,ptfbaux->NAME);
    } */
    ptfbaux = ptfbaux->NEXT;
  }

    /* ###------------------------------------------------------### */
    /*    Print out a block statement  with one guarded concurrent  */
    /* signal assignment for each BIABL of each BEREG       */
    /* ###------------------------------------------------------### */

  ptfbreg = pthedfbfig->BEREG;

  while (ptfbreg != 0)
  {
    ptbiabl = ptfbreg->BIABL;
    while (ptbiabl != 0)
      {
      if ( synopsys )
      (void) fprintf (fd,"  proc_%d : PROCESS ",nrlabel);
      else
      (void) fprintf (fd,"  label%d : BLOCK ",nrlabel);
      if (ptbiabl->CNDABL  != 0)
      {
        if ( ! synopsys )
        {
          (void) fprintf (fd,"(" );
          fbl_printablfile( fd, ptbiabl->CNDABL );
          (void) fprintf (fd," = '1')\n" );
        }
      }
      else
        fbl_toolbug (19,"fbl_decomp",ptfbreg->NAME,0);
  
      (void) fprintf (fd,"  BEGIN\n" );

      if ( synopsys )
      {
          (void) fprintf (fd,"   WAIT UNTIL (" );
          fbl_printablfile( fd, ptbiabl->CNDABL );
          (void) fprintf (fd," = '1');\n" );
      }

      (void) fprintf (fd,"    %s <= %s ",
                      fbl_vectorize(ptfbreg->NAME), 
                      ( synopsys ) ? "" : "GUARDED" );
      if (ptbiabl->VALABL  != 0)
        {
        fbl_printablfile( fd, ptbiabl->VALABL );
        (void)fprintf (fd,";\n" );
        }
      else
        fbl_toolbug (20,"fbl_decomp",ptfbreg->NAME,0);

      if ( synopsys )
      (void) fprintf( fd,"  END PROCESS proc_%d;\n",nrlabel);
      else
      (void) fprintf( fd,"  END BLOCK label%d;\n",nrlabel);
      ptbiabl = ptbiabl->NEXT;
      nrlabel++;
      }
  
      ptfbreg = ptfbreg->NEXT;
  }

    /* ###------------------------------------------------------### */
    /*    Print out a block statement  with one guarded concurrent  */
    /* signal assignment for each BIABL of each BEBUX       */
    /* ###------------------------------------------------------### */

  ptfbbux = pthedfbfig->BEBUX;
  while (ptfbbux != 0)
  {
    ptbiabl = ptfbbux->BIABL;
    while (ptbiabl != 0)
      {
      (void) fprintf (fd,"  label%d : BLOCK (",nrlabel);
      if (ptbiabl->CNDABL != 0)
        {
        fbl_printablfile( fd, ptbiabl->CNDABL );
        (void) fprintf (fd," = '1')\n" );
        }
      else
        fbl_toolbug (19,"fbl_decomp",ptfbbux->NAME,0);
  
      (void) fprintf (fd,"  BEGIN\n    %s <= GUARDED ",
                      fbl_vectorize(ptfbbux->NAME));
      if (ptbiabl->VALABL != 0)
        {
        fbl_printablfile( fd, ptbiabl->VALABL );
        (void) fprintf (fd,";\n" );
        }
      else
        fbl_toolbug (20,"fbl_decomp",ptfbbux->NAME,0);

      (void) fprintf (fd,"  END BLOCK label%d;\n",nrlabel);
      ptbiabl = ptbiabl->NEXT;
      nrlabel++;
      }
  
    ptfbbux = ptfbbux->NEXT;
  }

    /* ###------------------------------------------------------### */
    /*    Print out a block statement  with one guarded concurrent  */
    /* signal assignment for each BIABL of each BEBUS       */
    /* ###------------------------------------------------------### */

  ptfbbus = pthedfbfig->BEBUS;
  while (ptfbbus != 0)
  {
    ptbiabl = ptfbbus->BIABL;
    while (ptbiabl != 0)
      {
      (void) fprintf (fd,"\tlabel%d : BLOCK (",nrlabel);
      if (ptbiabl->CNDABL != 0)
        {
        fbl_printablfile ( fd, ptbiabl->CNDABL );
        (void) fprintf (fd," = '1')\n" );
        }
      else
        fbl_toolbug (19,"fbl_decomp",ptfbbus->NAME,0);
  
      (void) fprintf (fd,"\tBEGIN\n\t%s <= GUARDED ",fbl_vectorize(ptfbbus->NAME));
      if (ptbiabl->VALABL != 0)
        {
        fbl_printablfile( fd, ptbiabl->VALABL );
        (void) fprintf (fd,";\n" );
        }
      else
        fbl_toolbug (20,"fbl_decomp",ptfbbus->NAME,0);

      (void) fprintf (fd,"\tEND BLOCK label%d;\n",nrlabel);
      ptbiabl = ptbiabl->NEXT;
      nrlabel++;
      }
  
    ptfbbus = ptfbbus->NEXT;
  }

    /* ###------------------------------------------------------### */
    /*    Print out a concurrent signal assignment for each BEOUT   */
    /* ###------------------------------------------------------### */

  ptfbout = pthedfbfig->BEOUT;

  while (ptfbout != 0)
  {
    if (ptfbout->ABL != 0)
      {
      (void) fprintf (fd,"\n%s <= ",fbl_vectorize(ptfbout->NAME) );
      fbl_printablfile( fd, ptfbout->ABL );
      (void) fprintf (fd,";\n" );
      }
    else
    {
      fbl_error (40,ptfbout->NAME);
    }

    ptfbout = ptfbout->NEXT;
  }
    
  (void) fprintf (fd,"END;\n");
  fbl_message    (14,pthedfbfig->NAME,0);
  (void) fclose  (fd);

}
