/*------------------------------------------------------------\
|                                                             |
| Tool    :                     VASY                          |
|                                                             |
| File    :                 vasy_drvalc.c                     |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include VEX_H
# include VPN_H
# include VTL_H
# include RTN_H

# include "vasy_debug.h"
# include "vasy_error.h"
# include "vasy_shared.h"
# include "vasy_simprtl.h"
# include "vasy_drvvex.h"
# include "vasy_drvalc.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Private variables                      |
|                                                             |
\------------------------------------------------------------*/

  static FILE      *VasyFile        = (FILE *)0;
  static long       VasyNumberBlock = 0;

  static short VasyConvertType[ VEX_MAX_TYPE ] =
  {
    VEX_TYPE_SEVERITY          , /* SEVERITY          */
    VEX_TYPE_BIT               , /* BOOLEAN           */
    VEX_TYPE_BIT               , /* BIT               */
    VEX_TYPE_BIT_VECTOR        , /* INTEGER           */
    VEX_TYPE_BIT_VECTOR        , /* NATURAL           */
    VEX_TYPE_BIT_VECTOR        , /* BIT_VECTOR        */
    VEX_TYPE_BIT               , /* STD_ULOGIC        */
    VEX_TYPE_BIT               , /* STD_LOGIC         */
    VEX_TYPE_BIT_VECTOR        , /* STD_ULOGIC_VECTOR */
    VEX_TYPE_BIT_VECTOR        , /* STD_LOGIC_VECTOR  */
    VEX_TYPE_BIT               , /* X01               */
    VEX_TYPE_BIT               , /* X01Z              */
    VEX_TYPE_BIT               , /* UX01              */
    VEX_TYPE_BIT               , /* UX01Z             */
    VEX_TYPE_BIT_VECTOR        , /* UNSIGNED          */
    VEX_TYPE_BIT_VECTOR        , /* SIGNED            */
    VEX_TYPE_BIT               , /* SMALL_INT         */
    VEX_TYPE_REG_BIT           , /* REG_BIT           */
    VEX_TYPE_REG_VECTOR        , /* REG_VECTOR        */
    VEX_TYPE_MUX_BIT           , /* MUX_BIT           */
    VEX_TYPE_MUX_VECTOR        , /* MUX_VECTOR        */
    VEX_TYPE_WOR_BIT           , /* WOR_BIT           */
    VEX_TYPE_WOR_VECTOR        , /* WOR_VECTOR        */
    VEX_TYPE_BIT_VECTOR          /* ENUMERATE         */
  };

  static rtlfig_list *VasyRtlFigure = (rtlfig_list *)0;
  static rtlasg_list *VasyRtlAsg    = (rtlasg_list *)0;
  static long         VasyNumberDef = 0;

  static char         VasyBuffer[ 512 ];

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                   VasyAllianceTreatVexOperandAtom           |
|                                                             |
\------------------------------------------------------------*/

static vexexpr *VasyAllianceTreatVexOperandAtom( Operand )

  vexexpr *Operand;
{
  vexexpr      *VexAtom;
  rtldecl_list *RtlDeclar;
  rtlasg_list  *RtlAssign;
  char          Buffer[ 32 ];

  if ( ! IsVexNodeAtom( Operand ) )
  {
    sprintf( Buffer, "rtlatom_%ld", VasyNumberDef++ );
    VexAtom = createvexatomvec( Buffer, Operand->WIDTH - 1, 0 );

    if ( IsVexNodeSigned( Operand ) ) SetVexNodeSigned( VexAtom );

    RtlDeclar = addrtldecl( VasyRtlFigure, VexAtom, RTL_DECLAR_SIGNAL );
    RtlDeclar->BASE = VEX_TYPE_BIT_VECTOR;
    RtlDeclar->KIND = RTL_KIND_NONE;
    RtlDeclar->DIR  = RTL_DIR_INOUT;

    SetVasyRtlDeclarRead( RtlDeclar );
    SetVasyRtlDeclarAsg( RtlDeclar );

    RtlAssign = addrtlasgafter( VasyRtlFigure, VasyRtlAsg,
                                dupvexexpr( VexAtom ), RTL_ASG_COMBINATORIAL );
    RtlAssign->VEX_DATA = dupvexexpr( Operand );
    Operand = VexAtom;
  }

  return( Operand );
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyAllianceTreatVexArith                 |
|                                                             |
\------------------------------------------------------------*/

static vexexpr *VasyAllianceCreateVexCarry( Operand )

  vexexpr *Operand;
{
  vexexpr *VexAtom;
  short    Left;

  if ( ! IsVexAtomLiteral( Operand ) )
  {
    if ( IsVexNodeDown( Operand ) ) Left = Operand->LEFT - 1;
    else                            Left = Operand->LEFT + 1;

    VexAtom = createvexatomvec( GetVexAtomValue( Operand ), Left, Operand->RIGHT );
  }
  else
  {
    strcpy( VasyBuffer, GetVexAtomValue( Operand ) );

    if ( Operand->WIDTH > 2 ) VasyBuffer[ 1 ] = '"';
    else                      VasyBuffer[ 1 ] = '\'';

    VexAtom = createvexatomlit( &VasyBuffer[ 1 ] );
  }

  return( VexAtom );
}

static vexexpr *VasyAllianceTreatVexArith( VexExpr )

  vexexpr *VexExpr;
{
  rtldecl_list *RtlDeclCarry;
  rtldecl_list *RtlDeclSum;
  rtlasg_list  *RtlAsgCarry;
  rtlasg_list  *RtlAsgSum;
  vexexpr      *Operand1;
  vexexpr      *Operand2;
  vexexpr      *VexSum;
  vexexpr      *VexCarry;
  vexexpr      *VexAtomCarry;
  vexexpr      *VexAtomSum;
  vexexpr      *VexAnd1;
  vexexpr      *VexAnd2;
  vexexpr      *VexAnd3;
  vexexpr      *VexAtom0;
  vexexpr      *VexAtom1;
  vexexpr      *VexAtom2;
  vexexpr      *VexAtom3;
  vexexpr      *VexAtom4;
  char          Buffer[ 32 ];
  int           MaxWidth;
  long          Oper;

  VexAtom1 = (vexexpr *)0;
  VexAtom2 = (vexexpr *)0;
  VexAtom3 = (vexexpr *)0;
  VexAtom4 = (vexexpr *)0;

  Oper = GetVexOperValue( VexExpr );

  Operand1 = GetVexOperand( VexExpr->OPERAND       );
  Operand2 = GetVexOperand( VexExpr->OPERAND->NEXT );

  MaxWidth = Operand1->WIDTH;
  if ( Operand2->WIDTH > MaxWidth ) MaxWidth = Operand2->WIDTH;

  sprintf( Buffer, "rtlcarry_%ld", VasyNumberDef );
  VexAtomCarry = createvexatomvec( Buffer,  MaxWidth - 1, 0 );

  RtlDeclCarry = addrtldecl( VasyRtlFigure, dupvexexpr( VexAtomCarry ), RTL_DECLAR_SIGNAL );
  RtlDeclCarry->BASE = VEX_TYPE_BIT_VECTOR;
  RtlDeclCarry->KIND = RTL_KIND_NONE;
  RtlDeclCarry->DIR  = RTL_DIR_INOUT;

  SetVasyRtlDeclarRead( RtlDeclCarry );
  SetVasyRtlDeclarAsg( RtlDeclCarry );

  sprintf( Buffer, "rtlsum_%ld", VasyNumberDef++ );
  VexAtomSum = createvexatomvec( Buffer,  MaxWidth - 1, 0 );

  RtlDeclSum = addrtldecl( VasyRtlFigure, dupvexexpr( VexAtomSum ), RTL_DECLAR_SIGNAL );
  RtlDeclSum->BASE = VEX_TYPE_BIT_VECTOR;
  RtlDeclSum->KIND = RTL_KIND_NONE;
  RtlDeclSum->DIR  = RTL_DIR_INOUT;

  SetVasyRtlDeclarRead( RtlDeclSum );
  SetVasyRtlDeclarAsg( RtlDeclSum );
/*
** The two operands should be transformed to atomic expressions with the same widths
*/
  Operand1 = dupvexexpr( VasyAllianceTreatVexOperandAtom( Operand1 ) );
  Operand2 = dupvexexpr( VasyAllianceTreatVexOperandAtom( Operand2 ) );

  VexAtom0 = createvexatomvec( GetVexAtomValue( VexAtomCarry ), 0, 0 );

  if ( MaxWidth > 1 )
  {
    VexAtom1 = VasyAllianceCreateVexCarry( Operand1 );
    VexAtom2 = VasyAllianceCreateVexCarry( Operand2 );

    VexAtom3 = createvexatomvec( GetVexAtomValue( VexAtomCarry ), MaxWidth - 2, 0 );
    VexAtom4 = createvexatomvec( GetVexAtomValue( VexAtomCarry ), MaxWidth - 1, 1 );
  }

  if ( Oper == VEX_SUB )
  {
    Operand2 = createvexunaryexpr( VEX_NOT, MaxWidth, Operand2 );

    if ( MaxWidth > 1 )
    {
      VexAtom2 = createvexunaryexpr( VEX_NOT, MaxWidth - 1, VexAtom2 );
    }
  }

  VexSum = createvexbinexpr( VEX_XOR, MaxWidth, Operand1, Operand2 );
  VexSum = createvexbinexpr( VEX_XOR, MaxWidth,
                             VexSum, dupvexexpr( VexAtomCarry ) );

  RtlAsgSum = addrtlasg( VasyRtlFigure, dupvexexpr( VexAtomSum ), RTL_ASG_COMBINATORIAL );
  RtlAsgSum->VEX_DATA = simpvexexpr( VexSum );

  if ( MaxWidth > 1 )
  {
    VexAnd1 = createvexbinexpr( VEX_AND, MaxWidth - 1, dupvexexpr( VexAtom1 ),
                                                       dupvexexpr( VexAtom2 ) );
    VexAnd2 = createvexbinexpr( VEX_AND, MaxWidth - 1, VexAtom1, dupvexexpr( VexAtom3 ) );
    VexAnd3 = createvexbinexpr( VEX_AND, MaxWidth - 1, VexAtom2, VexAtom3 );

    VexCarry = createvexbinexpr( VEX_OR, MaxWidth - 1, VexAnd1 , VexAnd2 );
    VexCarry = createvexbinexpr( VEX_OR, MaxWidth - 1, VexCarry, VexAnd3 );

    RtlAsgCarry = addrtlasg( VasyRtlFigure, VexAtom4, RTL_ASG_COMBINATORIAL );
    RtlAsgCarry->VEX_DATA = simpvexexpr( VexCarry );
  }

  RtlAsgCarry = addrtlasg( VasyRtlFigure, VexAtom0,  RTL_ASG_COMBINATORIAL );

  if ( Oper == VEX_ADD )
  {
    RtlAsgCarry->VEX_DATA = createvexatomlit( VEX_ATOM_ZERO );
  }
  else
  {
    RtlAsgCarry->VEX_DATA = createvexatomlit( VEX_ATOM_ONE );
  }

  freevexexpr( VexExpr );

  return( VexAtomSum );
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyAllianceTreatVexNeg                   |
|                                                             |
\------------------------------------------------------------*/

static vexexpr *VasyAllianceTreatVexNeg( VexExpr )

  vexexpr *VexExpr;
{
  vexexpr *Operand;
  short    Width;

  Width   = VexExpr->WIDTH;
  Operand = GetVexOperand( VexExpr->OPERAND );
  SetVexOperand( VexExpr->OPERAND, NULL );
  freevexexpr( VexExpr );

  VexExpr = createvexoper( VEX_SUB, Width );
  addvexhexpr( VexExpr, Operand );
  Operand = createvexatombit( VEX_ATOM_ZERO );
  addvexhexpr( VexExpr, Operand );

  return( VasyAllianceTreatVexArith( VexExpr ) );
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyAllianceTreatVexRelational            |
|                                                             |
\------------------------------------------------------------*/

static vexexpr *VasyAllianceTreatVexRelational( VexExpr )

  vexexpr *VexExpr;
{
  rtldecl_list *RtlDeclCarry;
  rtldecl_list *RtlDeclSub;
  rtlasg_list  *RtlAsgCarry;
  rtlasg_list  *RtlAsgSub;
  vexexpr      *Operand1;
  vexexpr      *Operand2;
  vexexpr      *Operand3;
  vexexpr      *VexSub;
  vexexpr      *VexCarry;
  vexexpr      *VexAtomCarry;
  vexexpr      *VexAtomSub;
  vexexpr      *VexAnd1;
  vexexpr      *VexAnd2;
  vexexpr      *VexAnd3;
  vexexpr      *VexAtom0;
  vexexpr      *VexAtom1;
  vexexpr      *VexAtom2;
  vexexpr      *VexAtom3;
  vexexpr      *VexAtom4;
  vexexpr      *VexAtom5;
  vexexpr      *VexAtom6;
  vexexpr      *VexAtom7;
  char          Buffer[ 32 ];
  int           MaxWidth;
  long          Oper;

  Oper = GetVexOperValue( VexExpr );

  Operand1 = GetVexOperand( VexExpr->OPERAND       );
  Operand2 = GetVexOperand( VexExpr->OPERAND->NEXT );

  MaxWidth = Operand1->WIDTH;
  if ( Operand2->WIDTH > MaxWidth ) MaxWidth = Operand2->WIDTH;

  MaxWidth = MaxWidth + 1;

  sprintf( Buffer, "rtlcarry_%ld", VasyNumberDef );
  VexAtomCarry = createvexatomvec( Buffer,  MaxWidth - 1, 0 );

  RtlDeclCarry = addrtldecl( VasyRtlFigure, dupvexexpr( VexAtomCarry ), RTL_DECLAR_SIGNAL );
  RtlDeclCarry->BASE = VEX_TYPE_BIT_VECTOR;
  RtlDeclCarry->KIND = RTL_KIND_NONE;
  RtlDeclCarry->DIR  = RTL_DIR_INOUT;

  SetVasyRtlDeclarRead( RtlDeclCarry );
  SetVasyRtlDeclarAsg( RtlDeclCarry );

  sprintf( Buffer, "rtlltgt_%ld", VasyNumberDef++ );
  VexAtomSub = createvexatombit( Buffer );

  RtlDeclSub = addrtldecl( VasyRtlFigure, dupvexexpr( VexAtomSub ), RTL_DECLAR_SIGNAL );
  RtlDeclSub->BASE = VEX_TYPE_BIT;
  RtlDeclSub->KIND = RTL_KIND_NONE;
  RtlDeclSub->DIR  = RTL_DIR_INOUT;

  SetVasyRtlDeclarRead( RtlDeclSub );
  SetVasyRtlDeclarAsg( RtlDeclSub );
/*
** The two operands should be transformed to atomic expressions with the same widths
*/
  Operand1 = VasyAllianceTreatVexOperandAtom( Operand1 );
  Operand2 = VasyAllianceTreatVexOperandAtom( Operand2 );

  if ( ( Oper == VEX_GT ) ||
       ( Oper == VEX_LE ) )
  {
    Operand3 = Operand1;
    Operand1 = Operand2;
    Operand2 = Operand3;
  }

  VexAtom0 = createvexatomvec( GetVexAtomValue( VexAtomCarry ), 0, 0 );

  VexAtom1 = VasyAllianceCreateVexCarry( Operand1 );
  VexAtom2 = VasyAllianceCreateVexCarry( Operand2 );

  VexAtom2 = createvexunaryexpr( VEX_NOT, MaxWidth - 1, VexAtom2 );
  VexAtom3 = createvexatomvec( GetVexAtomValue( VexAtomCarry ), MaxWidth - 2, 0 );
  VexAtom4 = createvexatomvec( GetVexAtomValue( VexAtomCarry ), MaxWidth - 1, 1 );


  VexAtom5 = createvexatomvec( GetVexAtomValue( Operand1 ), MaxWidth - 1, MaxWidth - 1 );
  VexAtom6 = createvexatomvec( GetVexAtomValue( Operand2 ), MaxWidth - 1, MaxWidth - 1 );
  VexAtom6 = createvexunaryexpr( VEX_NOT, 1, VexAtom6 );
  VexAtom7 = createvexatomvec( GetVexAtomValue( VexAtomCarry ),  MaxWidth -1, MaxWidth - 1 );

  VexSub = createvexbinexpr( VEX_XOR, 1, VexAtom5, VexAtom6 );
  VexSub = createvexbinexpr( VEX_XOR, 1, VexSub  , VexAtom7 );

  if ( ( Oper == VEX_GE ) ||
       ( Oper == VEX_LE ) )
  {
    VexSub = createvexunaryexpr( VEX_NOT, 1, VexSub );
  }

  RtlAsgSub = addrtlasg( VasyRtlFigure, dupvexexpr( VexAtomSub ), RTL_ASG_COMBINATORIAL );
  RtlAsgSub->VEX_DATA = simpvexexpr( VexSub );

  VexAnd1 = createvexbinexpr( VEX_AND, MaxWidth - 1, dupvexexpr( VexAtom1 ),
                                                     dupvexexpr( VexAtom2 ) );
  VexAnd2 = createvexbinexpr( VEX_AND, MaxWidth - 1, VexAtom1, dupvexexpr( VexAtom3 ) );
  VexAnd3 = createvexbinexpr( VEX_AND, MaxWidth - 1, VexAtom2, VexAtom3 );

  VexCarry = createvexbinexpr( VEX_OR, MaxWidth - 1, VexAnd1 , VexAnd2 );
  VexCarry = createvexbinexpr( VEX_OR, MaxWidth - 1, VexCarry, VexAnd3 );

  RtlAsgCarry = addrtlasg( VasyRtlFigure, VexAtom4, RTL_ASG_COMBINATORIAL );
  RtlAsgCarry->VEX_DATA = simpvexexpr( VexCarry );

  RtlAsgCarry = addrtlasg( VasyRtlFigure, VexAtom0,  RTL_ASG_COMBINATORIAL );
  RtlAsgCarry->VEX_DATA = createvexatomlit( VEX_ATOM_ONE );

  freevexexpr( VexExpr );

  return( VexAtomSub );
}

/*------------------------------------------------------------\
|                                                             |
|                     VasyAllianceTreatVexAtomDontCare        |
|                                                             |
\------------------------------------------------------------*/

static vexexpr *VasyAllianceTreatVexAtomDontCare( VexAtom )
  
  vexexpr *VexAtom;
{
  char *Value;
  int   Index;

  Value = GetVexAtomValue( VexAtom );

  if ( strchr( Value, '-' ) )
  {
    for ( Index = 0; Value[ Index ] != '\0'; Index++ )
    {
      if ( Value[ Index ] != '-' ) VasyBuffer[ Index ] = Value[ Index ];
      else                         VasyBuffer[ Index ] = '0';
    }

    VasyBuffer[ Index ] = '\0';

    Value = namealloc( VasyBuffer );
    SetVexAtomValue( VexAtom, Value );
  }

  return( VexAtom );
}

/*------------------------------------------------------------\
|                                                             |
|                     VasyAllianceTreatVex                    |
|                                                             |
\------------------------------------------------------------*/

static vexexpr *VasyAllianceTreatVex( VexExpr )

  vexexpr *VexExpr;
{
  chain_list *ScanChain;
  vexexpr    *Operand;
  long        Oper;

  if ( VexExpr == (vexexpr *)0 )
  {
    return( VexExpr );
  }

  if ( IsVexNodeAtom( VexExpr ) )
  {
    if ( IsVexAtomLiteral( VexExpr ) )
    {
      VexExpr = VasyAllianceTreatVexAtomDontCare( VexExpr );
    }

    return( VexExpr );
  }
  else
  if ( IsVexNodeOper( VexExpr ) )
  {
    Oper = GetVexOperValue( VexExpr );

    if ( ( Oper == VEX_ADD ) ||
         ( Oper == VEX_SUB ) )
    {
      return( VasyAllianceTreatVexArith( VexExpr ) );
    }
    else
    if ( ( Oper == VEX_LT ) ||
         ( Oper == VEX_GT ) ||
         ( Oper == VEX_LE ) ||
         ( Oper == VEX_GE ) )
    {
      return( VasyAllianceTreatVexRelational( VexExpr ) );
    }
    else
    if ( Oper == VEX_NEG )
    {
      return( VasyAllianceTreatVexNeg( VexExpr ) );
    }

  }

  for ( ScanChain  = VexExpr->OPERAND;
        ScanChain != (chain_list *)0;
        ScanChain  = ScanChain->NEXT )
  {
    Operand = GetVexOperand( ScanChain );
    Operand = VasyAllianceTreatVex( Operand );
    SetVexOperand( ScanChain, Operand );
  }

  return( VexExpr );
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyAllianceTreatAsg                      |
|                                                             |
\------------------------------------------------------------*/

static void VasyAllianceTreatAsg( RtlAsg )

  rtlasg_list *RtlAsg;
{
  rtlbivex_list *RtlBiVex;

  VasyRtlAsg = RtlAsg;

  RtlAsg->VEX_DATA = VasyAllianceTreatVex( RtlAsg->VEX_DATA );

  for ( RtlBiVex  = RtlAsg->BIVEX;
        RtlBiVex != (rtlbivex_list *)0;
        RtlBiVex  = RtlBiVex->NEXT )
  {
    RtlBiVex->VEX_COND = VasyAllianceTreatVex( RtlBiVex->VEX_COND );
    RtlBiVex->VEX_DATA = VasyAllianceTreatVex( RtlBiVex->VEX_DATA );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceTreatFigure              |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceTreatFigure( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtldecl_list *RtlDeclar;
  rtlasg_list  *RtlAsg;
  vexexpr      *VexAtom;
  char         *AtomValue;
  int           Type;

  VasyRtlFigure = RtlFigure;

  for ( RtlAsg  = RtlFigure->ASSIGN;
        RtlAsg != (rtlasg_list *)0;
        RtlAsg  = RtlAsg->NEXT )
  {
    VasyAllianceTreatAsg( RtlAsg );
  }

  for ( Type = 0; Type < RTL_MAX_DECLAR_TYPE; Type++ )
  {
    for ( RtlDeclar  = RtlFigure->DECLAR[ Type ];
          RtlDeclar != (rtldecl_list *)0;
          RtlDeclar  = RtlDeclar->NEXT )
    {
      RtlDeclar->BASE = VasyConvertType[ RtlDeclar->BASE ];
      RtlDeclar->KIND = RTL_KIND_NONE;
    }
  }

  for ( RtlAsg  = RtlFigure->ASSIGN;
        RtlAsg != (rtlasg_list *)0;
        RtlAsg  = RtlAsg->NEXT )
  {
    if ( ( RtlAsg->TYPE == RTL_ASG_REGISTER ) ||
         ( RtlAsg->TYPE == RTL_ASG_TRISTATE ) )
    {
      VexAtom   = RtlAsg->VEX_ATOM;
      AtomValue = GetVexAtomValue( VexAtom );
      RtlDeclar = searchrtldecl( RtlFigure, AtomValue );

      if ( RtlAsg->TYPE == RTL_ASG_REGISTER )
      {
        RtlDeclar->KIND = RTL_KIND_REGISTER;

        if ( isvextypevector( RtlDeclar->BASE ) ) RtlDeclar->BASE = VEX_TYPE_REG_VECTOR;
        else                                      RtlDeclar->BASE = VEX_TYPE_REG_BIT;
      }
      else
      {
        RtlDeclar->KIND = RTL_KIND_BUS;

        if ( isvextypevector( RtlDeclar->BASE ) ) RtlDeclar->BASE = VEX_TYPE_MUX_VECTOR;
        else                                      RtlDeclar->BASE = VEX_TYPE_MUX_BIT;
      }
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAlliancePort                     |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAlliancePort( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtldecl_list *RtlDeclar;
  vexexpr      *VexPort;
  char         *PortName;
  char         *PortDir;
  char         *PortBase;
  char         *PortKind;
  char         *PortVecDir;

  RtlDeclar = RtlFigure->DECLAR[ RTL_DECLAR_PORT ];

  if ( RtlDeclar != (rtldecl_list *)0 )
  {
    fprintf( VasyFile, "PORT(\n" );

    while ( RtlDeclar != (rtldecl_list *)0 )
    {
      VexPort  = RtlDeclar->VEX_ATOM;
      PortName = GetVexAtomValue( VexPort );
      PortDir  = RTL_DIR_TYPE[ RtlDeclar->DIR ];
      PortKind = RTL_KIND_TYPE[ RtlDeclar->KIND ];
      PortBase = VEX_TYPE_NAME[ RtlDeclar->BASE ];

      fprintf( VasyFile, "  %s\t: %s %s", PortName, PortDir, PortBase );
  
      if ( isvextypevector( RtlDeclar->BASE ) )
      {
        if ( IsVexAtomUp( VexPort ) ) PortVecDir = "TO";
        else                          PortVecDir = "DOWNTO";
  
        fprintf( VasyFile, "(%d %s %d)", VexPort->LEFT, PortVecDir, VexPort->RIGHT );
      }

      if ( RtlDeclar->KIND != RTL_KIND_NONE ) fprintf( VasyFile, " %s", PortKind );

      if ( RtlDeclar->NEXT != (rtldecl_list *)0 )
      {
        fprintf( VasyFile, ";\n" );
      }
  
      RtlDeclar = RtlDeclar->NEXT;
    }

    fprintf( VasyFile, "\n);\n" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceDeclar                   |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceDeclar( RtlFigure, Type )

  rtlfig_list *RtlFigure;
  int          Type;
{
  rtldecl_list *RtlDeclar;
  vexexpr      *VexDeclar;
  char         *AtomValue;
  vexexpr      *VexInit;
  char         *DeclarName;
  char         *DeclarDir;
  char         *DeclarBase;
  char         *DeclarVecDir;
  char         *DeclarKind;

  for ( RtlDeclar  = RtlFigure->DECLAR[ Type ];
        RtlDeclar != (rtldecl_list *)0;
        RtlDeclar  = RtlDeclar->NEXT )
  {
    if ( ! IsVasyRtlDeclarUsed( RtlDeclar ) ) continue;

    VexDeclar  = RtlDeclar->VEX_ATOM;
    VexInit    = RtlDeclar->VEX_INIT;
    DeclarName = GetVexAtomValue( VexDeclar );
    DeclarDir  = RTL_DIR_TYPE[ RtlDeclar->DIR ];
    DeclarBase = VEX_TYPE_NAME[ RtlDeclar->BASE ];
    DeclarKind = RTL_KIND_TYPE[ RtlDeclar->KIND ];

    if ( Type == RTL_DECLAR_CONSTANT )
    {
      fprintf( VasyFile, "  CONSTANT %s\t: %s", DeclarName, DeclarBase );
    }
    else
    {
      fprintf( VasyFile, "  SIGNAL %s\t: %s", DeclarName, DeclarBase );
    }
  
    if ( isvextypevector( RtlDeclar->BASE ) )
    {
      if ( IsVexAtomUp( VexDeclar ) ) DeclarVecDir = "TO";
      else                            DeclarVecDir = "DOWNTO";
  
      fprintf( VasyFile, "(%d %s %d)", VexDeclar->LEFT, DeclarVecDir, VexDeclar->RIGHT );
    }

    if ( RtlDeclar->KIND != RTL_KIND_NONE ) fprintf( VasyFile, " %s", DeclarKind );

    if ( Type == RTL_DECLAR_CONSTANT )
    {
      fprintf( VasyFile, " := " );

      if ( RtlDeclar->BASE == VEX_TYPE_BOOLEAN )
      {
        if ( IsVexNodeAtom( VexInit ) )
        {
          AtomValue = GetVexAtomValue( VexInit );

          if ( AtomValue == VEX_ATOM_ONE ) fprintf( VasyFile, " TRUE" );
          else                             fprintf( VasyFile, " FALSE" );
        }
        else
        {
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexInit );
        }
      }
      else
      {
        VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexInit );
      }
    }

    fprintf( VasyFile, ";\n" );

    if ( RtlDeclar->NEXT == (rtldecl_list *)0 )
    {
      fprintf( VasyFile, "\n" );
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                 VasyDriveAllianceFsmDeclar                  |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceFsmDeclar( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtlfsm_list      *RtlFsm;
  rtlfsmstate_list *RtlState;

  for ( RtlFsm  = RtlFigure->FSM;
        RtlFsm != (rtlfsm_list *)0;
        RtlFsm  = RtlFsm->NEXT )
  {
    fprintf( VasyFile, "  TYPE STATE_TYPE_%s IS (", RtlFsm->NAME );

    for ( RtlState  = RtlFsm->STATE;
          RtlState != (rtlfsmstate_list *)0;
          RtlState  = RtlState->NEXT )
    {
      fprintf( VasyFile, " %s", RtlState->NAME );

      if ( RtlState->NEXT != (rtlfsmstate_list *)0 )
      {
        fprintf( VasyFile, "," );
      }
    }

    fprintf( VasyFile, " );\n" );
    fprintf( VasyFile, "  SIGNAL CS_%s : STATE_TYPE_%s;\n\n", RtlFsm->NAME, RtlFsm->NAME );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceEntity                   |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceEntity( RtlFigure )

  rtlfig_list *RtlFigure;
{
  fprintf( VasyFile, "ENTITY %s IS\n", RtlFigure->NAME );

  VasyDriveAlliancePort( RtlFigure );

  fprintf( VasyFile, "END %s;\n\n", RtlFigure->NAME );
}

/*------------------------------------------------------------\
|                                                             |
|                VasyDriveAllianceGetVexCond                  |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceVexCond( RtlFigure, VexExpr, Positive )

  rtlfig_list *RtlFigure;
  vexexpr     *VexExpr;
  int          Positive;
{
  vexexpr *VexCond;
  long     Oper;
  char     CharEqual0;
  char     CharEqual1;
  char     CharEqual;

  VexCond    = VexExpr;

  if ( Positive ) 
  {
    CharEqual0 = '0';
    CharEqual1 = '1';
  }
  else
  {
    CharEqual0 = '1';
    CharEqual1 = '0';
  }

  CharEqual = CharEqual1;

  if ( IsVexNodeOper( VexCond ) )
  {
    Oper = GetVexOperValue( VexCond );

    if ( Oper == VEX_NOT )
    {
      CharEqual = CharEqual0;
      VexCond   = dupvexexpr( GetVexOperand( VexCond->OPERAND ) );
    }
    else
    if ( isvexnegativeoper( Oper ) )
    {
      CharEqual = CharEqual0;
      Oper      = getvexnotoper( Oper );
      VexCond   = dupvexexpr( VexCond );

      SetVexOperValue( VexCond, Oper );
    }
  }

  fprintf( VasyFile, "(" );
  VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexCond );
  fprintf( VasyFile, " = '%c')", CharEqual );

  if ( VexCond != VexExpr )
  {
    freevexexpr( VexCond );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceAssign                   |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceAssign( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtldecl_list  *RtlDeclar;
  rtlsym        *RtlSymbol;
  rtlasg_list   *RtlAsg;
  rtlbivex_list *ScanBiVex;
  char          *AtomValue;
  char          *AtomName;
  vexexpr       *VexDeclar;
  vexexpr       *VexInit;
  vexexpr       *VexAtom;
  vexexpr       *VexSumAsync;
  vexexpr       *VexData;
  vexexpr       *VexAsync;
  vexexpr       *VexSync;
  int            Type;
  int            ScanPos;
  char           Literal;

  for ( Type = 0; Type < RTL_MAX_DECLAR_TYPE; Type++ )
  {
    for ( RtlDeclar  = RtlFigure->DECLAR[ Type ];
          RtlDeclar != (rtldecl_list *)0;
          RtlDeclar  = RtlDeclar->NEXT )
    {
      if ( ( RtlDeclar->DIR == RTL_DIR_IN      ) ||
           ( ! IsVasyRtlDeclarUsed( RtlDeclar ) ) ) continue;

      VexDeclar  = RtlDeclar->VEX_ATOM;
      VexInit    = RtlDeclar->VEX_INIT;

      if ( VexInit == (vexexpr *)0 ) continue;

      AtomName   = GetVexAtomValue( VexDeclar );
      AtomValue  = GetVexAtomValue( VexInit   );

      for ( ScanPos = 0; ScanPos < VexDeclar->WIDTH; ScanPos++ )
      {
        RtlSymbol = &RtlDeclar->DECL_SYM[ ScanPos ];

        if ( ( IsVasyRtlDeclarRead( RtlSymbol  ) ) &&
             ( ! IsVasyRtlDeclarAsg( RtlSymbol ) ) )
        {
          Literal = AtomValue[ ScanPos + 1 ];

          fprintf( VasyFile, "  " );

          if ( RtlSymbol->INDEX != -1 )
          {
            fprintf( VasyFile, "  %s(%d) <= '%c';\n", AtomName, RtlSymbol->INDEX, Literal );
          }
          else
          {
            fprintf( VasyFile, "  %s <= '%c';\n", AtomName, Literal );
          }
        }
      }
    }
  }

  for ( RtlAsg  = RtlFigure->ASSIGN;
        RtlAsg != (rtlasg_list *)0;
        RtlAsg  = RtlAsg->NEXT )
  {
    if ( RtlAsg->TYPE == RTL_ASG_COMBINATORIAL )
    {
      fprintf( VasyFile, "  " );
      VasyDriveVexExprVhdl( VasyFile, RtlFigure, RtlAsg->VEX_ATOM );
      fprintf( VasyFile, " <= " );
      VasyDriveVexExprVhdl( VasyFile, RtlFigure, RtlAsg->VEX_DATA );
      fprintf( VasyFile, ";\n" );
    }
    else
    if ( RtlAsg->TYPE == RTL_ASG_MULTIPLEXOR )
    {
      fprintf( VasyFile, "  WITH " );
      VasyDriveVexExprVhdl( VasyFile, RtlFigure, RtlAsg->VEX_DATA );
      fprintf( VasyFile, "  SELECT\n    " );
      VasyDriveVexExprVhdl( VasyFile, RtlFigure, RtlAsg->VEX_ATOM );
      fprintf( VasyFile, " <= " );

      for ( ScanBiVex  = RtlAsg->BIVEX;
            ScanBiVex != (rtlbivex_list *)0;
            ScanBiVex  = ScanBiVex->NEXT )
      {
        VasyDriveVexExprVhdl( VasyFile, RtlFigure, ScanBiVex->VEX_DATA );
        fprintf( VasyFile, " WHEN " );

        if ( ScanBiVex->VEX_COND != (vexexpr *)0 )
        {
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, ScanBiVex->VEX_COND );
        }
        else
        {
          fprintf( VasyFile, " OTHERS " );
        }

        if ( ScanBiVex->NEXT != (rtlbivex_list *)0 )
        {
          fprintf( VasyFile, ",\n     " );
        }
      }

      fprintf( VasyFile, ";\n" );
    }
    else
    if ( RtlAsg->TYPE == RTL_ASG_CONDITIONAL )
    {
      fprintf( VasyFile, "  " );
      VasyDriveVexExprVhdl( VasyFile, RtlFigure, RtlAsg->VEX_ATOM );
      fprintf( VasyFile, " <= " );

      for ( ScanBiVex  = RtlAsg->BIVEX;
            ScanBiVex != (rtlbivex_list *)0;
            ScanBiVex  = ScanBiVex->NEXT )
      {
        VasyDriveVexExprVhdl( VasyFile, RtlFigure, ScanBiVex->VEX_DATA );

        if ( ScanBiVex->NEXT != (rtlbivex_list *)0 )
        {
          fprintf( VasyFile, " WHEN " );
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, ScanBiVex->VEX_COND );
          fprintf( VasyFile, " ELSE\n     " );
        }
      }

      fprintf( VasyFile, ";\n" );
    }
    else
    if ( ( RtlAsg->TYPE == RTL_ASG_REGISTER ) ||
         ( RtlAsg->TYPE == RTL_ASG_TRISTATE ) )
    {
      VexAtom     = RtlAsg->VEX_ATOM;
      VexSumAsync = (vexexpr *)0;

      ScanBiVex = RtlAsg->BIVEX;

      while ( ScanBiVex != (rtlbivex_list *)0 )
      {
        VexData  = ScanBiVex->VEX_DATA;
        VexAsync = (vexexpr *)0;

        if ( ScanBiVex->TYPE == RTL_BIVEX_ASYNC_RESET )
        {
          VexAsync = ScanBiVex->VEX_COND;
          VexData  = createvexatomveclit( VEX_ZERO, VexAtom->WIDTH );
        }
        else
        if ( ScanBiVex->TYPE == RTL_BIVEX_ASYNC_SET )
        {
          VexAsync = ScanBiVex->VEX_COND;
          VexData = createvexatomveclit( VEX_ONE, VexAtom->WIDTH );
        }
        else
        if ( ScanBiVex->TYPE == RTL_BIVEX_ASYNC_WEN )
        {
          VexAsync = ScanBiVex->VEX_COND;
        }
        else break;

        fprintf( VasyFile, "  LABEL%ld : BLOCK ", VasyNumberBlock );
        VasyDriveAllianceVexCond( RtlFigure, VexAsync, 1 );
        fprintf( VasyFile, "\n" );
        fprintf( VasyFile, "  BEGIN\n" );
        fprintf( VasyFile, "    " );
        VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexAtom );
        fprintf( VasyFile, " <= GUARDED " );
        VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexData );
        fprintf( VasyFile, ";\n" );
        fprintf( VasyFile, "  END BLOCK LABEL%ld;\n", VasyNumberBlock );
        VasyNumberBlock++;

        if ( VexAsync != (vexexpr *)0 )
        {
          if ( VexSumAsync == (vexexpr *)0 )
          {
            VexSumAsync = dupvexexpr( VexAsync );
          }
          else
          {
            VexSumAsync = optimvexbinexpr( VEX_OR, 1, VexSumAsync, dupvexexpr( VexAsync ) );
          }
        }

        if ( VexData != ScanBiVex->VEX_DATA )
        {
          freevexexpr( VexData );
        }

        ScanBiVex = ScanBiVex->NEXT;
      }

      if ( ( ScanBiVex    != (rtlbivex_list *)0 ) &&
           ( RtlAsg->TYPE != RTL_ASG_TRISTATE   ) )
      {
        fprintf( VasyFile, "  LABEL%ld : BLOCK ", VasyNumberBlock );
        fprintf( VasyFile, " (" );

        if ( VexSumAsync != (vexexpr *)0 )
        {
          fprintf( VasyFile, " (" );
          VasyDriveAllianceVexCond( RtlFigure, VexSumAsync, 0 );
          fprintf( VasyFile, " AND " );
        }

        if ( ScanBiVex->TYPE == RTL_BIVEX_FALLING_EDGE )
        {
          VasyDriveAllianceVexCond( RtlFigure, ScanBiVex->VEX_COND, 0 );
          fprintf( VasyFile, " AND NOT(" );
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, ScanBiVex->VEX_COND );
          fprintf( VasyFile, "'STABLE)" );
        }
        else
        if ( ScanBiVex->TYPE == RTL_BIVEX_RISING_EDGE  )
        {
          VasyDriveAllianceVexCond( RtlFigure, ScanBiVex->VEX_COND, 1 );
          fprintf( VasyFile, " AND NOT(" );
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, ScanBiVex->VEX_COND );
          fprintf( VasyFile, "'STABLE)" );
        }

        if ( VexSumAsync != (vexexpr *)0 )
        {
          fprintf( VasyFile, " )" );
          freevexexpr( VexSumAsync );
        }

        fprintf( VasyFile, " )\n" );
        fprintf( VasyFile, "  BEGIN\n" );
        fprintf( VasyFile, "    " );
        VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexAtom );
        fprintf( VasyFile, " <= GUARDED " );

        ScanBiVex = ScanBiVex->NEXT;

        while ( ScanBiVex != (rtlbivex_list *)0 )
        {
          VexData = ScanBiVex->VEX_DATA;

          if ( ScanBiVex->TYPE == RTL_BIVEX_SYNC_RESET )
          {
            VexSync = ScanBiVex->VEX_COND;
            VexData  = createvexatomveclit( VEX_ZERO, VexAtom->WIDTH );
          }
          else
          if ( ScanBiVex->TYPE == RTL_BIVEX_SYNC_SET )
          {
            VexSync = ScanBiVex->VEX_COND;
            VexData = createvexatomveclit( VEX_ONE, VexAtom->WIDTH );
          }
          else
          if ( ScanBiVex->TYPE == RTL_BIVEX_SYNC_WEN )
          {
            VexSync = ScanBiVex->VEX_COND;
          }
          else break;

          VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexData );

          if ( VexSync != (vexexpr *)0 )
          {
            fprintf( VasyFile, " WHEN " );
            VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexSync );

            if ( ScanBiVex->NEXT != (rtlbivex_list *)0 )
            {
              fprintf( VasyFile, " ELSE\n     " );
            }
            else
            {
              fprintf( VasyFile, " ELSE " );
              VasyDriveVexExprVhdl( VasyFile, RtlFigure, VexAtom );
            }
          }

          if ( VexData != ScanBiVex->VEX_DATA )
          {
            freevexexpr( VexData );
          }

          ScanBiVex = ScanBiVex->NEXT;
        }

        fprintf( VasyFile, ";\n" );
        fprintf( VasyFile, "  END BLOCK LABEL%ld;\n", VasyNumberBlock );
        VasyNumberBlock++;
      }
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceFsm                      |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceFsm( RtlFigure )

  rtlfig_list *RtlFigure;
{
  rtlfsm_list      *RtlFsm;
  rtlfsmstate_list *FsmState;
  rtlfsmtrans_list *FsmTrans;
  int               ElseIf;
  rtlfsmasg_list   *FsmAsg;
  chain_list       *Support;
  chain_list       *ScanChain;

  for ( RtlFsm  = RtlFigure->FSM;
        RtlFsm != (rtlfsm_list *)0;
        RtlFsm  = RtlFsm->NEXT )
  {
    VasyError( VASY_NOT_YET_IMPLEMENTED_ERROR, "Driver Alliance: FSM" );
/*
**  Drive The transition process
*/
    fprintf( VasyFile, "  PROCESS ( " );

    /* TO BE DONE
    Support = getvexexprsupport( RtlFsm->VEX_COND );
    Support = unionvexexprsupport( Support, RtlFsm->VEX_WEN );
    */
    Support = (chain_list *)0;

    for ( ScanChain  = Support;
          ScanChain != (chain_list *)0;
          ScanChain  = ScanChain->NEXT )
    {
      fprintf( VasyFile, "%s", (char *)ScanChain->DATA );

      if ( ScanChain->NEXT != (chain_list *)0 )
      {
        fprintf( VasyFile, ", " );
      }
    }

    freechain( Support );

    fprintf( VasyFile, " )\n" );
    fprintf( VasyFile, "  BEGIN\n" );

    /* TO BE DONE
    fprintf( VasyFile, "    IF (" );
    VasyDriveAllianceVexCond( RtlFigure, RtlFsm->VEX_COND );
    fprintf( VasyFile, " AND " );
    VasyDriveVexExprVhdl( VasyFile, RtlFigure, RtlFsm->VEX_EDGE );
    fprintf( VasyFile, ")\n" );
    fprintf( VasyFile, "    THEN\n" );

    if ( RtlFsm->VEX_WEN != (vexexpr *)0 )
    {
      fprintf( VasyFile, "\n" );
      fprintf( VasyFile, "      IF " );
      VasyDriveAllianceVexCond( RtlFigure, RtlFsm->VEX_WEN );
      fprintf( VasyFile, "\n" );
      fprintf( VasyFile, "      THEN\n" );
    }
    */

    fprintf( VasyFile, "      CASE ( CS_%s ) IS\n", RtlFsm->NAME );

    for ( FsmState  = RtlFsm->STATE;
          FsmState != (rtlfsmstate_list *)0;
          FsmState  = FsmState->NEXT )
    {
      fprintf( VasyFile, "\n        WHEN %s =>\n\n", FsmState->NAME );

      ElseIf = 0;

      for ( ScanChain  = FsmState->FROM;
            ScanChain != (chain_list *)0;
            ScanChain  = ScanChain->NEXT )
      {
        FsmTrans = GetRtlFsmTrans( ScanChain );
        
        if ( FsmTrans->VEX_COND != (vexexpr *)0 )
        {
          if ( ! ElseIf )
          {
            fprintf( VasyFile, "          IF " );
          }
          else
          if ( ScanChain->NEXT != (chain_list *)0 )
          {
            fprintf( VasyFile, "          ELSIF " );
          }
          else
          {
            fprintf( VasyFile, "          ELSE\n" );
          }

          if ( ScanChain->NEXT != (chain_list *)0 )
          {
            VasyDriveAllianceVexCond( RtlFigure, FsmTrans->VEX_COND );
            fprintf( VasyFile, "\n          THEN\n" );
          }

          ElseIf = 1;
        }

        fprintf( VasyFile, "            CS_%s <= %s;\n", RtlFsm->NAME, FsmTrans->TO->NAME );

        for ( FsmAsg  = FsmTrans->ASSIGN;
              FsmAsg != (rtlfsmasg_list *)0;
              FsmAsg  = FsmAsg->NEXT )
        {
          fprintf( VasyFile, "            " );
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, FsmAsg->VEX_ATOM );
          fprintf( VasyFile, " <= " );
          VasyDriveVexExprVhdl( VasyFile, RtlFigure, FsmAsg->VEX_DATA );
          fprintf( VasyFile, ";\n" );
        }
      }

      if ( ElseIf )
      {
        fprintf( VasyFile, "          END IF;\n" );
      }
    }

    fprintf( VasyFile, "\n      END CASE;\n" );

    /* TO BE DONE
    if ( RtlFsm->VEX_WEN != (vexexpr *)0 )
    {
      fprintf( VasyFile, "      END IF;\n" );
    }
    */

    fprintf( VasyFile, "    END IF;\n" );
    fprintf( VasyFile, "  END PROCESS;\n\n" );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceArchi                    |
|                                                             |
\------------------------------------------------------------*/

static void VasyDriveAllianceArchi( RtlFigure )

  rtlfig_list *RtlFigure;
{
  fprintf( VasyFile, "ARCHITECTURE VBE OF %s IS\n\n", RtlFigure->NAME );

  VasyDriveAllianceDeclar( RtlFigure, RTL_DECLAR_CONSTANT );
  VasyDriveAllianceDeclar( RtlFigure, RTL_DECLAR_SIGNAL   );
  VasyDriveAllianceFsmDeclar( RtlFigure );

  fprintf( VasyFile, "BEGIN\n\n" );

  VasyDriveAllianceAssign( RtlFigure );
  VasyDriveAllianceFsm( RtlFigure );

  fprintf( VasyFile, "END VBE;\n" );
}

/*------------------------------------------------------------\
|                                                             |
|                   VasyDriveAllianceRtlFig                   |
|                                                             |
\------------------------------------------------------------*/

void VasyDriveAllianceRtlFig( RtlFigure, FileName )

  rtlfig_list  *RtlFigure;
  char         *FileName;
{
  if ( IsVasyDebugLevel0() )
  {
    VasyPrintf( stdout, "--> VasyDriveAllianceRtlFig %s\n", FileName );
  }

  if ( IsVasyDebugDriveStdout() )
  {
    VasyFile = stdout;
  }
  else
  {
    VasyFile = mbkfopen( FileName, "vbe", "w" );
  }

  if ( VasyFile != (FILE *)0 )
  {
/*
** Special treatments for Alliance
*/
    VasyDriveAllianceTreatFigure( RtlFigure );
    VasyDriveAllianceEntity( RtlFigure );
    VasyDriveAllianceArchi( RtlFigure );

    if ( ! IsVasyDebugDriveStdout() )
    {
      fclose( VasyFile );
    }
  }

  if ( IsVasyDebugLevel0() )
  {
    VasyPrintf( stdout, "<-- VasyDriveAllianceRtlFig %s\n\n", FileName );
  }
}
