/*------------------------------------------------------------\
|                                                             |
| Tool    :                     VASY                          |
|                                                             |
| File    :                  vasy_main.c                      |
|                                                             |
| Author  :                 Jacomme Ludovic                   |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include VEX_H
# include BDD_H
# include VPN_H
# include VTL_H
# include RTN_H

# include "vasy_error.h"
# include "vasy_debug.h"
# include "vasy_parse.h"
# include "vasy_preanal.h"
# include "vasy_func.h"
# include "vasy_elabo.h"
# include "vasy_reduce.h"
# include "vasy_analys.h"
# include "vasy_simprtl.h"
# include "vasy_drvalc.h"
# include "vasy_drvsyn.h"
# include "vasy_drvvlog.h"
# include "vasy_drvrtl.h"
# include "vasy_main.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Vasy Usage                          |
|                                                             |
\------------------------------------------------------------*/

void VasyUsage()
{
  fprintf( stderr, "\t\tvasy [Options] Input_name Output_name\n\n" );

  fprintf( stdout, "\t\tOptions : -D [Level] [File] Sets Debug mode on\n"   );
  fprintf( stdout, "\t\t          -V                Sets Verbose mode on\n" );
  fprintf( stdout, "\t\t          -I [extention]    Input file format\n" );
  fprintf( stdout, "\t\t          -S                Uses Std_logic instead of Bit\n" );
  fprintf( stdout, "\t\t          -i                Drives initials values\n"  );
  fprintf( stdout, "\t\t          -v                Verilog output\n"  );
  fprintf( stdout, "\t\t          -a                Alliance output\n" );
  fprintf( stdout, "\t\t          -s                Standard VHDL output\n" );
  fprintf( stdout, "\t\t          -r                Rtl output\n" );
  fprintf( stdout, "\n" );

  exit( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                            Main                             |
|                                                             |
\------------------------------------------------------------*/

int main( argc, argv )

   int   argc;
   char *argv[];
{
  vpnfig_list *VpnFigure;
  rtlfig_list *RtlFigure;

  char        *InputFileName;
  char        *OutputFileName;
  char        *InputFormat;
  char         Option;
  int          Number;
  int          Index;

  int          FlagVerbose  = 0;
  int          FlagDebug    = 0;
  int          FlagOption   = 0;
  int          FlagDrive    = 2;
  int          FlagStdLogic = 0;
  int          FlagInitial  = 0;

  alliancebanner_with_contrib(
      "VASY", VASY_VERSION, "VHDL Analysis for Synthesizer",
      "1999", ALLIANCE_VERSION, "Ludovic Jacomme", "Frederic Petrot" );

  mbkenv();
  autenv();
  ablenv();
  bddenv();
  vexenv();
  vpnenv();

  if ( argc < 2 ) VasyUsage();

  InputFileName  = (char *)0;
  OutputFileName = (char *)0;

  for ( Number = 1; Number  < argc; Number++ )
  {
    if ( argv[ Number ][ 0 ] == '-' )
    {
      for ( Index = 1; argv[ Number ][ Index ] != '\0'; Index++ )
      {
        Option = argv[ Number ][ Index ];

        if ( FlagDebug )
        {
          if ( ! VasyDebugAddItem( &argv[ Number ][ Index ], FlagDebug ) )
          {
            VasyUsage();
          }

          break;
        }
        else
        if ( FlagOption )
        {
          if ( ! VasyDebugSetOption( &argv[ Number ][ Index ] ) )
          {
            VasyUsage();
          }

          break;
        }
        else
        {
          if ( Option == 'I' )
          {
            Number = Number + 1;

            if ( Number < argc )
            {
              if ( argv[ Number ][ 0 ] == '.' )
              {
                InputFormat = &argv[ Number ][ 1 ];
              }
              else
              {
                InputFormat = argv[ Number ];
              }

              VPN_IN = namealloc( InputFormat );
            }
            else VasyUsage();

            break;
          }
          
          switch ( Option )
          {
            case 'V' : FlagVerbose = 1;
            break;
            case 'S' : FlagStdLogic = 1;
            break;
            case 'i' : FlagInitial = 1;
            break;
            case 'v' : FlagDrive   = 0;
            break;
            case 'a' : FlagDrive   = 1;
            break;
            case 's' : FlagDrive   = 2;
            break;
            case 'r' : FlagDrive   = 3;
            break;
            case 'D' :

              Index  = Index + 1;
              Option = argv[ Number ][ Index ];

              if ( Option == '0' ) FlagDebug = VASY_DEBUG_LEVEL0;
              else
              if ( Option == '1' ) FlagDebug = VASY_DEBUG_LEVEL1;
              else
              if ( Option == '2' ) FlagDebug = VASY_DEBUG_LEVEL2;
              else
              VasyUsage();
            break;

            case 'O' :
              FlagOption = 1;
            break;

            default  : VasyUsage();
          }
        }
      }

      FlagDebug  = 0;
      FlagOption = 0;
    }
    else
    if ( InputFileName  == (char *)0 ) InputFileName  = argv[ Number ];
    else
    if ( OutputFileName == (char *)0 ) OutputFileName = argv[ Number ];
    else
    VasyUsage();
  }

  if ( InputFileName == (char *)0 ) VasyUsage();

  InputFileName = autbasename( InputFileName, VPN_IN );

  if ( OutputFileName == (char *)0 )
  {
    OutputFileName = autallocblock( strlen( InputFileName ) + 10 );
    sprintf( OutputFileName, "%s_v", InputFileName );
  }

  if ( IsVasyDebugStatistics() ) VasyDebugStartChrono(0);
  
  fprintf( stdout, "\t--> Run VHDL Compiler\n" );
  fprintf( stdout, "\t--> Compile file %s\n", InputFileName );

  VpnFigure = VasyParseVpnFig( InputFileName );

  if ( FlagVerbose )
  {
    fprintf( stdout, "\t--> Preliminary Analysis\n" );
  }

  VasyPreAnalysisVpnFig( VpnFigure );

  if ( FlagVerbose )
  {
    fprintf( stdout, "\t--> In-Line function\n" );
  }

  VasyFuncVpnFig( VpnFigure );

  if ( FlagVerbose )
  {
    fprintf( stdout, "\t--> Elaboration\n" );
  }

  VasyElaborateVpnFig( VpnFigure );

  if ( FlagVerbose )
  {
    fprintf( stdout, "\t--> Petri Net Reduction\n" );
  }

  VasyReduceVpnFig( VpnFigure );

  if ( FlagVerbose )
  {
    fprintf( stdout, "\t--> Analysis and Synthesis\n" );
  }
  
  RtlFigure = VasyAnalysisVpnFig( VpnFigure );

  if ( FlagVerbose )
  {
    fprintf( stdout, "\t--> Simplify RTL figure\n" );
  }
  
  VasySimplifyRtlFig( RtlFigure );

  if ( FlagDrive == 0 )
  {
    fprintf( stdout, "\t--> Drive Verilog file %s\n", OutputFileName );

    VasyDriveVerilogRtlFig( RtlFigure, OutputFileName );
  }
  else
  if ( FlagDrive == 1 )
  {
    fprintf( stdout, "\t--> Drive Alliance file %s\n", OutputFileName );

    VasyDriveAllianceRtlFig( RtlFigure, OutputFileName );
  }
  else
  if ( FlagDrive == 2 )
  {
    fprintf( stdout, "\t--> Drive Standard VHDL file %s\n", OutputFileName );

    VasyDriveSynopsysRtlFig( RtlFigure, OutputFileName, FlagStdLogic, FlagInitial );
  }
  else
  {
    fprintf( stdout, "\t--> Drive Rtl file %s\n", OutputFileName );

    VasyDriveRtlRtlFig( RtlFigure, OutputFileName );
  }

  if ( IsVasyDebugStatistics() )
  {
    VasyPrintf( stdout, "--> Total time : %ld sec\n", VasyDebugReadChrono(0) );
  }

  fprintf( stdout, "\n" );

  return( 0 );
}
