/*------------------------------------------------------------\
|                                                             |
| Tool    :                     VASY                          |
|                                                             |
| File    :                  vasy_parse.c                     |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include VEX_H
# include VBH_H
# include VBL_H
# include VPN_H
# include VTL_H

# include "vasy_shared.h"
# include "vasy_parse.h"
# include "vasy_debug.h"
# include "vasy_loop.h"
# include "vasy_array.h"
# include "vasy_error.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Private variables                      |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       VasyParseVpnFig                       |
|                                                             |
\------------------------------------------------------------*/

vpnfig_list *VasyParseVpnFig( FileName )

  char *FileName;
{
  vbfig_list    *VbhFigure;
  vpnfig_list   *VpnFigure;
  vpnproc_list  *VpnProc;
  vpnproc_list  *DelProc;
  vpnproc_list  *FirstProc;
  vpntrans_list *VpnTrans;
  vpntrans_list *AsgTrans;
  vpntrans_list *FirstTrans;
  vpntrans_list *LastTrans;
  vpntrans_list *WaitTrans;
  vpnplace_list *VpnPlace;
  vpnplace_list *AsgPlace;
  vpnplace_list *LastPlace;
  vpnarc        *VpnArc;
  vpnact_list   *VpnAction;
  vpndecl_list  *VpnDeclar;
  vpnsym        *VpnSymbol;
  authtable     *HashSens;
  authtable     *HashDef;
  authelem      *Element;
  chain_list    *ScanChain;
  chain_list    *SensList;
  chain_list    *DefList;
  vexexpr       *VexAtom;
  vexexpr       *VexExpr;
  int            Type;

  if ( IsVasyDebugLevel0() )
  {
    VasyPrintf( stdout, "--> VasyParseVpnFig %s\n", FileName );
  }

  if ( IsVasyDebugStatistics() )
  {
    VasyDebugStartChrono(1);
  }

  VpnFigure = addvpnfig( FileName );
  VbhFigure = getvbfig( FileName, VPN_IN );

  VasyUnrollLoopVbhFig( VbhFigure );
  VasyExpandArrayVbhFig( VbhFigure );

  if ( IsVasyDebugStatistics() )
  {
    VasyPrintf( stdout, "--> Compilation : %ld sec\n", VasyDebugReadChrono(1) );
    VasyDebugStartChrono(1);
  }

  VvhVbh2Vpn( VbhFigure, VpnFigure, 1 );
  vbh_frevbfig( VbhFigure );

  if ( IsVasyDebugStatistics() )
  {
    VasyPrintf( stdout, "--> Translation : %ld sec\n", VasyDebugReadChrono(1) );
  }

  if ( VpnFigure->FUNCTION != (vpnfunc_list *)0 )
  {
    VasyErrorLine( VASY_NOT_YET_IMPLEMENTED_ERROR, VpnFigure->FUNCTION->LINE, "Function" );
  }

  HashSens  = createauthtable( 100 );
  HashDef   = createauthtable( 1000 );

  SensList  = (chain_list    *)0;
  DefList   = (chain_list    *)0;
  LastPlace = (vpnplace_list *)0;
  LastTrans = (vpntrans_list *)0;
  WaitTrans = (vpntrans_list *)0;
  FirstProc = (vpnproc_list  *)0;

  VpnProc = VpnFigure->PROCESS;

  while ( VpnProc != (vpnproc_list *)0 )
  {
    if ( IsVpnProcGuarded( VpnProc ) )
    {
      VpnTrans = VpnProc->FIRST;
      VpnArc   = GetVpnArc( VpnTrans->PLACE_IN );
      VpnPlace = GetVpnArcSourcePlace( VpnArc );
      VpnPlace->TYPE  = VPN_PLACE_ASSIGN;
      VpnPlace->TOKEN = 0;
      VpnTrans->TYPE  = VPN_TRANS_IMMEDIATE;

      VpnArc = GetVpnArc( VpnPlace->TRANS_IN );
      delvpnarc( VpnFigure, VpnArc );

      if  ( WaitTrans != (vpntrans_list *)0 )
      {
        VpnArc = addvpnarctrans( VpnFigure, WaitTrans, VpnPlace );
      }

      for ( ScanChain  = VpnProc->DEF_DECL;
            ScanChain != (chain_list *)0;
            ScanChain  = ScanChain->NEXT )
      {
        VpnDeclar = (vpndecl_list *)ScanChain->DATA;
        Element   = searchauthelem( HashDef, (char *)VpnDeclar );

        if ( Element == (authelem *)0 )
        {
          addauthelem( HashDef, (char *)VpnDeclar, 0 );
          DefList = addchain( DefList, (void *)VpnDeclar );
        }
      }

      freechain( VpnProc->DEF_DECL );
      VpnProc->DEF_DECL = (chain_list *)0;

      for ( VpnTrans  = VpnProc->TRANS;
            VpnTrans != (vpntrans_list *)0;
            VpnTrans  = VpnTrans->NEXT )
      {
        if ( VpnTrans->TYPE == VPN_TRANS_INF_WAIT ) break;
      }

      for ( ScanChain  = VpnTrans->WAIT_SYM;
            ScanChain != (chain_list *)0;
            ScanChain  = ScanChain->NEXT )
      {
        VpnSymbol = (vpnsym *)ScanChain->DATA;
        Element   = searchauthelem( HashSens, (char *)VpnSymbol );

        if ( Element == (authelem *)0 )
        {
          addauthelem( HashSens, (char *)VpnSymbol, 0 );
          SensList = addchain( SensList, (void *)VpnSymbol );
        }
      }

      VpnTrans->TYPE = VPN_TRANS_IMMEDIATE;
      freechain( VpnTrans->WAIT_SYM );
      VpnTrans->WAIT_SYM = (chain_list *)0;

      VpnArc   = GetVpnArc( VpnTrans->PLACE_IN );
      VpnPlace = GetVpnArcSourcePlace( VpnArc );
      VpnPlace->TYPE = VPN_PLACE_ASSIGN;

      WaitTrans = VpnTrans;

      VpnTrans = VpnProc->TRANS;

      while ( VpnTrans->NEXT != (vpntrans_list *)0 )
      {
        VpnTrans = VpnTrans->NEXT;
      }

      VpnPlace = VpnProc->PLACE;

      while ( VpnPlace->NEXT != (vpnplace_list *)0 )
      {
        VpnPlace = VpnPlace->NEXT;
      }

      if ( FirstProc != (vpnproc_list *)0 )
      {
        LastTrans->NEXT      = VpnProc->TRANS;
        VpnProc->TRANS->PREV = &LastTrans->NEXT;
        LastTrans            = VpnTrans;
        VpnProc->TRANS       = (vpntrans_list *)0;

        LastPlace->NEXT      = VpnProc->PLACE;
        VpnProc->PLACE->PREV = &LastPlace->NEXT;
        LastPlace            = VpnPlace;
        VpnProc->PLACE       = (vpnplace_list *)0;

        DelProc = VpnProc;
        VpnProc = VpnProc->NEXT;

        delvpnproc( VpnFigure, DelProc );
      }
      else
      {
        FirstProc = VpnProc;
        LastPlace = VpnPlace;
        LastTrans = VpnTrans;

        freechain( FirstProc->DEF_DECL );
        FirstProc->DEF_DECL = (chain_list *)0;

        VpnProc = VpnProc->NEXT;
      }
    }
    else
    {
      VpnProc = VpnProc->NEXT;
    }
  }

  if ( FirstProc != (vpnproc_list *)0 )
  {
    FirstTrans = FirstProc->FIRST;
    VpnArc     = GetVpnArc( FirstTrans->PLACE_IN );
    VpnPlace   = GetVpnArcSourcePlace( VpnArc );
    VpnArc     = addvpnarctrans( VpnFigure, WaitTrans, VpnPlace );

    VpnPlace->TYPE  = VPN_PLACE_PROCESS;
    VpnPlace->TOKEN = 1;

    FirstProc->TYPE     = VPN_PROCESS_SEQUENTIAL_MASK;
    FirstProc->DEF_DECL = DefList;

    WaitTrans->WAIT_SYM = SensList;
    WaitTrans->TYPE     = VPN_TRANS_INF_WAIT;

    AsgPlace       = addvpnprocplace( VpnFigure, FirstProc, "vasy_asg_bus" );
    AsgPlace->TYPE = VPN_PLACE_ASSIGN;
    AsgTrans       = addvpnproctrans( VpnFigure, FirstProc, "vasy_asg_bus" );
    AsgTrans->TYPE = VPN_TRANS_IMMEDIATE;
    VpnArc         = addvpnarcplace( VpnFigure, AsgPlace, AsgTrans );

    VpnArc    = GetVpnArc( FirstTrans->PLACE_OUT );
    VpnPlace  = GetVpnArcTargetPlace( VpnArc );
    delvpnarc( VpnFigure, VpnArc );

    VpnArc = addvpnarctrans( VpnFigure, FirstTrans, AsgPlace );
    VpnArc = addvpnarctrans( VpnFigure, AsgTrans  , VpnPlace );

    for ( Type = 0; Type < VPN_MAX_DECLAR_TYPE; Type++ )
    {
      for ( VpnDeclar = VpnFigure->DECLAR[ Type ];
            VpnDeclar != (vpndecl_list *)0;
            VpnDeclar  = VpnDeclar->NEXT )
      {
        if ( VpnDeclar->KIND == VPN_KIND_BUS )
        {
          VexAtom   = dupvexexpr( VpnDeclar->VEX_ATOM );
          VexExpr   = createvexatomveclit( VEX_TRISTATE, VexAtom->WIDTH );
          VpnAction = addvpnactasg( VpnFigure, AsgTrans, VexAtom, VexExpr );

          AsgTrans->TYPE = VPN_TRANS_ACT_EXEC;
        }
      }
    }
  }

  destroyauthtable( HashSens  );
  destroyauthtable( HashDef   );

  for ( VpnProc  = VpnFigure->PROCESS;
        VpnProc != (vpnproc_list *)0;
        VpnProc  = VpnProc->NEXT )
  {
    if ( IsVasyDebugLevel1() )
    {
      VasyPrintf( stdout, "+++ VasyAddProcInfo %s\n", VpnProc->NAME );
    }

    VasyAddVpnProcInfo( VpnProc );
  }

  if ( IsVasyDebugLevel1() )
  {
    VasyDebugSaveVpnFig( VpnFigure );
  }

  if ( IsVasyDebugLevel0() )
  {
    VasyPrintf( stdout, "<-- VasyParseVpnFig %s\n\n", FileName );
  }

  return( VpnFigure );
}
