/*------------------------------------------------------------\
|                                                             |
| Tool    :                     VASY                          |
|                                                             |
| File    :                  vasy_reduce.c                    |
|                                                             |
| Author  :                 Jacomme Ludovic                   |
|                                                             |
| Date    :                   25.08.97                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include VEX_H
# include VPN_H
# include VTL_H
# include RTN_H

# include "vasy_shared.h"
# include "vasy_redact.h"
# include "vasy_redinst.h"
# include "vasy_redwait.h"
# include "vasy_reduce.h"

# include "vasy_debug.h"
# include "vasy_error.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         VasyReduceVpnFig                    |
|                                                             |
\------------------------------------------------------------*/

void VasyReduceVpnFig( VpnFigure )

  vpnfig_list *VpnFigure;
{
  vpnproc_list *VpnProc;
  int           Reduced;

  if ( IsVasyDebugLevel0() )
  {
    VasyPrintf( stdout, "--> VasyReduceVpnFig %s\n", VpnFigure->NAME );
    VasyDebugSaveVpnFig( VpnFigure );
  }
/*
**  Partial reduction instruction by instruction
*/
  if ( IsVasyDebugStatistics() )
  {
    VasyPrintf( stdout, "--> Process     : %ld\n", GetVpnNumProc( VpnFigure ) );
    VasyPrintf( stdout, "--> Transition  : %ld\n", GetVpnNumTrans( VpnFigure ) );
    VasyPrintf( stdout, "--> Place       : %ld\n", GetVpnNumPlace( VpnFigure ) );
    VasyDebugStartChrono(1);
  }

  for ( VpnProc  = VpnFigure->PROCESS;
        VpnProc != (vpnproc_list *)0;
        VpnProc  = VpnProc->NEXT )
  {
    if ( IsVasyVpnProcToAnalyse( VpnProc ) )
    {
      VasyRedInstVpnProc( VpnFigure, VpnProc );
    }
  }

  if ( IsVasyDebugStatistics() )
  {
    VasyPrintf( stdout, "--> Local reduction : %ld sec\n", VasyDebugReadChrono(1) );
    VasyPrintf( stdout, "--> Process     : %ld\n", GetVpnNumProc( VpnFigure ) );
    VasyPrintf( stdout, "--> Transition  : %ld\n", GetVpnNumTrans( VpnFigure ) );
    VasyPrintf( stdout, "--> Place       : %ld\n", GetVpnNumPlace( VpnFigure ) );
  }

  if ( IsVasyDebugLevel0() )
  {
    VasyDebugSaveVpnFig( VpnFigure );
  }
/*
**  Global reduction wait to wait
*/
  if ( IsVasyDebugStatistics() ) VasyDebugStartChrono(1);

  Reduced = 0; 

  for ( VpnProc  = VpnFigure->PROCESS;
        VpnProc != (vpnproc_list *)0;
        VpnProc  = VpnProc->NEXT )
  {
    if ( IsVasyVpnProcToAnalyse( VpnProc ) )
    {
      Reduced += VasyRedWaitVpnProc( VpnFigure, VpnProc );
    }
  }

  if ( IsVasyDebugStatistics() )
  {
    VasyPrintf( stdout, "--> Process minimal : %d\n", Reduced );

    VasyPrintf( stdout, "--> Global reduction : %ld sec\n", VasyDebugReadChrono(1) );
    VasyPrintf( stdout, "--> Process          : %ld\n", GetVpnNumProc( VpnFigure ) );
    VasyPrintf( stdout, "--> Transition       : %ld\n", GetVpnNumTrans( VpnFigure ) );
    VasyPrintf( stdout, "--> Place            : %ld\n", GetVpnNumPlace( VpnFigure ) );
  }

  if ( IsVasyDebugLevel0() )
  {
    VasyDebugSaveVpnFig( VpnFigure );
    VasyPrintf( stdout, "<-- VasyReduceVpnFig %s\n", VpnFigure->NAME );
  }
}
