
/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: vbh_free.c						*/
/* date		: Aug 24 1992						*/
/* version	: v102							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: contains a specific function used to delete a list	*/
/*		  of behaviour data structures				*/
/*									*/
/* ###--------------------------------------------------------------### */

#include MUT_H 
#include AUT_H
#include VEX_H 
#include VBH_H 

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbfig						*/
/* description	: delete a list of BEFIG structures and all objects	*/
/*		  pointed by any os BEFIGs in the list			*/
/* called func.	: autfreeblock						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbfig (listvbfig)

struct vbfig *listvbfig;	/* list of vbfig to be deleted		*/

{
  struct vbfig *ptvbfig;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listvbfig != NULL)
  {
    vbh_frevbatr (listvbfig->BEATR);
    vbh_frevbaux (listvbfig->BEAUX);
    vbh_frevbcst (listvbfig->BECST);
    vbh_frevbpor (listvbfig->BEPOR);
    vbh_frevbpcs (listvbfig->BEPCS);
    vbh_frevbfun (listvbfig->BEFUN);
    vbh_frevbtyp (listvbfig->BETYP);

    destroyauthtable( listvbfig->HASH_TYPE );

    ptvbfig   = listvbfig;
    listvbfig = listvbfig->NEXT;
    autfreeblock (ptvbfig);
  }

}

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbcst						*/
/* description	: delete a list of BECST structures and all objects	*/
/*		  pointed by any os BECSTs in the list			*/
/* called func.	: autfreeheap						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbcst (listvbcst)

struct vbcst *listvbcst;	/* list of vbcst to be deleted		*/

{
  struct vbcst *ptvbcst;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listvbcst != NULL)
    {
    freevexexpr (listvbcst->VEX);
    freevexexpr (listvbcst->TARGET);

    ptvbcst   = listvbcst;
    listvbcst = listvbcst->NEXT;
    autfreeheap (ptvbcst, sizeof(struct vbcst));
    }

}

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbatr						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: autfreeheap						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbatr (listvbatr)

struct vbatr *listvbatr;	/* list of vbatr to be deleted		*/

{
  struct vbatr *ptvbatr;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listvbatr != NULL)
    {
    ptvbatr   = listvbatr;
    listvbatr = listvbatr->NEXT;
    autfreeheap (ptvbatr, sizeof(struct vbatr));
    }

}

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbaux						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: autfreeheap						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbaux (listvbaux)

struct vbaux *listvbaux;	/* list of vbaux to be deleted		*/

{
  struct vbaux *ptvbaux;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listvbaux != NULL)
    {
    freevexexpr (listvbaux->VEX);
    freevexexpr (listvbaux->TARGET);

    ptvbaux   = listvbaux;
    listvbaux = listvbaux->NEXT;
    autfreeheap (ptvbaux, sizeof(struct vbaux));
    }

}

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbarg						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: autfreeheap						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbarg (listvbarg)

struct vbarg *listvbarg;	/* list of vbarg to be deleted		*/

{
  struct vbarg *ptvbarg;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, delete the object itself	*/
	/* ###------------------------------------------------------### */

  while (listvbarg != NULL)
  {
    freevexexpr (listvbarg->TARGET);

    ptvbarg   = listvbarg;
    listvbarg = listvbarg->NEXT;
    autfreeheap (ptvbarg, sizeof(struct vbarg));
  }
}

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbpor						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: autfreeheap						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbpor (listvbpor)

struct vbpor *listvbpor;	/* list of vbpor to be deleted		*/

{
  struct vbpor *ptvbpor;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, delete the object itself	*/
	/* ###------------------------------------------------------### */

  while (listvbpor != NULL)
  {
    freevexexpr (listvbpor->TARGET);

    ptvbpor   = listvbpor;
    listvbpor = listvbpor->NEXT;
    autfreeheap (ptvbpor, sizeof(struct vbpor));
  }
}

void vbh_frevbifs( ScanIfs )

  vbifs_list *ScanIfs;
{
  freevexexpr( ScanIfs->CND );
  vbh_frevbins( ScanIfs->CNDTRUE  );
  vbh_frevbins( ScanIfs->CNDFALSE );

  autfreeheap( ScanIfs, sizeof(struct vbifs) );
}

void vbh_frevbfor( ScanFor )

  vbfor_list *ScanFor;
{
  freevexexpr( ScanFor->VARIABLE );
  freevexexpr( ScanFor->LEFT     );
  freevexexpr( ScanFor->RIGHT    );

  vbh_frevbins( ScanFor->INSTRUCTION );
  autfreeheap( ScanFor, sizeof(struct vbfor) );
}

void vbh_frevblop( ScanLop )

  vblop_list *ScanLop;
{
  vbh_frevbins( ScanLop->INSTRUCTION );
  autfreeheap( ScanLop, sizeof(struct vblop) );
}

void vbh_frevbwhi( ScanWhi )

  vbwhi_list *ScanWhi;
{
  freevexexpr( ScanWhi->CND );
  vbh_frevbins( ScanWhi->INSTRUCTION );
  autfreeheap( ScanWhi, sizeof(struct vbwhi) );
}

void vbh_frevbext( ScanExit )

  vbext_list *ScanExit;
{
  freevexexpr( ScanExit->CND );
  autfreeheap( ScanExit, sizeof(struct vbext) );
}

void vbh_frevbnxt( ScanNext )

  vbnxt_list *ScanNext;
{
  freevexexpr( ScanNext->CND );
  autfreeheap( ScanNext, sizeof(struct vbnxt) );
}

void vbh_frevbret( ScanReturn )

  vbret_list *ScanReturn;
{
  freevexexpr( ScanReturn->RET );
  autfreeheap( ScanReturn, sizeof(struct vbret) );
}

void vbh_frevbcal( ScanCall )

  vbcal_list *ScanCall;
{
  freevexexpr( ScanCall->CALL );
  autfreeheap( ScanCall, sizeof(struct vbcal) );
}

void vbh_frevbasg( ScanAsg )

  vbasg_list *ScanAsg;
{
  freevexexpr( ScanAsg->TARGET );
  freevexexpr( ScanAsg->VEX    );

  autfreeheap( ScanAsg, sizeof(struct vbasg) );
}

void vbh_frevbvar( ScanVar )

  vbvar_list *ScanVar;
{
  freevexexpr( ScanVar->TARGET );
  freevexexpr( ScanVar->VEX    );

  autfreeheap( ScanVar, sizeof(struct vbvar) );
}

void vbh_frevbwas( ScanWas )

  vbwas_list *ScanWas;
{
  freechain( ScanWas->SEN    );
  freevexexpr( ScanWas->CND );

  autfreeheap( ScanWas, sizeof(struct vbwas) );
}

void vbh_frevbcas( ScanCas )

  vbcas_list *ScanCas;
{
  vbcho_list *ScanCho;
  int         Offset;

  freevexexpr( ScanCas->VEX );

  for ( Offset = 0; Offset < ScanCas->SIZE; Offset++ )
  {
    ScanCho = &ScanCas->CHOICE[ Offset ];
    vbh_frevbins( ScanCho->INSTRUCTION );
  }

  autfreeblock( ScanCas->CHOICE );
  autfreeheap( ScanCas, sizeof(struct vbcas) );
}

void vbh_frevbagr( ScanAgr )

  vbagr_list *ScanAgr;
{
  vbh_frevbins( ScanAgr->ASSIGN );
  autfreeheap( ScanAgr, sizeof(struct vbagr) );
}

void vbh_frevbins( HeadType )

  ptype_list *HeadType;
{
  ptype_list *ScanType;

  for ( ScanType  = HeadType;
        ScanType != (ptype_list *)0;
        ScanType  = ScanType->NEXT )
  {
    switch ( ScanType->TYPE )
    {
      case VBH_BEIFS : vbh_frevbifs( ScanType->DATA );
      break;
  
      case VBH_BEASG : vbh_frevbasg( ScanType->DATA );
      break;
  
      case VBH_BEVAR : vbh_frevbvar( ScanType->DATA );
      break;
  
      case VBH_BECAS : vbh_frevbcas( ScanType->DATA );
      break;
  
      case VBH_BEWAS : vbh_frevbwas( ScanType->DATA );
      break;
  
      case VBH_BEWHI : vbh_frevbwhi( ScanType->DATA );
      break;

      case VBH_BELOP : vbh_frevblop( ScanType->DATA );
      break;
  
      case VBH_BEFOR : vbh_frevbfor( ScanType->DATA );
      break;
  
      case VBH_BENXT : vbh_frevbnxt( ScanType->DATA );
      break;
  
      case VBH_BEEXT : vbh_frevbext( ScanType->DATA );
      break;

      case VBH_BECAL : vbh_frevbcal( ScanType->DATA );
      break;

      case VBH_BERET : vbh_frevbret( ScanType->DATA );
      break;

      case VBH_BEAGR : vbh_frevbagr( ScanType->DATA );
    }
  }

  freeptype( HeadType );
}

/* ###--------------------------------------------------------------### */
/* function     : vbh_frevbpcs                                          */
/* description  : delete a list of BEPCS structures and all objects     */
/*                pointed by any os BEPCSs in the list                  */
/* called func. : autfreeheap                                          */
/* ###--------------------------------------------------------------### */

void vbh_frevbpcs (listvbpcs)

struct vbpcs *listvbpcs;        /* list of vbpcs to be deleted          */

{
  struct vbpcs *ptvbpcs;

        /* ###------------------------------------------------------### */
        /*    for each object of the list, first delete pointed objects */
        /* then, delete the object itself                               */
        /* ###------------------------------------------------------### */

  while (listvbpcs != NULL)
  {
    freechain( listvbpcs->SENSITIVITY );

    vbh_frevbins( listvbpcs->VARIABLE );
    vbh_frevbins( listvbpcs->INSTRUCTION );
 
    ptvbpcs   = listvbpcs;
    listvbpcs = listvbpcs->NEXT;
    autfreeheap (ptvbpcs, sizeof(struct vbpcs) );
  }
}

/* ###--------------------------------------------------------------### */
/* function     : vbh_frevbfun                                          */
/* description  : delete a list of BEFUN structures and all objects     */
/*                pointed by any os BEFUNs in the list                  */
/* called func. : autfreeheap                                          */
/* ###--------------------------------------------------------------### */

void vbh_frevbfun (listvbfun)

struct vbfun *listvbfun;        /* list of vbfun to be deleted          */

{
  struct vbfun *ptvbfun;

        /* ###------------------------------------------------------### */
        /*    for each object of the list, first delete pointed objects */
        /* then, delete the object itself                               */
        /* ###------------------------------------------------------### */

  while (listvbfun != NULL)
  {
    vbh_frevbarg( listvbfun->RETURN   );
    vbh_frevbarg( listvbfun->ARGUMENT );

    vbh_frevbins( listvbfun->VARIABLE );
    vbh_frevbins( listvbfun->INSTRUCTION );
 
    ptvbfun   = listvbfun;
    listvbfun = listvbfun->NEXT;
    autfreeheap (ptvbfun, sizeof(struct vbfun) );
  }
}

/* ###--------------------------------------------------------------### */
/* function	: vbh_frevbtyp						*/
/* description	: delete a list of BETYP structures and all objects	*/
/*		  pointed by any os BETYPs in the list			*/
/* called func.	: autfreeheap						*/
/* ###--------------------------------------------------------------### */

void vbh_frevbtyp (listvbtyp)

struct vbtyp *listvbtyp;	/* list of vbtyp to be deleted		*/

{
  struct vbtyp *ptvbtyp;

	/* ###------------------------------------------------------### */
	/*    for each object of the list delete the object itself	*/
	/* ###------------------------------------------------------### */

  while (listvbtyp != NULL)
    {
    ptvbtyp   = listvbtyp;
    listvbtyp = listvbtyp->NEXT;
    autfreeheap (ptvbtyp, sizeof(struct vbtyp) );
    }

}
