/*------------------------------------------------------------\
|                                                             |
| Tool    :                   XSCH                            |
|                                                             |
| File    :                   View.c                          |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                   01.06.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <math.h>
# include <signal.h>
# include <Xm/Xm.h>
 
# include MUT_H
# include AUT_H
# include MLO_H
# include SCL_H
# include XSB_H
# include XSC_H
# include XMX_H

# include "XMX_view.h" 

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

    static long         XschGraphicX1;
    static long         XschGraphicX2;
    static long         XschGraphicY1;
    static long         XschGraphicY2;
    static XFontStruct *XschTextFont;
    static GC           XschTextGC;
    static GC           XschDrawGC;
    static char         XschInterrupt = 0;

    static xschview_list *XschHeadView = (xschview_list *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       Alloc Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         XschAllocView                       |
|                                                             |
\------------------------------------------------------------*/

xschview_list *XschAllocView()
{
  return( (xschview_list *)autallocheap( sizeof( xschview_list ) ) );
}

/*------------------------------------------------------------\
|                                                             |
|                        Free Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          XschFreeView                       |
|                                                             |
\------------------------------------------------------------*/

void XschFreeView( FreeView )

   xschview_list *FreeView;
{
  autfreeheap( (char *)FreeView, sizeof( xschview_list ) );
}

/*------------------------------------------------------------\
|                                                             |
|                        XschAddViewLater                     |
|                                                             |
\------------------------------------------------------------*/

void XschAddViewLater( Obj )

   xschobj_list *Obj;
{
  xschview_list *View;

  autbegin();

  View = XschAllocView();

  View->OBJECT  = Obj;
  View->NEXT    = XschHeadView;
  XschHeadView  = View;

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                        XschDelView                          |
|                                                             |
\------------------------------------------------------------*/

void XschDelView()
{
  xschview_list *DelView;
  xschview_list *View;

  autbegin();

  View          = XschHeadView;
  XschHeadView = (xschview_list *)NULL;

  while( View != (xschview_list *)NULL )
  {
    DelView  = View;
    View     = View->NEXT;

    XschFreeView( DelView );
  }

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                     XschInterruptDisplay                    |
|                                                             |
\------------------------------------------------------------*/

void XschInterruptDisplay()
{
  XEvent       Event;
  KeySym       Key;
  char         Text;

  autbegin();

  signal( SIGALRM, XschInterruptDisplay );
  alarm( 1 );

  if ( XCheckTypedEvent( XschGraphicDisplay, KeyPress, &Event ) )
  {
    XLookupString( &Event.xkey, &Text, 1, &Key, 0 );

    if ( Text == '\003' ) 
    {
      XschInterrupt = 1;

      XBell( XschGraphicDisplay, 0 );
    }
  }

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                     XschFlushEventDisplay                   |
|                                                             |
\------------------------------------------------------------*/

void XschFlushEventDisplay()
{
  XEvent Event;
  
  autbegin();

  while ( XCheckTypedEvent( XschGraphicDisplay, KeyPress, &Event ) );

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                       XschComputeHexagon                    |
|                                                             |
\------------------------------------------------------------*/

char XschComputeHexagon( X1, Y1, X2, Y2, Points )

  long    X1;
  long    Y1;
  long    X2;
  long    Y2;
  XPoint *Points;
{
  long StepX;
  long StepY;

  StepY = ( Y2 - Y1 ) >> 1;
  StepX = StepY >> 1;

  if ( ( StepY > 0 ) &&
       ( StepX > 0 ) )
  {
    Points[ 0 ].x = X1 + StepX;
    Points[ 0 ].y = Y1;
    Points[ 1 ].x = X1;
    Points[ 1 ].y = Y1 + StepY;
    Points[ 2 ].x = X1 + StepX;
    Points[ 2 ].y = Y2;
    Points[ 3 ].x = X2 - StepX;
    Points[ 3 ].y = Y2;
    Points[ 4 ].x = X2;
    Points[ 4 ].y = Y1 + StepY;
    Points[ 5 ].x = X2 - StepX;
    Points[ 5 ].y = Y1;
    Points[ 6 ].x = X1 + StepX;
    Points[ 6 ].y = Y1;

    return( XSCH_TRUE );
  }

  return( XSCH_FALSE );
}

/*------------------------------------------------------------\
|                                                             |
|                       XschDisplayArrow                      |
|                                                             |
\------------------------------------------------------------*/

char XschComputeArrow( X1, Y1, X2, Y2, Points )

  long    X1;
  long    Y1;
  long    X2;
  long    Y2;
  XPoint *Points;
{
  float LineX;
  float LineY;
  float HalfX;
  float HalfY;
  float CrossX;
  float CrossY;
  float Norm;

  LineX = (float)( X2 - X1 );
  LineY = (float)( Y2 - Y1 );

  if ( ( LineX != 0.0 ) ||
       ( LineY != 0.0 ) )
  {
    Norm    = sqrt( ( LineX * LineX ) + ( LineY * LineY ) );
    Norm    = Norm  / XschUnitGridStep;
    LineX   = LineX / Norm;
    LineY   = LineY / Norm;
    HalfX   = LineX / 2;
    HalfY   = LineY / 2;
    CrossX  = - HalfY;
    CrossY  =   HalfX;

    Points[ 0 ].x = X1 + LineX;
    Points[ 0 ].y = Y1 + LineY;
    Points[ 1 ].x = Points[ 0 ].x + LineX + HalfX + CrossX;
    Points[ 1 ].y = Points[ 0 ].y + LineY + HalfY + CrossY;
    Points[ 2 ].x = Points[ 0 ].x + LineX + HalfX - CrossX;
    Points[ 2 ].y = Points[ 0 ].y + LineY + HalfY - CrossY;

    return( XSCH_TRUE );
  }

  return( XSCH_FALSE );
}

/*------------------------------------------------------------\
|                                                             |
|                     XschDisplayOneObject                    |
|                                                             |
\------------------------------------------------------------*/

void XschDisplayOneObject( Obj )

   xschobj_list *Obj;
{
  XPoint  Points[ 7 ];
  int     Index;
  long    X1r;
  long    X2r;
  long    Y1r;
  long    Y2r;
  long    DeltaX;
  long    DeltaY;

  if ( IsXschLineLeft( Obj ) )
  {
    X1r = Obj->X + Obj->DX;
    X2r = Obj->X;
  }
  else
  {
    X1r = Obj->X;
    X2r = Obj->X + Obj->DX;
  }

  if ( IsXschLineDown( Obj ) )
  {
    Y1r  = Obj->Y + Obj->DY;
    Y2r  = Obj->Y;
  }
  else
  {
    Y1r  = Obj->Y;
    Y2r  = Obj->Y + Obj->DY;
  }

  X1r  = (float)( X1r ) * XschUnitGridStep;
  Y1r  = (float)( Y1r ) * XschUnitGridStep;
  X2r  = (float)( X2r ) * XschUnitGridStep;
  Y2r  = (float)( Y2r ) * XschUnitGridStep;

  X1r  = ( X1r / XSCH_UNIT ) - XschPixelGridX;
  X2r  = ( X2r / XSCH_UNIT ) - XschPixelGridX;
  Y1r  = ( Y1r / XSCH_UNIT ) - XschPixelGridY;
  Y2r  = ( Y2r / XSCH_UNIT ) - XschPixelGridY;

  DeltaX = X2r - X1r;
  DeltaY = Y2r - Y1r;

  if ( DeltaX <= 0 ) DeltaX = 1;
  if ( DeltaY <= 0 ) DeltaY = 1;

  switch ( GetXschObjectType( Obj ) )
  {
    case XSCH_OBJECT_LINE :

      XDrawLine( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschDrawGC,
                 X1r, XschGraphicDy - Y1r,
                 X2r, XschGraphicDy - Y2r );
    break;

    case XSCH_OBJECT_ARROW :

      XDrawLine( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschDrawGC,
                 X1r, XschGraphicDy - Y1r,
                 X2r, XschGraphicDy - Y2r );

      if ( XschComputeArrow( X1r, Y1r, X2r, Y2r, Points ) )
      {
        Points[ 0 ].y = XschGraphicDy - Points[ 0 ].y;
        Points[ 1 ].y = XschGraphicDy - Points[ 1 ].y;
        Points[ 2 ].y = XschGraphicDy - Points[ 2 ].y;

        XFillPolygon( XschGraphicDisplay,
                      XschGraphicPixmap, 
                      XschDrawGC, Points, 3, 
                      Convex,
                      CoordModeOrigin );
      }

    break;

    case XSCH_OBJECT_HEXAGON :

      if ( XschComputeHexagon( X1r, Y1r, X2r, Y2r, Points ) )
      {
        for ( Index = 0; Index < 7; Index++ )
        {
           Points[ Index ].y = XschGraphicDy - Points[ Index ].y;
        }

        XDrawLines( XschGraphicDisplay,
                    XschGraphicPixmap,
                    XschDrawGC, Points, 7,
                    CoordModeOrigin );
      }

    break;

    case XSCH_OBJECT_CIRCLE :

     XDrawArc( XschGraphicDisplay,
               XschGraphicPixmap,
               XschDrawGC,
               X1r, XschGraphicDy - Y2r, 
               DeltaX, DeltaY, 0, 23040 );
    break;

    case XSCH_OBJECT_RECTANGLE :

      XDrawRectangle( XschGraphicDisplay,
                      XschGraphicPixmap,
                      XschDrawGC,
                      X1r, XschGraphicDy - Y2r,
                      DeltaX, DeltaY );

    break;

    case XSCH_OBJECT_TRIANGLE :

      XDrawLine( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschDrawGC,
                 X1r, XschGraphicDy - Y1r,
                 X2r, ( XschGraphicDy - Y1r - ( DeltaY >> 1 ) ) );

      XDrawLine( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschDrawGC,
                 X1r, XschGraphicDy - Y1r,
                 X1r, XschGraphicDy - Y2r );

      XDrawLine( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschDrawGC,
                 X1r, XschGraphicDy - Y2r,
                 X2r, ( XschGraphicDy - Y1r - ( DeltaY >> 1 ) ) );

    break;
  }
}

/*------------------------------------------------------------\
|                                                             |
|                     XschDisplayOneString                    |
|                                                             |
\------------------------------------------------------------*/

void XschDisplayOneString( Obj )

   xschobj_list *Obj;
{
  long    X1r;
  long    X2r;
  long    Y1r;
  long    Y2r;
  long    DeltaX;
  long    DeltaY;
  long    WidthText;
  long    HeightText;
  int     Length;

  if ( IsXschLineLeft( Obj ) )
  {
    X1r = Obj->X + Obj->DX;
    X2r = Obj->X;
  }
  else
  {
    X1r = Obj->X;
    X2r = Obj->X + Obj->DX;
  }

  if ( IsXschLineDown( Obj ) )
  {
    Y1r  = Obj->Y + Obj->DY;
    Y2r  = Obj->Y;
  }
  else
  {
    Y1r  = Obj->Y;
    Y2r  = Obj->Y + Obj->DY;
  }

  X1r  = (float)( X1r ) * XschUnitGridStep;
  Y1r  = (float)( Y1r ) * XschUnitGridStep;
  X2r  = (float)( X2r ) * XschUnitGridStep;
  Y2r  = (float)( Y2r ) * XschUnitGridStep;

  X1r  = ( X1r / XSCH_UNIT ) - XschPixelGridX;
  X2r  = ( X2r / XSCH_UNIT ) - XschPixelGridX;
  Y1r  = ( Y1r / XSCH_UNIT ) - XschPixelGridY;
  Y2r  = ( Y2r / XSCH_UNIT ) - XschPixelGridY;

  DeltaX = X2r - X1r;
  DeltaY = Y2r - Y1r;

  if ( DeltaX <= 0 ) DeltaX = 1;
  if ( DeltaY <= 0 ) DeltaY = 1;

  Length     = strlen( Obj->NAME );
  WidthText  = XTextWidth( XschTextFont, Obj->NAME, Length );
  HeightText = XschTextFont->ascent;

  if ( Obj->TYPE == XSCH_OBJECT_TEXT_RIGHT )
  {
    XDrawString( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschTextGC,
                 X1r,
                 XschGraphicDy - (( Y1r + Y2r - HeightText ) >> 1),
                 Obj->NAME, Length );
  }
  else
  if ( Obj->TYPE == XSCH_OBJECT_TEXT_LEFT )
  {
    XDrawString( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschTextGC,
                 X2r - WidthText,
                 XschGraphicDy - (( Y1r + Y2r - HeightText ) >> 1),
                 Obj->NAME, Length );
  }
  else
  {
    XDrawString( XschGraphicDisplay,
                 XschGraphicPixmap,
                 XschTextGC,
                 ( X2r + X1r - WidthText ) >> 1,
                 XschGraphicDy - (( Y1r + Y2r - HeightText ) >> 1),
                 Obj->NAME, Length );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                     XschDisplayFigure                       |
|                                                             |
\------------------------------------------------------------*/

void XschDisplayFigure( GraphicX1, GraphicY1, GraphicX2, GraphicY2 )

   long GraphicX1;
   long GraphicY1;
   long GraphicX2;
   long GraphicY2;
{
  xschview_list *View;
  xschobj_list  *Obj;
  long           X1;
  long           Y1;
  long           X2;
  long           Y2;
  int            Layer;

char DisplayText;
/*\
  char           DisplayTextConnector;
  char           DisplayTextSignal;
\*/

  autbegin();

  DisplayText = 1;
/*\
  DisplayTextConnector = XSCH_ACTIVE_NAME_TABLE[ XSCH_ACTIVE_NAME_IO     ];
  DisplayTextSignal    = XSCH_ACTIVE_NAME_TABLE[ XSCH_ACTIVE_NAME_VECTOR ];
\*/

  XschGraphicX1 = GraphicX1;
  XschGraphicX2 = GraphicX2;
  XschGraphicY1 = GraphicY1;
  XschGraphicY2 = GraphicY2;

  if ( XschUnitGridStep > 10.0 )
  {
    XschTextGC       = XschLargeTextGC;
    XschTextFont     = XschLargeTextFont;
  }
  else
  if ( XschUnitGridStep > 6.0 )
  {
    XschTextGC       = XschMediumTextGC;
    XschTextFont     = XschMediumTextFont;
  }
  else
  if ( XschUnitGridStep > 3.0 )
  {
    XschTextGC       = XschSmallTextGC;
    XschTextFont     = XschSmallTextFont;
  }
  else
  {
    DisplayText = XSCH_FORCE_DISPLAY;
  }

  XschClearGraphicWindow( GraphicX1, 
                           XschGraphicDy - GraphicY2,
                           GraphicX2 - GraphicX1,
                           GraphicY2 - GraphicY1 );

  if ( XschFigure == (xschfig_list *)NULL ) return;
  
  X1 = GraphicX1 + XschPixelGridX;
  X2 = GraphicX2 + XschPixelGridX;
  Y1 = GraphicY1 + XschPixelGridY;
  Y2 = GraphicY2 + XschPixelGridY;

  X1 = ( X1 / XschUnitGridStep );
  Y1 = ( Y1 / XschUnitGridStep );
  X2 = ( X2 / XschUnitGridStep );
  Y2 = ( Y2 / XschUnitGridStep );

  if ( X2 >= 0 ) { X2 = X2 + 1; }
  if ( Y2 >= 0 ) { Y2 = Y2 + 1; }
  if ( X1 <= 0 ) { X1 = X1 - 1; }
  if ( Y1 <= 0 ) { Y1 = Y1 - 1; }

  X1 = X1 * XSCH_UNIT;
  X2 = X2 * XSCH_UNIT;
  Y1 = Y1 * XSCH_UNIT;
  Y2 = Y2 * XSCH_UNIT;

  XschInterrupt = 0;

  signal( SIGALRM, XschInterruptDisplay );
  alarm( 1 );

  for ( Layer = 0; Layer < XSCH_MAX_LAYER; Layer++ )
  {
    if ( XschInterrupt ) break;

    if ( XSCH_ACTIVE_LAYER_TABLE[ Layer ] == 0 ) continue;

    XschDrawGC = XschLayerDrawGC[ Layer ];

    for ( Obj  = XschFigure->OBJECT[ Layer ];
          Obj != (xschobj_list *)NULL;
          Obj  = Obj->NEXT )
    {
      if ( XschInterrupt ) break;

      if ( XSCH_ACTIVE_LAYER_TABLE[ Obj->LAYER ] == 0 ) continue;
  
      if ( ( Obj->X           <= X2 ) &&
           ( Obj->Y           <= Y2 ) &&
           ( Obj->X + Obj->DX >= X1 ) && 
           ( Obj->Y + Obj->DY >= Y1 ) )
      {
        if ( ( IsXschAccept( Obj  ) ) ||
             ( IsXschConnect( Obj ) ) )
        {
          XschAddViewLater( Obj );
        }
        else
        {
          XschDisplayOneObject( Obj );

          if ( Obj->NAME != (char *)0 )
          {
            if ( DisplayText )
            {
              if ( ( ( IsXschSchCon( Obj                             ) && 
                       XSCH_ACTIVE_NAME_TABLE[ XSCH_CONNECTOR_NAME ] ) ) ||
                   ( ( IsXschSchBox( Obj                             ) && 
                       XSCH_ACTIVE_NAME_TABLE[ XSCH_INSTANCE_NAME  ] ) ) )
              {
                XschDisplayOneString( Obj );
              }
            }
          }
        }
      }
    }
  }

  for ( View  = XschHeadView;
        View != (xschview_list *)NULL;
        View  = View->NEXT )
  {
    Obj = View->OBJECT;

    if ( IsXschAccept( Obj ) )
    {
      XschDrawGC = XschAcceptDrawGC;
    } 
    else
    {
      XschDrawGC = XschConnectDrawGC;
    } 
    
    XschDisplayOneObject( Obj );

    if ( Obj->NAME != (char *)0 )
    {
/*\
      if ( ( IsXschConnector( Obj ) && DisplayTextConnector ) ||
           ( IsXschSignal( Obj )    && DisplayTextSignal    ) )
\*/
      {
        XschDisplayOneString( Obj );
      }
    }
  }

  XschDelView();

  alarm( 0 );
  XschFlushEventDisplay();

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                     XschDisplayObject                       |
|                                                             |
\------------------------------------------------------------*/

void XschDisplayObject( Obj )

   xschobj_list *Obj;
{
  xschobj_list *ScanObj;
  long          Xmin;
  long          Ymin;
  long          Xmax;
  long          Ymax;

  autbegin();

  Xmin = Obj->X;
  Ymin = Obj->Y;
  Xmax = Xmin + Obj->DX;
  Ymax = Ymin + Obj->DY;

  ScanObj = Obj->LINK;

  while ( ScanObj != Obj )
  {
    if ( Xmin > ScanObj->X )
    {
      Xmin = ScanObj->X;
    }

    if ( Xmax < ( ScanObj->X + ScanObj->DX ) )
    {
      Xmax = ScanObj->X + ScanObj->DX;
    }

    if ( Ymin > ScanObj->Y )
    { 
      Ymin = ScanObj->Y;
    }

    if ( Ymax < ( ScanObj->Y + ScanObj->DY ) )
    { 
      Ymax = ScanObj->Y + ScanObj->DY;
    }

    ScanObj = ScanObj->LINK;
  }

  Xmin = (float)( Xmin - XSCH_UNIT ) * XschUnitGridStep;
  Ymin = (float)( Ymin - XSCH_UNIT ) * XschUnitGridStep;
  Xmax = (float)( Xmax + XSCH_UNIT ) * XschUnitGridStep;
  Ymax = (float)( Ymax + XSCH_UNIT ) * XschUnitGridStep;

  Xmin = ( Xmin / XSCH_UNIT ) - XschPixelGridX - 1;
  Xmax = ( Xmax / XSCH_UNIT ) - XschPixelGridX + 1;
  Ymin = ( Ymin / XSCH_UNIT ) - XschPixelGridY - 1;
  Ymax = ( Ymax / XSCH_UNIT ) - XschPixelGridY + 1;

  if ( Xmin < 0 ) Xmin = 0;
  if ( Ymin < 0 ) Ymin = 0;

  if ( Xmax > XschGraphicDx ) Xmax = XschGraphicDx;
  if ( Ymax > XschGraphicDy ) Ymax = XschGraphicDy;

  if ( ( Xmax > 0              ) &&
       ( Ymax > 0              ) &&
       ( Xmin < XschGraphicDx ) &&
       ( Ymin < XschGraphicDy ) )
  {
    XschDisplayFigure( Xmin, Ymin , Xmax, Ymax );

    XschRefreshGraphicWindow( Xmin, XschGraphicDy - Ymax, 
                              Xmax - Xmin, Ymax - Ymin );
  }

  autend();
}
