/*------------------------------------------------------------\
|                                                             |
| Tool    :                    XSCH                           |
|                                                             |
| File    :                   XFS.h                           |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                   01.06.96                        |
|                                                             |
\------------------------------------------------------------*/

# ifndef XSCH_XFS
# define XSCH_XFS 

# include SCL_H
/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/

# define XSCH_OBJECT_CIRCLE      0x0000
# define XSCH_OBJECT_LINE        0x0001
# define XSCH_OBJECT_RECTANGLE   0x0002
# define XSCH_OBJECT_ARROW       0x0003
# define XSCH_OBJECT_HEXAGON     0x0004
# define XSCH_OBJECT_TEXT_CENTER 0x0005
# define XSCH_OBJECT_TEXT_LEFT   0x0006
# define XSCH_OBJECT_TEXT_RIGHT  0x0007
# define XSCH_OBJECT_TRIANGLE    0x0008
# define XSCH_OBJECT_TYPE        0x000F

# define XSCH_OBJECT_SCHCON      0x0010
# define XSCH_OBJECT_SCHNET      0x0020
# define XSCH_OBJECT_SCHBOX      0x0040

# define XSCH_OBJECT_LINE_DOWN   0x0100
# define XSCH_OBJECT_LINE_LEFT   0x0200

# define XSCH_OBJECT_SELECT      0x0400
# define XSCH_OBJECT_ACCEPT      0x0800
# define XSCH_OBJECT_CONNECT     0x1000

/*------------------------------------------------------------\
|                                                             |
|                            Macros                           |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                           Object                            |
|                                                             |
\------------------------------------------------------------*/


# define GetXschObjectType( X ) ( ( X )->TYPE & XSCH_OBJECT_TYPE )

# define IsXschLineDown( X )  ( ( X )->TYPE & XSCH_OBJECT_LINE_DOWN )
# define IsXschLineLeft( X )  ( ( X )->TYPE & XSCH_OBJECT_LINE_LEFT )
# define IsXschSchCon( X )    ( ( X )->TYPE & XSCH_OBJECT_SCHCON    )
# define IsXschSchNet( X )    ( ( X )->TYPE & XSCH_OBJECT_SCHNET    )
# define IsXschSchBox( X )    ( ( X )->TYPE & XSCH_OBJECT_SCHBOX    )

# define SetXschLineDown( X )  ( ( X )->TYPE |= XSCH_OBJECT_LINE_DOWN )
# define SetXschLineLeft( X )  ( ( X )->TYPE |= XSCH_OBJECT_LINE_LEFT )
# define SetXschSchCon( X )    ( ( X )->TYPE |= XSCH_OBJECT_SCHCON    )
# define SetXschSchNet( X )    ( ( X )->TYPE |= XSCH_OBJECT_SCHNET    )
# define SetXschSchBox( X )    ( ( X )->TYPE |= XSCH_OBJECT_SCHBOX    )

# define ClearXschLineDown( X )  ( ( X )->TYPE &= ~XSCH_OBJECT_LINE_DOWN )
# define ClearXschLineLeft( X )  ( ( X )->TYPE &= ~XSCH_OBJECT_LINE_LEFT )
# define ClearXschSchCon( X )    ( ( X )->TYPE &= ~XSCH_OBJECT_SCHCON    )
# define ClearXschSchNet( X )    ( ( X )->TYPE &= ~XSCH_OBJECT_SCHNET    )
# define ClearXschSchBox( X )    ( ( X )->TYPE &= ~XSCH_OBJECT_SCHBOX    )

/*------------------------------------------------------------\
|                                                             |
|                           Select                            |
|                                                             |
\------------------------------------------------------------*/

# define IsXschSelect( X )    ( ( X )->TYPE & XSCH_OBJECT_SELECT   )
# define SetXschSelect( X )   ( ( X )->TYPE |= XSCH_OBJECT_SELECT  )
# define ClearXschSelect( X ) ( ( X )->TYPE &= ~XSCH_OBJECT_SELECT )

/*------------------------------------------------------------\
|                                                             |
|                           Accept                            |
|                                                             |
\------------------------------------------------------------*/

# define IsXschAccept( X )    ( ( X )->TYPE & XSCH_OBJECT_ACCEPT   )
# define SetXschAccept( X )   ( ( X )->TYPE |= XSCH_OBJECT_ACCEPT  )
# define ClearXschAccept( X ) ( ( X )->TYPE &= ~XSCH_OBJECT_ACCEPT )

/*------------------------------------------------------------\
|                                                             |
|                           Connect                           |
|                                                             |
\------------------------------------------------------------*/

# define IsXschConnect( X )    ( ( X )->TYPE & XSCH_OBJECT_CONNECT   )
# define SetXschConnect( X )   ( ( X )->TYPE |= XSCH_OBJECT_CONNECT  )
# define ClearXschConnect( X ) ( ( X )->TYPE &= ~XSCH_OBJECT_CONNECT )

/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Object                           |
|                                                             |
\------------------------------------------------------------*/

  typedef struct xschobj_list
  {
    struct xschobj_list *NEXT;
    struct xschobj_list *LINK;
    char                *NAME;
    short                LAYER;
    long                 TYPE;
    long                 X;
    long                 Y;
    long                 DX;
    long                 DY;
    void                *USER;

  } xschobj_list;

  typedef struct xschfig_list
  {
    struct xschfig_list *NEXT;
    char                *NAME;
    xschobj_list        *OBJECT [ XSCH_MAX_LAYER ];

  } xschfig_list;

/*------------------------------------------------------------\
|                                                             |
|                            Select                           |
|                                                             |
\------------------------------------------------------------*/

  typedef struct xschselect_list 
  {
    struct xschselect_list *NEXT;
    xschobj_list           *OBJECT;

  } xschselect_list;

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  extern long            XschBoundXmin;
  extern long            XschBoundYmin;
  extern long            XschBoundXmax;
  extern long            XschBoundYmax; 
  extern char            XschRecomputeBound;

  extern xschfig_list    *XschFigure;
  extern schfig_list     *XschFigureSch;

  extern xschselect_list *XschHeadSelect;
  extern xschselect_list *XschHeadConnect;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Add Functions                      |
|                                                             |
\------------------------------------------------------------*/

  extern xschfig_list   *XschAddFigure();
  extern xschobj_list   *XschAddCircle();
  extern xschobj_list   *XschAddLine();
  extern xschobj_list   *XschAddRectangle();
  extern xschobj_list   *XschAddTriangle();
  extern xschobj_list   *XschAddHexagon();
  extern xschobj_list   *XschAddText();

/*------------------------------------------------------------\
|                                                             |
|                          Del Functions                      |
|                                                             |
\------------------------------------------------------------*/

  extern void XschDelFigure();

/*------------------------------------------------------------\
|                                                             |
|                   Flag Set & Clear Functions                |
|                                                             |
\------------------------------------------------------------*/

  extern void XschSelectObject();
  extern void XschUnselectObject();
  extern void XschAcceptObject();
  extern void XschRejectObject();
  extern void XschConnectObject();
  extern void XschDisconnectObject();

/*------------------------------------------------------------\
|                                                             |
|                          Select Functions                   |
|                                                             |
\------------------------------------------------------------*/

  extern void XschAddSelect();
  extern void XschDelSelect();
  extern void XschPurgeSelect();

/*------------------------------------------------------------\
|                                                             |
|                          Connect Functions                  |
|                                                             |
\------------------------------------------------------------*/

  extern void XschAddConnect();
  extern void XschDelConnect();

/*------------------------------------------------------------\
|                                                             |
|                          Bound Compute                      |
|                                                             |
\------------------------------------------------------------*/

  extern char XschComputeBound();

/*------------------------------------------------------------\
|                                                             |
|                        Load Functions                       |
|                                                             |
\------------------------------------------------------------*/

 extern void XschLoadFigure();

/*------------------------------------------------------------\
|                                                             |
|                    Error Message Functions                  |
|                                                             |
\------------------------------------------------------------*/

  extern char *XschGetErrorMessage();
  extern void  XschExitErrorMessage();
  extern void  XschInitializeErrorMessage();

/*------------------------------------------------------------\
|                                                             |
|                    Informations Functions                   |
|                                                             |
\------------------------------------------------------------*/

  extern char *XschGetInformations();

# endif
