/*------------------------------------------------------------\
|                                                             |
| Tool    :                   XSCH                            |
|                                                             |
| File    :                 Object.c                          |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                   01.06.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include <signal.h>
# include <string.h>
# include <setjmp.h>
# include MUT_H
# include AUT_H
# include MLO_H
# include SCL_H
# include XSB_H 
# include XSC_H
# include "XSC_place.h"
# include "XSC_object.h"
# include "XSC_schem.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  xschfig_list *XschFigure = (xschfig_list *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                           Functions                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Alloc Functions                     |
|                                                             |
\------------------------------------------------------------*/

xschfig_list *XschAllocFigure()
{
  return( (xschfig_list *)autallocblock( sizeof( xschfig_list ) ) );
}

xschobj_list *XschAllocObject()
{
  return( (xschobj_list *)autallocheap( sizeof( xschobj_list ) ) );
}

/*------------------------------------------------------------\
|                                                             |
|                         Free Functions                      |
|                                                             |
\------------------------------------------------------------*/

void XschFreeFigure( Figure )

  xschfig_list *Figure;
{
  autfreeblock( Figure );
}

void XschFreeObject( Object )

  xschobj_list *Object;
{
  autfreeheap( Object, sizeof( xschobj_list ) );
}

/*------------------------------------------------------------\
|                                                             |
|                         Add Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          XschAddCircle                      |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddCircle( X, Y, R, Layer )

  long  X;
  long  Y;
  long  R;
  short Layer;
{
  xschobj_list *Object;

  autbegin();

  Object        = XschAllocObject();
  Object->TYPE  = XSCH_OBJECT_CIRCLE;
  Object->LAYER = Layer;
  Object->NEXT  = XschFigure->OBJECT[ Layer ];
  Object->X     = X - R;
  Object->Y     = Y - R;
  Object->DX    = ( R << 1 );
  Object->DY    = ( R << 1 );

  XschFigure->OBJECT[ Layer ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddLine                        |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddLine( X1, Y1, X2, Y2, LayerFrom, LayerTo )

  long  X1;
  long  Y1;
  long  X2;
  long  Y2;
  short LayerFrom;
  short LayerTo;
{
  xschobj_list *Object;

  autbegin();

  Object        = XschAllocObject();
  Object->LAYER = LayerFrom;
  Object->TYPE  = XSCH_OBJECT_LINE;
  Object->NEXT  = XschFigure->OBJECT[ LayerTo ];

  if ( X1 > X2 )
  {
    SetXschLineLeft( Object );

    Object->X  = X2;
    Object->DX = X1 - X2;
  }
  else
  {
    Object->X  = X1;
    Object->DX = X2 - X1;
  }

  if ( Y1 > Y2 )
  {
    SetXschLineDown( Object );

    Object->Y  = Y2;
    Object->DY = Y1 - Y2;
  }
  else
  {
    Object->Y  = Y1;
    Object->DY = Y2 - Y1;
  }

  XschFigure->OBJECT[ LayerTo ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddArrow                       |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddArrow( X1, Y1, X2, Y2, LayerFrom, LayerTo )

  long  X1;
  long  Y1;
  long  X2;
  long  Y2;
  short LayerFrom;
  short LayerTo;
{
  xschobj_list *Object;

  autbegin();

  Object        = XschAllocObject();
  Object->LAYER = LayerFrom;
  Object->TYPE  = XSCH_OBJECT_ARROW;
  Object->NEXT  = XschFigure->OBJECT[ LayerTo ];

  if ( X1 > X2 )
  {
    SetXschLineLeft( Object );

    Object->X  = X2;
    Object->DX = X1 - X2;
  }
  else
  {
    Object->X  = X1;
    Object->DX = X2 - X1;
  }

  if ( Y1 > Y2 )
  {
    SetXschLineDown( Object );

    Object->Y  = Y2;
    Object->DY = Y1 - Y2;
  }
  else
  {
    Object->Y  = Y1;
    Object->DY = Y2 - Y1;
  }

  XschFigure->OBJECT[ LayerTo ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddTriangle                    |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddTriangle( X, Y, DX, DY, Layer )

  long  X;
  long  Y;
  long  DX;
  long  DY;
  short Layer;
{
  xschobj_list  *Object;

  autbegin();

  Object        = XschAllocObject();
  Object->TYPE  = XSCH_OBJECT_TRIANGLE;
  Object->LAYER = Layer;
  Object->NEXT  = XschFigure->OBJECT[ Layer ];
  Object->X     = X;
  Object->Y     = Y;
  Object->DX    = DX;
  Object->DY    = DY;

  XschFigure->OBJECT[ Layer ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddRectangle                   |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddRectangle( X, Y, DX, DY, Layer )

  long  X;
  long  Y;
  long  DX;
  long  DY;
  short Layer;
{
  xschobj_list  *Object;

  autbegin();

  Object        = XschAllocObject();
  Object->TYPE  = XSCH_OBJECT_RECTANGLE;
  Object->LAYER = Layer;
  Object->NEXT  = XschFigure->OBJECT[ Layer ];
  Object->X     = X;
  Object->Y     = Y;
  Object->DX    = DX;
  Object->DY    = DY;

  XschFigure->OBJECT[ Layer ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddText                        |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddText( X, Y, Type, Name, Layer )

  long  X;
  long  Y;
  long  Type;
  char *Name;
  short Layer;
{
  xschobj_list *Object;
  long          Width;

  autbegin();

  Width = strlen( Name ) * XSCH_UNIT * 2 / 3;

  if ( Type == XSCH_OBJECT_TEXT_CENTER ) 
  {
    X = X - ( Width >> 1 );
  }

  if ( Type == XSCH_OBJECT_TEXT_LEFT  )
  {
    X = X - Width;
  }

  Object        = XschAllocObject();
  Object->LAYER = Layer;
  Object->TYPE  = Type;
  Object->NEXT  = XschFigure->OBJECT[ Layer ];
  Object->NAME  = namealloc( Name );
  Object->X     = X;
  Object->Y     = Y;
  Object->DX    = Width;


  XschFigure->OBJECT[ Layer ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddHexagon                     |
|                                                             |
\------------------------------------------------------------*/

xschobj_list *XschAddHexagon( X, Y, DX, DY, Layer )

  long  X;
  long  Y;
  long  DX;
  long  DY;
  short Layer;
{
  xschobj_list  *Object;

  autbegin();

  Object        = XschAllocObject();
  Object->TYPE  = XSCH_OBJECT_HEXAGON;
  Object->LAYER = Layer;
  Object->NEXT  = XschFigure->OBJECT[ Layer ];
  Object->X     = X;
  Object->Y     = Y;
  Object->DX    = DX;
  Object->DY    = DY;

  XschFigure->OBJECT[ Layer ] = Object;

  autend();
  return( Object );
}

/*------------------------------------------------------------\
|                                                             |
|                          XschAddFigure                      |
|                                                             |
\------------------------------------------------------------*/

xschfig_list *XschAddFigure()
{
  autbegin();

  XschFigure = XschAllocFigure();
  XschFigure->NAME = XschFigureSch->NAME;

  XschTraceFigure(XschFigureSch);

  autend();
  return( XschFigure );
}

/*------------------------------------------------------------\
|                                                             |
|                         Load Functions                      |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          XschLoadFigure                     |
|                                                             |
\------------------------------------------------------------*/

void XschLoadFigure( FileName )

  char *FileName;
{
  int Mode;

  autbegin();

  if ( XschCnsMode ) Mode = SCH_FROM_CNS;
  else               Mode = SCH_FROM_LO;

  XschFigureSch = (schfig_list *)Xschgetschfig( FileName , Mode );

  if ( XschFigureSch != (schfig_list *)NULL )
  {
    XschAddFigure();
  }

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                         Del Functions                       |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          XschDelFigure                      |
|                                                             |
\------------------------------------------------------------*/

void XschDelFigure()
{
  xschobj_list   *ScanObject;
  xschobj_list   *DelObject;
  short           Layer;

  if ( XschFigure == (xschfig_list *)NULL ) return;

  autbegin();

  for ( Layer = 0; Layer < XSCH_MAX_LAYER; Layer++ )
  {
    ScanObject = XschFigure->OBJECT[ Layer ];

    while ( ScanObject != (xschobj_list *)NULL )
    {
      DelObject  = ScanObject;
      ScanObject = ScanObject->NEXT;

      XschFreeObject( DelObject );
    }
  }

  XschFreeFigure( XschFigure );

  delschfig( XschFigureSch );

  XschFigure    = (xschfig_list *)NULL;
  XschFigureSch = (schfig_list  *)NULL;

  autend();
}
