/*------------------------------------------------------------\
|                                                             |
| Tool    :                    XVPN                           |
|                                                             |
| File    :                  Colors.c                         |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <Xm/Xm.h>
 
# include MUT_H
# include AUT_H
# include XSB_H
# include XMX_H
# include "XMX_color.h" 

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  GC   XvpnBackgroundGC;
  GC   XvpnGridGC;
  GC   XvpnAcceptDrawGC;
  GC   XvpnConnectDrawGC;
  GC   XvpnLayerDrawGC[ XVPN_MAX_LAYER ];

  GC   XvpnSmallTextGC;
  GC   XvpnMediumTextGC;
  GC   XvpnLargeTextGC;
  GC   XvpnXorGC;

  XFontStruct *XvpnLargeTextFont;
  XFontStruct *XvpnMediumTextFont;
  XFontStruct *XvpnSmallTextFont;

/*------------------------------------------------------------\
|                                                             |
|                      Layer Color Define                     |
|                                                             |
\------------------------------------------------------------*/

 static char **XvpnColorName[ 5 ] =

  {
    &XVPN_BACKGROUND_COLOR_NAME,
    &XVPN_FOREGROUND_COLOR_NAME,
    &XVPN_CURSOR_COLOR_NAME,
    &XVPN_ACCEPT_COLOR_NAME,
    &XVPN_CONNECT_COLOR_NAME
  };

  static int XvpnColor[ XVPN_MAX_COLOR ];

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      XvpnInitializeColors                   |
|                                                             |
\------------------------------------------------------------*/

void XvpnInitializeColors()

{
  XColor   ForgetIt;
  XColor   GetIt;
  int      Counter;
  Colormap ColorMap;
  char     Error;
  char    *ColorName;

  autbegin();

  ColorMap  = DefaultColormapOfScreen ( XtScreen ( XvpnGraphicWindow ) );
  Error     = False;

  for ( Counter = 0; Counter < XVPN_MAX_COLOR; Counter++ )
  {
    if ( Counter < 5 )
    {
       ColorName = *XvpnColorName[ Counter ];
    }
    else
    {
      ColorName = XVPN_LAYER_NAME_TABLE[ Counter - 5 ][ 1 ];
    }

    if ( ColorName != (char *)NULL )
    {
      if ( ! XAllocNamedColor( XvpnGraphicDisplay, 
                               ColorMap, ColorName, 
                              &GetIt, &ForgetIt ) ) 
      {
        fprintf ( stderr, "\nXmx100: Color name '%s' is not in X11 database !\n",
                          ColorName );

        Error = True;
      }

      XvpnColor[ Counter ] = GetIt.pixel;
    }
    else 
    {
      XvpnColor[ Counter ] = XvpnColor[ 0 ];
    }
  }

  if ( Error == True ) exit( 1 );

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                  XvpnInitializeGraphicContext               |
|                                                             |
\------------------------------------------------------------*/

void XvpnInitializeGraphicContext()

{
  XGCValues GraphicValue;
  int       Counter;
  Window    Root;

  autbegin();

  XvpnLargeTextFont = XLoadQueryFont( XvpnGraphicDisplay,
                                       XVPN_LARGE_TEXT_FONT_NAME );

  XvpnMediumTextFont = XLoadQueryFont( XvpnGraphicDisplay,
                                        XVPN_MEDIUM_TEXT_FONT_NAME );

  XvpnSmallTextFont = XLoadQueryFont( XvpnGraphicDisplay,
                                       XVPN_SMALL_TEXT_FONT_NAME );

  GraphicValue.foreground = XvpnColor[ 0 ];  /* Black */
  GraphicValue.background = XvpnColor[ 0 ];
  GraphicValue.plane_mask = AllPlanes;

  Root = RootWindowOfScreen ( XtScreen ( XvpnGraphicWindow ) ),

  XvpnBackgroundGC = XCreateGC ( XvpnGraphicDisplay, 
                                  Root,
                                  ( GCForeground |
                                    GCBackground |
                                    GCPlaneMask ),
                                    &GraphicValue );

  GraphicValue.foreground = XvpnColor[ 1 ];  /* White */

  XvpnGridGC = XCreateGC ( XvpnGraphicDisplay, 
                            Root,
                            ( GCForeground |
                              GCBackground |
                              GCPlaneMask ),
                              &GraphicValue );

  GraphicValue.foreground = XvpnColor[ 1 ];  /* Text  */

  XvpnSmallTextGC = XCreateGC ( XvpnGraphicDisplay, 
                                 Root,
                                 ( GCForeground |
                                   GCBackground |
                                   GCPlaneMask ),
                                   &GraphicValue );

  XvpnMediumTextGC = XCreateGC ( XvpnGraphicDisplay, 
                                  Root,
                                  ( GCForeground |
                                    GCBackground |
                                    GCPlaneMask ),
                                   &GraphicValue );

  XvpnLargeTextGC = XCreateGC ( XvpnGraphicDisplay, 
                                 Root,
                                 ( GCForeground |
                                   GCBackground |
                                   GCPlaneMask ),
                                  &GraphicValue );

  XSetFont( XvpnGraphicDisplay,
            XvpnSmallTextGC, XvpnSmallTextFont->fid );

  XSetFont( XvpnGraphicDisplay,
            XvpnMediumTextGC, XvpnMediumTextFont->fid );

  XSetFont( XvpnGraphicDisplay,
            XvpnLargeTextGC, XvpnLargeTextFont->fid );

  GraphicValue.background = XvpnColor[ 2 ];  /* Xor Color */
  GraphicValue.foreground = XvpnColor[ 2 ];  /* Xor Color */
  GraphicValue.function   = GXxor;

  XvpnXorGC = XCreateGC ( XvpnGraphicDisplay, 
                           Root,
                           ( GCForeground |
                             GCBackground |
                             GCFunction   |
                             GCPlaneMask ),
                             &GraphicValue );

  GraphicValue.background = XvpnColor[ 0 ];  /* Black        */
  GraphicValue.foreground = XvpnColor[ 3 ];  /* Accept Color */

  XvpnAcceptDrawGC = XCreateGC ( XvpnGraphicDisplay,
                                 Root,
                                 ( GCForeground |
                                   GCBackground |
                                   GCPlaneMask ),
                                   &GraphicValue );

  GraphicValue.foreground = XvpnColor[ 4 ];  /* Connect Color */

  XvpnConnectDrawGC = XCreateGC ( XvpnGraphicDisplay,
                                  Root,
                                  ( GCForeground |
                                    GCBackground |
                                    GCPlaneMask ),
                                    &GraphicValue );

  for ( Counter = 0; Counter < XVPN_MAX_LAYER; Counter++ )
  {
    GraphicValue.foreground = XvpnColor[ Counter + 5 ];

    XvpnLayerDrawGC[ Counter ] =

         XCreateGC ( XvpnGraphicDisplay,
                     Root,
                     ( GCForeground |
                       GCBackground |
                       GCPlaneMask  ),
                      &GraphicValue  );
  }

  autend();
}
