/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include MLO_H
#include MUT_H 
#include PPT_H
#include PAT_H
#include "genbist.h"
#include "utile.h"
#include "environment.h"
#include "scan_path.h"


lofig_list *ptfig;     /* pointer to the global netlist */

int gates_nbr1  = 0;      /* number of gates in the input netlist */
int gates_nbr2  = 0;      /* number of gates in the target netlist */
int sig_nbr1    = 0;      /* number of signals in the input netlist */
int sig_nbr2    = 0;      /* number of gates in the target netlist */
int scan_length  ;        /* size of the scan-path for bilbo and misr/lfsr architecures */ 


int nbr_inputs  = 0;      /* number of input signals */
int nbr_outputs = 0;      /* number of output signals */ 
int mem_nbr     = 0;      /* number of Dff in the netlist */
int nbr_scan    = 0;      /* number of dffs in the scan chain */

lofig_list *misr_fig = NULL;  /* pointer to the misr generated model */
lofig_list *lfsr_fig = NULL;  /* pointer to the lfsr generated model */
lofig_list *mux_fig = NULL;   /* pointer to the mux  generated model */
lofig_list *bilbo_fig = NULL; /* pointer to the bilbo generated model */

struct chain *scan_ptr = NULL;/* list of the ffs in the scan path */
struct chain *scan_order = NULL;/* list of the ff in the scanpath_order file */
struct chain *mem_ptr = NULL;/* list of memory cells */
struct chain *in_ptr;/* list of input connectors */
struct chain *out_ptr ;/* list of output connectors */



/*********************************** parameter variables *********************************************/

char *info_filename = NULL;
char *genpat_filename = NULL;
char *bistgenpat_filename = NULL;
char *scanorder_filename = NULL;
char *pat_filename       = NULL;

int bilbo         = 0; /* true for -b option to select bilbo architeture */
int hierarchical  = 0; /* true for -h option */
int name_saves    = 0; /* true for -c option */
int nobist        = 1; /* true for -nobist option */
int scan_insert   = 1; /* true for -s option */



/*******************************************************************************************************************/
/*                                                                                                                 */
/*                                         maion program                                                           */
/*                                                                                                                 */
/*******************************************************************************************************************/

main(argc,argv)
int argc;
char *argv[];
{
    locon_list *ptrc;
    losig_list *lfsr_scan_in, *lfsrscout_misrscin,*misr_scan_out, *ptr_sig, *scan_signal;
    chain_list *ptr, *ptr1, *ptr2, *ptr3, *ptr_i1, *ptr_i0;
	
    int i,j;
    int nbr_sig1;         /* number of signals in the input netlist */
    int nbr_gates1;       /* number of gates in the input netlist */

	char *input_file, *output_file;
    char *option;
    char *insname;		
	
    int comb_circuit = 0; /* true if the circuit doesn't contain register gates */
	int _exchange = 0;
	int nbr_cycles = 100;       /* number of cycles to execute in bist mode */
    
/***************************************************** Get parameters ***********************************************/
	

	if ((argc < 3) || (argc >4)) 
		Usage(argv[0]);
	else
	{
		input_file = argv[1];
		output_file = argv[2];
	        if(argc == 4) scanorder_filename =argv[3];
	}
	
	
	mbkenv();
	if((ptfig = getlofig(input_file,'A'))==NULL)
	{ 
		printf("unable to open figure %s \n",input_file);
		exit(1);
	}
	
			
/**********************************************************************************************************************/
/*                                                                                                                    */
/*                                                                                                                    */
/*                                                                                                                    */
/**********************************************************************************************************************/

/* get_signals_names(info_filename); */

    get_environment_variables();
	
/* exit(1);*/

	
	get_special_signals(ptfig);

	get_external_connectors(ptfig);

	max_index = get_max_index(ptfig);

	mem_ptr = get_mem_cells(ptfig);

	if ( mem_ptr == NULL) comb_circuit = 1;


if (scan_insert)
{
	insert_scan(ptfig,comb_circuit,scanorder_filename);
}

if (scan_insert || !nobist )
{
        ptfig->NAME = namealloc(output_file);
        printf("\n");
        printf("  saving figure file ......\n");
        savelofig(ptfig);
}

    freechain(mem_ptr);
    freechain(in_ptr);
    freechain(out_ptr);
    freechain(scan_ptr);

exit(0);
}




